/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Copyright Projet JRL-Japan, 2007
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*
* File:      SoTGenericRobot.cc
* Project:   SOT
* Author:    Nicolas Mansard from Sovan Hack from Francois Kiss 
*
* Version control
* ===============
*
*  $Id$
*
* Description
* ============
*
*
* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* -------------------------------------------------------------------------- */
/* --- INCLUDE -------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */

#include "AFSoT/SoTGenericRobot.hh"

#include <AFScene/MultiBody.hh>
#include <AFUtils/AFMacros.hh>

/* -------------------------------------------------------------------------- */
/* --- SoTGenericRobot Definition ------------------------------------------- */
/* -------------------------------------------------------------------------- */

namespace AFSoT
{

  SoTGenericRobot::SoTGenericRobot( AFScene::MultiBody* mb,
				    const std::string & name )
    : SoTMultiBody(mb, name )
    ,_controlSIN(NULL,name + "::input(vector40)::control")
    ,_attitudeSIN(NULL,name + "::input(vector3)::attitudeIN")
    ,_stateSOUT( name + "::output(vector46)::state" )
    ,_attitudeSOUT( name + "::output(matrixRot)::attitude" )

    ,_suspend( true )
    ,_mb(mb)
  {
    _numParam = _mb->getNumParameters();
    _initQ.Resize(_numParam);
    _initQ.Fill(0.0);
    _mbState.Resize(_numParam);
    _controlLaw.resize(_numParam);
    
    signalRegistration( _controlSIN
			<< _attitudeSIN
			<< _stateSOUT
			<<_attitudeSOUT );
  }

  SoTGenericRobot::~SoTGenericRobot() {}
  void SoTGenericRobot::initialize( ) {}
  void SoTGenericRobot::newStep() {}

  void SoTGenericRobot::runStep(unsigned iter)
  {
    _iter = iter;
    writeState(iter);
    try {_periodicCallBefore.run( _iter ); } catch (...) {}
    readControlLaw();
    try { _periodicCallAfter.run( _iter ); } catch (...) {}
  }

  void SoTGenericRobot::getControlLaw(VNL::Vector<double> & v)
  {
    v.Resize( _controlLaw.size() );
    for (unsigned i = 0; i < _numParam; ++i)
      { v[i] = _controlLaw(i); }
  }
  
  void SoTGenericRobot::writeState(unsigned iter)
  {
    _mbState.Resize(_numParam);
    _mb->getStateVector(_mbState);
    ml::Vector mlrs(_numParam+6);
    for( int i=0;i<6;++i ) {mlrs(i)=0.;}
    for( unsigned int i=0;i<_numParam;++i ) {mlrs(i+6) = _mbState[i]; }
    _stateSOUT.setConstant( mlrs );
    _stateSOUT.setTime( iter );
  }

  void SoTGenericRobot::readControlLaw ()
  {
    _controlLaw.fill(.0);
    try{
      if(! _suspend ) _controlLaw = _controlSIN(_iter);
    } catch( const sotExceptionAbstract& e ) {sotDEBUG(1) << e;}
    catch( ... ) { sotDEBUG(1) << "Unknown catched." <<endl; }
    sotDEBUG(25) << _controlLaw <<endl;

    for( unsigned int i=0;i<_numParam;++i )
      {
	if( isnan(_controlLaw(i)) ) 
	  {	
	    sotDEBUG(1) << "!! Plug ->MC broken."<<endl;
	    _controlSIN.unplug();
	    _controlLaw.fill(.0); 
	    break;
	  }
      }
  }

  /* ------------ SoT functions ------------------ */
  void SoTGenericRobot::commandLine( const std::string& cmdLine,
				     std::istringstream& cmdArgs,
				     std::ostream& os )
  {
    if( cmdLine =="help" )
      {
	os << "Plugin " << name << ": " << std::endl
	   << " - play/pause: \tStart/Stop writing the controlSIN"
	   << " in the robot MC." << endl
	   << " - reinit [from {MC|RS}]: \treinit the current pos value (default MC)" <<endl
	   << " - periodicCall{Before|After}:\tAccess to the periodic caller. "
	   << "\'Before\' and \'After\' refeere to the MC computation." <<endl;
	sotEntity::commandLine(cmdLine,cmdArgs,os);
      }
    else if( cmdLine == "print" ) 
      {
	os << getName() << " [mode=" << ((_suspend)?"pause":"play") <<"]."<<endl;
      }
    else if( cmdLine == "pause" ) { pause(); }
    else if( cmdLine == "play" )  { play(); }
    else if( cmdLine == "reinit" )
      { /* for compatibility only. */    }
    // Kept two tests for compatibility reason. 
    else if(( cmdLine == "periodicCall" )||( cmdLine == "periodicCallAfter" ))
      { 
	string cmd2; cmdArgs >> cmd2;
	_periodicCallAfter .commandLine( cmd2,cmdArgs,os );
      }
    else if( cmdLine == "periodicCallBefore" )
      { 
	string cmd2; cmdArgs >> cmd2;
	_periodicCallBefore.commandLine( cmd2,cmdArgs,os );
      }
    else sotEntity::commandLine(cmdLine,cmdArgs,os);
  }

  void SoTGenericRobot::play( void ) { _suspend = false; }
  void SoTGenericRobot::pause( void ) { _suspend = true;}
}
