/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Copyright Projet JRL-Japan, 2007
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*
* File:      SotMainProgram.cc
* Project:   SOT
* Author:    Sovan Hack from Francois Kiss from Nicolas Mansard from Paul Evrard
*
* Version control
* ===============
*
*  $Id$
*
* Description
* ============
*
*
* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* -------------------------------------------------------------------------- */
/* --- INCLUDE -------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */

#include "AFSoT/SoTHRP2.hh"

#include <AFScene/MultiBody.hh>
#include <AFUtils/AFMacros.hh>

/* -------------------------------------------------------------------------- */
/* --- SoTHRP2 Definition --------------------------------------------------- */
/* -------------------------------------------------------------------------- */

namespace AFSoT
{

	SoTHRP2::SoTHRP2 (
		AFScene::MultiBody* mb,
		const std::string & name
		) : SoTMultiBody(mb, name )
		,_controlSIN(NULL,name + "::input(vector40)::control")
		,_attitudeSIN(NULL,name + "::input(vector3)::attitudeIN")
		,_zmpSIN(NULL,name + "::input(vector3)::zmp")
		,_stateSOUT( name + "::output(vector46)::state" )
		,_attitudeSOUT( name + "::output(matrixRot)::attitude" )
		,_pseudoTorqueSOUT( name + "::output(vector)::ptorque" )
		,_previousControlSOUT( name + "::output(vector)::previousControl" )

		,_suspend( true )
		,_withForceSignals()
		,_activatePseudoTorqueSignal( false )
		,_activatePreviousControlSignal( false )

		,_mb(mb)
	{
		_numParam = _mb->getNumParameters();
		_initQ.Resize(_numParam);
		_initQ.Fill(0.0);
		_mbState.Resize(_numParam);
		_mlmc.resize(_numParam);
		_mlmcVNL.Resize(_numParam);

		_forcesSOUT.resize(4);

		for( int i=0;i<4;++i ){ _withForceSignals[i] = false; }
		_forcesSOUT[0] =
			new sotSignal<ml::Vector, int>(name + "::output(vector6)::forceRLEG");
		_forcesSOUT[1] =
			new sotSignal<ml::Vector, int>(name + "::output(vector6)::forceLLEG");
		_forcesSOUT[2] =
			new sotSignal<ml::Vector, int>(name + "::output(vector6)::forceRARM");
		_forcesSOUT[3] =
			new sotSignal<ml::Vector, int>(name + "::output(vector6)::forceLARM");

		signalRegistration( _controlSIN
			<< _attitudeSIN
			<< _zmpSIN
			<< _stateSOUT
			<<*(_forcesSOUT[0])
			<<*(_forcesSOUT[1])
			<<*(_forcesSOUT[2])
			<<*(_forcesSOUT[3])
			<<_attitudeSOUT
			<<_pseudoTorqueSOUT
			<<_previousControlSOUT);
	}

	SoTHRP2::~SoTHRP2()
	{}

	void SoTHRP2::initialize( )
	{
	}

	void SoTHRP2::newStep()
	{
	}

	void SoTHRP2::runStep(unsigned iter)
	{
	  _iter = iter;
	  writeState(iter);

	  /* --- FORCES --- */
	  ml::Vector mlforces(6);
	  for( int i=0;i<4;++i )
	    {
	      if( _withForceSignals[i] == true )
		{
		  for( int j=0;j<6;++j )
		    mlforces(j) = 0;//??How can we do that with AF??   
		  _forcesSOUT[i]->setConstant( mlforces );
		  _forcesSOUT[i]->setTime( _iter );
		}
	    }
		
	  /* --- PERIODIC CALL --- */
	  try {_periodicCallBefore.run( _iter ); } catch (...) {}


	  /* --- Compute the control law. ------------------------------------------- */
	  readMC ( );

	  /* --- PERIODIC CALL --- */
	  try { _periodicCallAfter.run( _iter ); } catch (...) {}
	}

	void SoTHRP2::getControlLaw(VNL::Vector<double> & v)
	{
	  readMC ( );
	  v = _mlmcVNL;
	}


	void SoTHRP2::writeState(unsigned iter)
	{
	  _mbState.Resize(_numParam);
	  _mb->getStateVector(_mbState);
	  ml::Vector mlrs(_numParam+6);
	  for( int i=0;i<6;++i ) {mlrs(i)=0.;}
	  for( unsigned int i=0;i<_numParam;++i ) {mlrs(i+6) = _mbState[i]; }
	  _stateSOUT.setConstant( mlrs );
	  _stateSOUT.setTime( iter );
	}

	void SoTHRP2::readMC ()
	{
	  _mlmc.fill(.0);
	  try{
	    if(! _suspend ) _mlmc = _controlSIN(_iter);
	  }catch( const sotExceptionAbstract& e ) {sotDEBUG(1) << e;}
	  catch( ... ) { sotDEBUG(1) << "Unknown catched." <<endl; }
	  sotDEBUG(25) << _mlmc <<endl;

		//	if( mlmc.size()!=ROBOT_SIZE ) 
		//	{
		//		sotDEBUG(1) << "!! Size incorrect: Plug ->MC broken."<<endl;
		//		controlSIN.unplug();
		//		mlmc.resize(ROBOT_SIZE); mlmc.fill(0.); 
		//	}

	  for( unsigned int i=0;i<_numParam;++i )
	    {
	      if( isnan(_mlmc(i)) ) 
		{	
		  sotDEBUG(1) << "!! Plug ->MC broken."<<endl;
		  _controlSIN.unplug();
		  _mlmc.fill(.0); 
		  break;
		}
	    }
	  
	  /* Correction of the angles orientation for compatibility with
	   * the OpenHRP-2 model (used in the SOT by default). */
	  /* For left hand. */
	  if( _numParam>=40 )
	    {
	      _mlmc(36) = _mlmc(38) = _mlmc(29);
	      _mlmc(35) = _mlmc(37) = _mlmc(39) = - _mlmc(29);
	    }

	  /* For right hand. */
	  if( _numParam>=35 )
	    {
	      _mlmc(31) = _mlmc(33) = _mlmc(22);
	      _mlmc(30) = _mlmc(32) = _mlmc(34) = - _mlmc(22);
	    }

	  /* Translate the vector from MAAL/SOT to VNL/AF. */
	  for (unsigned i = 0; i < _numParam; ++i)
	    _mlmcVNL[i] = _mlmc(i);
	}

	//------------ SoT functions ------------------//
	void SoTHRP2::commandLine( const std::string& cmdLine,
				   std::istringstream& cmdArgs,
				   std::ostream& os )
	{
		if( cmdLine =="help" )
		{
			os << "Plugin " << name << ": " << std::endl
				<< " - play/pause: \tStart/Stop writing the controlSIN"
				<< " in the robot MC." << endl
				<< " - reinit [from {MC|RS}]: \treinit the current pos value (default MC)" <<endl
				<< " - withForces <i> \tCopy the force sensor data from sensor " 
				<< "<i> in the corresponding SOUT." <<endl
				<< " - whichForces: \tDisplay the activated force signals." <<endl
				<< " - with{PseudoTorque|PreviousControl} [{true|false}]: "
				<< "activate/inactivate the signal." << endl
				<< " - periodicCall{Before|After}:\tAccess to the periodic caller. "
				<< "\'Before\' and \'After\' refeere to the MC computation." <<endl
				<< " - timestep [<value>]: get/set the timestep (=0.005)." << std::endl;
			sotEntity::commandLine(cmdLine,cmdArgs,os);
		}
		else if( cmdLine == "print" ) 
		{
			os << getName() << " [mode=" << ((_suspend)?"pause":"play") <<"]."<<endl;
		}
		else if( cmdLine == "pause" ) { pause(); }
		else if( cmdLine == "play" )  { play(); }
		else if( cmdLine == "reinit" )
		{ /* for compatibility only. */    }
		else if( cmdLine == "withForces" )
		{
			int index;
			cmdArgs >> index;
			if((index >= 0) && (index < 4))
				cmdArgs >> _withForceSignals[index];
		}
		else if( cmdLine == "whichForces" )
		{
			os << "Force signals: " << endl; 
			for( unsigned int i=0;i<4;++i )
			{
				os << "\t- Force " << i << ": ";
				if( _withForceSignals[i] )
				{ os << "Active";	  } else { os << "Inactive"; } 
				os << endl;
			}
		}
		else if( cmdLine == "withPseudoTorque" )
		{
			cmdArgs >> std::ws; if( cmdArgs.good() )
			{
				std::string val; cmdArgs>>val; 
				_activatePseudoTorqueSignal = ( (val=="true")||(val=="1") );
			} else {
				os << "withPseudoTorque = " << _activatePseudoTorqueSignal 
				   << std::endl;
			}
		}
		else if( cmdLine == "withPreviousControl" )
		{
			/* For compatibility only. */
		}
		// Kept two tests for compatibility reason. 
		else if(( cmdLine == "periodicCall" )||( cmdLine == "periodicCallAfter" ))
		{ 
			string cmd2; cmdArgs >> cmd2;
			_periodicCallAfter .commandLine( cmd2,cmdArgs,os );
		}
		else if( cmdLine == "periodicCallBefore" )
		{ 
			string cmd2; cmdArgs >> cmd2;
			_periodicCallBefore.commandLine( cmd2,cmdArgs,os );
		}
		//else if( cmdLine == "timestep" )
		//{ 
		//	cmdArgs >> std::ws;
		//	if( cmdArgs.good() ) { cmdArgs >> timestep; } 
		//	else { os << "timestep = " << timestep << std::endl; }

		//	cmdArgs >> std::ws; 
		//	if( cmdArgs.good() ) 
		//	{
		//		std::string strarg; cmdArgs >> strarg; 
		//		if( strarg=="fixed=true" ) { fixedTimestep = true; } 
		//		else if( strarg=="fixed=false" ) { fixedTimestep = false; }
		//		else { os << "fixed = " << fixedTimestep << std::endl; }
		//	} 
		//}
		else sotEntity::commandLine(cmdLine,cmdArgs,os);
	}


	void SoTHRP2::play( void ) 
	{ 
		_suspend = false; 
	}

	void SoTHRP2::pause( void ) 
	{
		_suspend = true; 
	}
}
