/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Copyright Projet JRL-Japan, 2007
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
*
* File:      SoTWrapper.cc
* Project:   SOT
* Author:    Nicolas Mansard from Paul Evrard
*
* Version control
* ===============
*
*  $Id$
*
* Description
* ============
*
*
* ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* -------------------------------------------------------------------------- */
/* --- INCLUDE -------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */


/* --- SOT --- */
#include <sot/sotDebug.h>
#include <sot/sotInterpretor.h>


/* --- AFSoT --- */
#include "AFSoT/SoTWrapper.hh"

/* --- AFScene --- */
#include <AFScene/DefaultSceneLoader.hh>
#include <AFScene/Universe.hh>
#include <AFScene/MultiBody.hh>

#ifdef WIN32
#include <Windows.h>
#else 
#include <dlfcn.h> 
#endif

#include <pthread.h>


namespace 
{
	static void* launcherThreadShell( void * argVoid )
	{
		const bool & catchPlease=(const bool&)argVoid;

		do
		{
			try
			{
				std::cout << "In the shell ... " << std::endl;
				sotShell.shell(std::cin,std::cout,std::string(""));
			} 
			catch( const sotExceptionAbstract& exc )
			{
				std::cout << exc << std::endl;
				std::cin.clear();
				if( catchPlease ) 
				{
					std::cout << "in launcherThreadShell: Catch and restart." << std::endl; 
				}
			}
		}  while( catchPlease );
		return NULL;
	}
}

/* -------------------------------------------------------------------------- */
/* --- sotPluginKey --------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
#ifdef WIN32
typedef HMODULE sotPluginKey;
#else
typedef void* sotPluginKey;
#endif

namespace AFSoT
{
	/* -------------------------------------------------------------------------- */
	/* --- CONSTRUCTION --------------------------------------------------------- */
	/* -------------------------------------------------------------------------- */
	//const std::string SoTWrapper::CLASS_NAME = "AMELIF";

	SoTWrapper::SoTWrapper( void )
	  :pluginLoader()
	  ,_iter(0)
	  ,_minTimeStep(0.0001)
# ifdef SOT_CHECK_TIME 
	  ,timeArray()
	  ,stateArray()
	  ,timeIndex()
# endif  

	{
		_multiBodiesSotFull.clear();
		_multiBodiesSot.clear();


		{
#ifndef WIN32
			sotPluginKey dlib = dlopen( "libsot-0.so",RTLD_NOW|RTLD_GLOBAL); 
			if( NULL==dlib ) 
			{
				sotDEBUG(5) << "Failure while loading: " <<dlerror() <<endl;
				SOT_THROW sotExceptionFactory( sotExceptionFactory::DYNAMIC_LOADING,
					"Error while dlopen. ",dlerror() );
			}
#else
			sotPluginKey dlib = LoadLibrary ( "sot-0.dll"); 
			if( NULL==dlib ) 
			{
				// Retrieve the system error message for the last-error code
				LPTSTR pszMessage;
				DWORD dwLastError = GetLastError(); 
				FormatMessage(
					FORMAT_MESSAGE_ALLOCATE_BUFFER | 
					FORMAT_MESSAGE_FROM_SYSTEM |
					FORMAT_MESSAGE_IGNORE_INSERTS,
					NULL,
					dwLastError,
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
					(LPTSTR)&pszMessage,
					0, NULL );


				sotDEBUG(5) << "Failure while loading: " << pszMessage <<endl;


				SOT_THROW sotExceptionFactory( sotExceptionFactory::DYNAMIC_LOADING,
					"Error while dlopen. ", pszMessage );

				LocalFree(pszMessage);
			}
#endif
		}
	}

	SoTWrapper::~SoTWrapper()
	{
		for (unsigned i = 0; i < _multiBodiesSotFull.size(); ++i)
			if(_multiBodiesSotFull[i] != 0x0)
			  delete _multiBodiesSotFull[i];
	}

	void SoTWrapper::initialize( void )
	{
		_iter = 0;

		/* Pointer to the robot. */
		for (unsigned i = 0; i < _multiBodiesSotFull.size(); ++i)
			_multiBodiesSotFull[i]->initialize();
		sotShell.referencePluginLoader(&pluginLoader);

		//
		//runShellPromptThread();
	}

	/* -------------------------------------------------------------------------- */
	/* --- SHELL ---------------------------------------------------------------- */
	/* -------------------------------------------------------------------------- */

	// Script part
	void SoTWrapper::scriptPath( const string &path )
	{
		std::ostringstream oss; oss << "run " << path;
		std::istringstream iss( oss.str() );
		script( iss );
	}
	void SoTWrapper::script( const string &strm )
	{
		std::ostringstream oss; oss << strm;
		std::istringstream iss( oss.str() );
		script( iss );
	}
	void SoTWrapper::script( std::istringstream &strm )
	{
		sotDEBUGIN(5);

		string cmdLine;
		strm >>skipws>> cmdLine ;

		try {
			sotShell.cmd( cmdLine,strm, std::cout );
		} catch( const sotExceptionAbstract& e ) { std::cout << "!! " << e <<endl; }

		sotDEBUGOUT(5);
	}

	void SoTWrapper::runShellPromptThread( const bool & catchPlease )
	{
		pthread_t threadShell;
		pthread_create (&threadShell, NULL, &launcherThreadShell,(void*)true);
	}

	/* -------------------------------------------------------------------------- */
	/* --- LOOP ----------------------------------------------------------------- */
	/* -------------------------------------------------------------------------- */

	void SoTWrapper::newStep( )
	{
		for (unsigned i = 0; i < _multiBodiesSotFull.size(); ++i)
			_multiBodiesSotFull[i]->newStep();
	}



	void SoTWrapper::runStep(double step)
	{
		/* --- RS --- */
		for (unsigned i = 0; i < _multiBodiesSotFull.size(); ++i)
			_multiBodiesSotFull[i]->runStep (_iter);


		double step2 = (step/_minTimeStep);
		double b_sup = ceil(step/_minTimeStep);
		double b_inf = floor(step/_minTimeStep);
		if (b_sup - step2 > step2 -b_inf)
			_iter += b_inf;
		else
			_iter += b_sup;
	}

	/* -------------------------------------------------------------------------- */
	/* --- FILL THE WRAPPER ----------------------------------------------------- */
	/* -------------------------------------------------------------------------- */

 
	AFSoT::SoTMultiBody* SoTWrapper::getMultiBodySot(int id) const 
	{
		if((size_t) id >= _multiBodiesSot.size())
			return 0x0;

		return _multiBodiesSot[id];
	}

	void SoTWrapper::getMultiBodiesSot(std::vector<AFSoT::SoTMultiBody*> & v) const 
	{
		v = _multiBodiesSotFull;
	}

	void SoTWrapper::setMinTimeStep(double step)
	{
		_minTimeStep = step;
	}

	double SoTWrapper::getMinTimeStep() const
	{
		return _minTimeStep;
	}
}
