#include "sad3dof.hh"

#include "AFCollision/PQP/CollisionDetectionPQP.hh"

#include "AFUtils/Configuration.hh"
#include "AFUtils/AFConstants.hh"

#include <AFDynamics/FixedRootMultiBodyDA.hh>
#include "AFSoT/SoTHRP2.hh"

#include "sot/sotSignalTimeDependant.h"

//HRP-2 standing the object
namespace sad3dofNamespace {

	sad3dof::sad3dof()
	:_DefaultTimeStep(0.005)
	,_sceneLoader(NULL)
	,_universe(NULL)
	,_SoT(NULL)
	{
	  //AFUtils::Timer* _timer = new AFUtils::Timer();
	  _moyDet = 0;
	  _moySim = 0;
	  _numStep = 0;
	  _simulationTimeElapsed = 0;
	}

	sad3dof::~sad3dof()
	{
	}

	void sad3dof::initUniverse()
	{
	  /* Construct the scene from XML. */
	  _sceneLoader = new AFScene::DefaultSceneLoader(_appInterface);
	  _universe = _sceneLoader->createUniverse("AFResources/xml/Universe3dof.xml", true);

	  /* Get the reference to the hrp2 multi-body object. */
	  _hrp2 = _universe->getMultiBodyFromUserDefId(0);
	  _hrp2->computeDGM();

	  /* Create the collision detection module. */
	  _collisionDetection = new AFCollision::CollisionDetectionPQP;
	  int groupId = _collisionDetection->generateNewGroupId();
	  _collisionDetection->addPair(_universe->getMultiBodyFromUserDefId(0),
				       _universe->getBodyFromUserDefId(0), groupId );

	  /* Create the dynamic simulation module. */
	  _simLoader 
	    = new AFDynamics::DefaultDynamicsSimulatorLoader
	    <AFDynamics::DefaultFreeBodyDA,AFDynamics::FixedRootMultiBodyDA>
	    (*_universe);
	  _simulator = _simLoader->createDynamicsSimulator();
	  _simulator->setCollisionDetection(_collisionDetection, -1);

	  /* Add the multibody to the dynamic simulator. */
	  _simulator->add(_hrp2);

	  /* Build a controller for the hrp2. */
	  std::string controllerFile, actuatorFile;
	  AFUtils::ConfigNode* config 
	    = _appInterface->getConfiguration()->findNode("/Configuration", "Control");
	  if(config != 0x0)
	    {
	      config->getString(controllerFile, "3dof_PDControllerData");
	      config->getString(actuatorFile, "3dof_ActuatorData");
	      std::cout << "Run with config <" << controllerFile
			<< "> and <" << actuatorFile << ">." << std::endl;
	    } else { std::cout << "No specific config file." << std::endl; }

	  /* Initialisation the robot specific params. */
	  /* -> Init robot configuration. */
	  VNL::Vector<double> q;
	  q.Resize(_hrp2->getNumParameters());
	  q.Fill(0);
	  _hrp2->setStateVector(q);
	  /* -> Init robot inertia. */
	  _simulator->getMultiBodyDA(_hrp2->getId())
	    ->setJointMotorInertia(actuatorFile.c_str());
	  
	  /* Add the listener toward the dynamic simulator. */
	  _interactionDisplayer = new InteractionDisplayer(*_universe);
	  _simulator->getInteractionPublisher()
	    .addInteractionListener(_interactionDisplayer);

	  /* Create and fill the sot. */
	  _SoT = new AFSoT::SoTWrapper ();
	  // Add an SOT entity called "OpenHRP" associated to the robot.
	  _sotHRP2 = _SoT->addMultiBody<AFSoT::SoTHRP2>(_hrp2, "OpenHRP");
	  // Create the controller.
	  _controller = new PDController();
	  _controller->setSize(_hrp2->getNumDoF());
	  _controller->setTimeStep(_DefaultTimeStep);
	  _controller->setMultiBody(_hrp2, controllerFile.c_str());
	  _controller->setSotMultiBody(_sotHRP2);
	}

	//Set the initial position of the hrp2
	void sad3dof::initProgram()
	{
		//initialization of the state of the Hrp2
		std::string fileInitialConfigPath;
		AFUtils::ConfigNode* configInitPos = _appInterface->getConfiguration()->findNode("/Configuration", "InitRobotConfig");
		if(configInitPos != 0x0)
		{
			configInitPos->getString(fileInitialConfigPath, "3dof_initialPosition");
			std::cout << "Inital-pos config path = " << fileInitialConfigPath << std::endl;
		}

		std::ifstream fileInitialConfig (fileInitialConfigPath.c_str());
		double numParam = _hrp2->getNumParameters( );
		VNL::Vector<double> position (numParam);
		position.Fill(0.0);
		if (fileInitialConfig)
		{
			for (unsigned i = 0; i < numParam; ++i)
				fileInitialConfig >> position[i];
			fileInitialConfig.close();
		}
		else
		{
		  // Configuration of the robot in "half-sitting"
		  double POS_INIT [] = { 0.0 , 0.0 , -26.0  , 50  , -24  , 0.0 ,
					 0.0 , 0.0 , -26.0  , 50  , -24  , 0.0 ,
					 0.0 , 0.0 , 0.0 , 0.0 ,
					 15  , -10 , 0.0 , -30 , 0.0 , 0.0  , 0.0  ,
					 15  , 10  , 0.0 , -30 , 0.0 , 0.0  , 0.0 };

		  unsigned int minSize = std::min( sizeof(POS_INIT),position.size() );
		  for( unsigned int i=0;i<minSize;++i )
		    { position[i]=POS_INIT[i]* DEG2RAD; }
		}

		//apply the posture to the robot simulation for initialization
		_hrp2->setStateVector(position);

		// Synchronize the controller with the multibody
		_controller->update();

		//Get data from the file AFConfig.xml --> AMELIF Preferences
		AFUtils::ConfigNode* config = _appInterface->getConfiguration()->findNode("/Configuration", "Simulation");
		if(config != 0x0)
		{
		  double tmp  = _DefaultTimeStep;
		  if(config->getDouble(_DefaultTimeStep, "timeFactor") != 0)
		    _DefaultTimeStep = tmp;
		}
		_controller->setTimeStep(_DefaultTimeStep);
		std::cout << " _DefaultTimeStep " << _DefaultTimeStep << std::endl;

		_interactionDisplayer->newStep();

		_hrp2->computeDGM();
		_collisionDetection->newStep();
		_collisionDetection->detectCollisions(-1, AFCollision::CD_ADDED_PAIRS, true, true);

		//initialize the sot
		_SoT->initialize();

		/* Run initial scripts from AFConfig.xml. */
		std::string path; const std::string labelname="path";
		std::pair<AFUtils::Configuration::tree_iterator,
	                  AFUtils::Configuration::tree_iterator> snodes 
		  = _appInterface->getConfiguration()
		  ->pathNodes("/Configuration/Control/sot/scripts");
		
		for( AFUtils::Configuration::tree_iterator it=snodes.first;
		     it!=snodes.second;++it )
		  {
		    if( 0==it->second->getString(path,labelname) )
		      {
			std::cout << "Run " << path << " ... " << std::flush;
			_SoT->scriptPath(path);
			std::cout << " \t\t\t [OK]" << std::endl;
		      }
		  }

	}

	int sad3dof::run(double step)
	{
	  //step = std::min( _DefaultTimeStep,step );
	  step=_DefaultTimeStep;

		_hrp2->computeDGM();

		// One run execution correspond to one step of the SoT whatever the time dt is
		_SoT->newStep();
		_SoT->setMinTimeStep(_DefaultTimeStep);
		_SoT->runStep(_DefaultTimeStep);

		/* Computation of the control law. */
		VNL::Vector<double> torqueCommand(_controller->computeTorque());
		_hrp2->setJointTorques(torqueCommand);

		/* Re-initialize all. */
		_interactionDisplayer->newStep();
		_collisionDetection->newStep();
		_simulator->newStep();

		/* Detect the collision. */
		_collisionDetection->detectCollisions(-1, AFCollision::CD_ADDED_PAIRS, 
						      true, false);
		
		/* Run the dynamic simulation. */
		_simulator->runStep(step);
		_interactionDisplayer->runDisplay();

		return 0;
	}

	void sad3dof::quitProgram()
	{
		//delete _interactionDisplayer;
		delete _simLoader;
		delete _collisionDetection;
		delete _sceneLoader;
	}
	std::string sad3dof::getName()
	{
		return "sad3dof";
	}
}


