#include "sadMainProgramAbstract.hh"

#include "AFCollision/PQP/CollisionDetectionPQP.hh"

#include "AFUtils/Configuration.hh"
#include "AFUtils/AFConstants.hh"

#include "AFSoT/SoTGenericRobot.hh"

#include "sot/sotSignalTimeDependant.h"

//HRP-2 standing the object
namespace sadMainProgram {

  sadMainProgramAbstract::sadMainProgramAbstract()
    :_verbose(true)
    ,_timeStep(_DEFAULT_TIME_STEP)
    ,_referenceName("")
    ,_sceneLoader(NULL)
    ,_universe(NULL)
    ,_SoT(NULL)
  {
  }
  
  sadMainProgramAbstract::~sadMainProgramAbstract()
  {
  }

  /* --- STATIC INIT AND ACCESS --------------------------------------------- */
  const std::string sadMainProgramAbstract::_SOT_SCRIPT_PATH("/Configuration/Control/sot/scripts");
  const std::string sadMainProgramAbstract::_INITIAL_POSITION_XML_MARKUP("initialPosition");
  const std::string sadMainProgramAbstract::_UNIVERSE_XML_MARKUP("universeFilename");
  const double  sadMainProgramAbstract::_DEFAULT_TIME_STEP = 5e-3;

  const std::string sadMainProgramAbstract::getUniverseFileName( void ) const
  {
    //return sadMainProgramAbstract::_UNIVERSE_XML_FILENAME;  
    std::string res;
    AFUtils::ConfigNode* configInitPos 
      = _appInterface->getConfiguration()->findNode("/Configuration", "Universe");
    if(configInitPos != 0x0)
      {
	configInitPos->getString(res,getUniverseXmlMarkup());
      } else { res = ""; }
    return res;
  }
  const std::string sadMainProgramAbstract::getSotScriptPath( void ) const
  { return sadMainProgramAbstract::_SOT_SCRIPT_PATH;  }
  const std::string sadMainProgramAbstract::getInitialPositionXmlMarkup( void ) const
  { 
    std::string res;
    if( 0==_referenceName.length() ) res = "";
    else { res = _referenceName + "_"; }
    res += sadMainProgramAbstract::_INITIAL_POSITION_XML_MARKUP; 
    return res;
  }
  const std::string sadMainProgramAbstract::getUniverseXmlMarkup( void ) const
  { 
    std::string res;
    if( 0==_referenceName.length() ) res = "";
    else { res = _referenceName + "_"; }
    res += sadMainProgramAbstract::_UNIVERSE_XML_MARKUP;
    return res;
  }
  const double sadMainProgramAbstract::getDefaultTimeStep( void ) const
  { return sadMainProgramAbstract::_DEFAULT_TIME_STEP; }

  const int sadMainProgramAbstract::getMultiBodyId( void ) { return 0; }

  /* --- CONSTRUCTION ------------------------------------------------------- */
  void sadMainProgramAbstract::constructSadUniverse(void)
  {
    setReferenceName();
    /* Construct the scene from XML. */
    _sceneLoader = new AFScene::DefaultSceneLoader(_appInterface);
    const std::string& universeFileName = getUniverseFileName();
    _universe = _sceneLoader->createUniverse(universeFileName, true);
    if( _verbose ) 
      { std::cout << "Loaded Universe " << universeFileName << std::endl; }
  }

  void sadMainProgramAbstract::selecSimulatedMultiBodies( void )
  {
    /* Get the reference to the hrp2 multi-body object. */
    _mainMultiBody = _universe->getMultiBodyFromUserDefId(getMultiBodyId());
    if( _verbose )
      { std::cout << "Main Multibody is number " << getMultiBodyId() << std::endl; }
    _mainMultiBody->computeDGM();
  }

  void sadMainProgramAbstract::selecSotMultiBodies( void )
  {
    // Add an SOT entity called "OpenHRP" associated to the robot.
    _sotMainMultiBody = _SoT->addMultiBody<AFSoT::SoTGenericRobot>(_mainMultiBody, "OpenHRP");
  }

  void sadMainProgramAbstract::constructSot( void )
  {
    _SoT = new AFSoT::SoTWrapper ();
    selecSotMultiBodies();
  }

  void sadMainProgramAbstract::constructSimulation( void )
  {
    selecSimulatedMultiBodies();
  }

  void sadMainProgramAbstract::setReferenceName( void )
  {
    AFUtils::ConfigNode* configInitPos 
      = _appInterface->getConfiguration()->findNode("/Configuration", "Control");
    if(configInitPos != 0x0)
      {
	configInitPos->getString(_referenceName,"robotName");
	if( _verbose ) 
	  { std::cout << "Robot name is '" << _referenceName<<"'" << std::endl;}
      } else { _referenceName = ""; }
  }
  /* --- INITIALIZATION ----------------------------------------------------- */
  void sadMainProgramAbstract::initMultiBodiesSpecificParams( void ) {}
  
  void sadMainProgramAbstract::initMultiBodiesConfiguration( void )
  { 
    /* Initialization of the state of the MainMultiBody. */
    std::string fileInitialConfigPath;
    AFUtils::ConfigNode* configInitPos 
      = _appInterface->getConfiguration()->findNode("/Configuration", "InitRobotConfig");
    if(configInitPos != 0x0)
      {
	configInitPos->getString(fileInitialConfigPath, getInitialPositionXmlMarkup());
	if( _verbose ) 
	  { std::cout << "Initial-pos config path " << "'" << fileInitialConfigPath
		      << "'" << std::endl;}
      }

    std::ifstream fileInitialConfig (fileInitialConfigPath.c_str());
    double numParam = _mainMultiBody->getNumParameters( );
    VNL::Vector<double> position (numParam);
    if (fileInitialConfig)
      {
	for (unsigned i = 0; i < numParam; ++i)
	  fileInitialConfig >> position[i];
	fileInitialConfig.close();
      }
    else {  position.Fill(0.0);   }

    //apply the posture to the robot simulation for initialization
    _mainMultiBody->setStateVector(position);
  }

  void  sadMainProgramAbstract::initTimeFactor( void )
  {
    AFUtils::ConfigNode* config 
      = _appInterface->getConfiguration()->findNode("/Configuration", "Simulation");
    if(config != 0x0)
      {
	if(config->getDouble(_timeStep, "timeFactor") != 0)
	  { 
	    _timeStep = _DEFAULT_TIME_STEP; 
	  }
      }
    if( _verbose ) 
      {std::cout << "Applied time step = " << _timeStep << std::endl;}
	
  }

  /* Run initial scripts from AFConfig.xml. */
  void  sadMainProgramAbstract::runInitialSotScripts( void )
  {
    std::string path; const std::string labelname="path";
    std::pair<AFUtils::Configuration::tree_iterator,
      AFUtils::Configuration::tree_iterator> snodes 
      = _appInterface->getConfiguration()->pathNodes( getSotScriptPath() );
		
    for( AFUtils::Configuration::tree_iterator it=snodes.first;it!=snodes.second;++it )
      {
	if( 0==it->second->getString(path,labelname) )
	  {
	    if( _verbose )
	      { std::cout << "Run " << path;  
		/* Flush after path, otherwise not displayed. */
		std::cout <<std::flush << " ... \t" << std::endl; }
	    _SoT->scriptPath(path);
	    if( _verbose ) { std::cout << " \t\t\t [OK]" << std::endl; }
	  }
      }
  }
  void sadMainProgramAbstract::initSadUniverse( void )
  {
    initMultiBodiesSpecificParams();
    initMultiBodiesConfiguration();
    _mainMultiBody->computeDGM();
    initTimeFactor();
  }
  void sadMainProgramAbstract::initSot( void )
  {
    _SoT->initialize();
    /* Run SOT Init. */
    runInitialSotScripts();
    _SoT->runShellPromptThread();
  }

  /* --- MAIN LOOP ---------------------------------------------------------- */

  const double  sadMainProgramAbstract::
  computeTimeStep( const double& AFTimeStep ) const
  {
    return _timeStep;
  }
  void sadMainProgramAbstract::
  runSadUniverse( const double& timeStep )
  {
    _mainMultiBody->computeDGM();
  }
  void sadMainProgramAbstract::
  runSot( const double& timeStep )
  {
    /* The execution corresponds to one step of the SoT whatever the time dt is. */
    _SoT->newStep();
    _SoT->setMinTimeStep(timeStep);
    _SoT->runStep(timeStep);
  }

  /* --- DESTRUCTION -------------------------------------------------------- */
  void sadMainProgramAbstract::
  destructSot( void )
  {}
  void sadMainProgramAbstract::
  destructSadUniverse( void )
  {
  }
  /* --- AF INHERITANCE ----------------------------------------------------- */
  /* AF main program initialization */
  void sadMainProgramAbstract::initUniverse()
  {
    constructSadUniverse(); 
    constructSimulation();
    constructSot();
    constructController();
  }

  void sadMainProgramAbstract::initProgram()
  {
    initSadUniverse();
    initController();
    initSimulation();
    initSot();
  }


  int sadMainProgramAbstract::run(double AFstep)
  {
    
    const double step = computeTimeStep( AFstep );
    runSadUniverse(step);
    runSot(step);
    runController(step);
    runSimulation(step);

    return 0;
  }

  void sadMainProgramAbstract::quitProgram()
  {
    destructSot();
    destructSimulation();
    destructController();
    destructSadUniverse();
  }
  std::string sadMainProgramAbstract::getName()
  {
    return "sadMainProgramAbstract";
  }
}


