#include "sadMainProgramDynamicAbstract.hh"
#include "AFCollision/PQP/CollisionDetectionPQP.hh"
#include "AFUtils/Configuration.hh"
#include <AFDynamics/FixedRootMultiBodyDA.hh>

namespace sadMainProgram {

  sadMainProgramDynamicAbstract::sadMainProgramDynamicAbstract()
  {
  }
  
  sadMainProgramDynamicAbstract::~sadMainProgramDynamicAbstract()
  {
  }

  const std::string sadMainProgramDynamicAbstract::_ACTUATOR_CONFIG_XML_MARKUP
  = "ActuatorData";
  const std::string sadMainProgramDynamicAbstract::
  getActuatorConfigXmlMarkup( void ) const
  { 
    std::string res;
    if( 0==_referenceName.length() ) res = "";
    else { res = _referenceName + "_"; }
    res += sadMainProgramDynamicAbstract::_ACTUATOR_CONFIG_XML_MARKUP; 
    return res;
  }

  const std::string sadMainProgramDynamicAbstract::_FIXED_ROOT_XML_MARKUP
  = "fixedRoot";
  const std::string sadMainProgramDynamicAbstract::
  getFixedRootXmlMarkup( void ) const
  { 
    std::string res;
    if( 0==_referenceName.length() ) res = "";
    else { res = _referenceName + "_"; }
    res += sadMainProgramDynamicAbstract::_FIXED_ROOT_XML_MARKUP; 
    return res;
  }

  const bool sadMainProgramDynamicAbstract::
  getFixedRootOption( void ) const
  {   
    std::string opts;
    AFUtils::ConfigNode* config 
      = _appInterface->getConfiguration()->findNode("/Configuration", "Universe");
    if(config != 0x0)
      {
	config->getString(opts,getFixedRootXmlMarkup()); 
	return ( "yes"==opts );
      } else { return false; }

  }

  /* --- CONSTRUCTION ------------------------------------------------------- */
  void sadMainProgramDynamicAbstract::constructSimulation(void)
  {
    /* Create the collision detection module. */
    _collisionDetection = new AFCollision::CollisionDetectionPQP;
    
    /* Create the dynamic simulation module. */
    if( getFixedRootOption() )
      {
	_simLoader 
	  = new AFDynamics::DefaultDynamicsSimulatorLoader
	  <AFDynamics::DefaultFreeBodyDA,AFDynamics::FixedRootMultiBodyDA>
	  (*_universe);
      }
    else  
      {
	_simLoader 
	  = new AFDynamics::DefaultDynamicsSimulatorLoader
	  <AFDynamics::DefaultFreeBodyDA,AFDynamics::DefaultMultiBodyDA>
	  (*_universe);
      }
    _simulator = _simLoader->createDynamicsSimulator();
    _simulator->setCollisionDetection(_collisionDetection, -1);

    /* Add the listener toward the dynamic simulator. */
    _interactionDisplayer = new InteractionDisplayer(*_universe);
    _simulator->getInteractionPublisher()
      .addInteractionListener(_interactionDisplayer);

    sadMainProgramAbstract::constructSimulation();
  }

  void sadMainProgramDynamicAbstract::selecSimulatedMultiBodies( void )
  {
    sadMainProgramAbstract::selecSimulatedMultiBodies();

    int groupId = _collisionDetection->generateNewGroupId();
    _collisionDetection->addPair(_mainMultiBody,
				 _universe->getBodyFromUserDefId(0),groupId );

    /* Add the multibody to the dynamic simulator. */
    _simulator->add(_mainMultiBody);
  }
 
  /* --- INITIALIZATION ----------------------------------------------------- */
  void sadMainProgramDynamicAbstract::initMultiBodiesSpecificParams( void )
  {
    sadMainProgramAbstract::initMultiBodiesSpecificParams();

    /* -> Init robot inertia. */
    std::string actuatorFile;
    AFUtils::ConfigNode* config 
      = _appInterface->getConfiguration()->findNode("/Configuration", "Control");
    if(config != 0x0)
      {
	config->getString(actuatorFile,getActuatorConfigXmlMarkup()); 
      }
    _simulator->getMultiBodyDA(_mainMultiBody->getId())
      ->setJointMotorInertia(actuatorFile.c_str());

    if( _verbose )
      { std::cout <<"Parsed actuator-param file '" << actuatorFile
		  << "'"<<std::endl; }
  }

  void sadMainProgramDynamicAbstract::initSimulation( void )
  {
    /* Initialize the dyn/collision modules. */
    _interactionDisplayer->newStep();
    _mainMultiBody->computeDGM();
    _collisionDetection->newStep();
    _collisionDetection->detectCollisions(-1, AFCollision::CD_ADDED_PAIRS,
					  true, true);
  }

  /* --- MAIN LOOP ---------------------------------------------------------- */
  
  void sadMainProgramDynamicAbstract::
  runSimulation( const double& timeStep )
  {
    /* Step-initialize all the modules. */
    _interactionDisplayer->newStep();
    _collisionDetection->newStep();
    _simulator->newStep();
    
    /* Detect the collision. */
    _collisionDetection->detectCollisions(-1, AFCollision::CD_ADDED_PAIRS, 
					  true, false);
    
    /* Run the dynamic simulation. */
    _simulator->runStep(timeStep);
    _interactionDisplayer->runDisplay();
  }

  void sadMainProgramDynamicAbstract::
  runController( const double& timeStep )
  {
    /* Computation of the torque-based control law. */
    VNL::Vector<double> torqueCommand(_controller->computeTorque());
    _mainMultiBody->setJointTorques(torqueCommand);
  }

  /* --- DESTRUCTION -------------------------------------------------------- */
  void sadMainProgramDynamicAbstract::
  destructSimulation( void )
  {
    delete _interactionDisplayer;
    delete _simLoader;
    delete _collisionDetection;
  }
  void sadMainProgramDynamicAbstract::
  destructController( void )
  {
    delete _controller;
  }
  std::string sadMainProgramDynamicAbstract::getName()
  {
    return "AfSotAbstractDyn";
  }
}


