/*
 * =====================================================================================
 *
 *       Filename:  sadPDController.hh
 *
 *    Description:  A sadPD Controller specialized in AMELIF and HRP2
 *
 *        Created:  03/31/2009 10:03:53 PM
 *
 *         Author:  Hak Sovannara (shak@laas.fr), 
 *
 * =====================================================================================
 */

#ifndef sadPD_CONTROLLER_HH
#define sadPD_CONTROLLER_HH

#include "sadAbstractController.hh"

#include "AFSoT/SoTMultiBody.hh" 

namespace AFScene
{
  class MultiBody;
}

/*! /brief basic PD controller
 * Kp and Kd parameters of the controller
 * dt the time-step for derivation
 *
 * We assume all joints have 1 DOF so size(Kp) = size(Kd) = size(Tau)...
 * Initialization:
 *      setSize(SIZE);
 *      setTimeStep(DT);
 *      
 *      setMultiBody(MB, CFILE);
 *      setSoTHRP2(SOTHRP2);
 *
 * input: a desired configuration/velocity computed by the Stack of Tasks
 *        an actual configuration/velocity given by the AMELIF multibody
 *
 * output: a torque command for the robot
 */
class sadPDController : public sadAbstractController
{
public:

  sadPDController();
  sadPDController(int dim);

  ~sadPDController();

public:
  /* 
   * Set the size of every vector (_q, _qdes, _Kp...)
   * it is the number of degree of freedom of the multibody
   * (initialize also everything to 0)
   * Must be called first;
   * */
  void setSize(int dimension);

  /*
   * Set the time step
   */
  void setTimeStep(double dt);

  /*
   * Compute the torque command to be applied
   * given the actual configuration and desired configuration,
   * the time-step and the constants Kp Kd.
   *
   * return  pointer to the torque
   */
  VNL::Vector<double> control();

  //----------AMELIF SPECIFIC PART---------

  /*
   * Set the position to be achieved via servoing
   */
  void setDesiredPosition(const VNL::Vector<double>& qdes);

  /*
   * Set the speed to be achieved via servoing
   */
  void setDesiredSpeed(const VNL::Vector<double>& dqdes);

  /*
   * Constant parameters for the Proportional part
   * Size of the vector correspond to the number of degree of freedom
   * of the multibody
   */
  void setKp(const VNL::Vector<double>& Kp);

  /*
   * Constant parameters for the Derivation part
   * Size of the vector correspond to the number of degree of freedom
   * of the multibody
   */
  void setKd(const VNL::Vector<double>& Kd);

  /*
   * Get the position to be achieved
   */
  void getDesiredPosition(VNL::Vector<double>& qdes) const;

  /*
   * Get the speed to be achieved
   */
  void getDesiredSpeed(VNL::Vector<double>& dqdes) const;

  /*
   * Get the Kp constants
   */
  void getKp(VNL::Vector<double>& Kp) const;

  /*
   * Get the Kd constants
   */
  void getKd(VNL::Vector<double>& Kd) const;


  /*
   * Update the desired position to be equal to the actual position
   * ie for initialization purpose
   */
  void update();

  /* 
   * Set the multibody which will be used, this multibody belongs to the
   * scene we want to simulate. It contains the actual position/speed
   * of the robot
   *
   * const char* params is a string representing a file 
   *                    which contains Kp and Kd constant
   * */
  void setMultiBody(AFScene::MultiBody* mb, const char* params);

  /*
   * Set the sotMultibody,
   * it contains the desired position/speed (computed in sot)
   */
  void setSotMultiBody(AFSoT::SoTMultiBody* sotMB);
  
  /* 
   * Read the config file, and file in the Kp/Kd. 
   * Return false in case of error.
   */
  bool parseConfigFile( const std::string& paramsFilename );


  /*
   * Set the controller inputs from multibody and sotMultibody
   * and perform the torque command computation:
   * return pointer to the torque
   */
  VNL::Vector<double> computeTorque();

  //--------------------------------

private:

  /*
   * Dimension of all vectors
   */
  int _dimension;

  /*
   * Actual/Mesured configuration/speed
   */
  VNL::Vector<double> _q;
  VNL::Vector<double> _dq;

  /*
   * Desired configuration/speed
   */
  VNL::Vector<double> _qdes;
  VNL::Vector<double> _dqdes;

  /*
   * Controller parameters
   */
  VNL::Vector<double> _Kp;
  VNL::Vector<double> _Kd;

  /*
   * Torque command
   */
  VNL::Vector<double> _tau;

  /*
   * Timestep
   */
  double _dt;

  //----------AMELIF SPECIFIC PART---------
		
  /*
   * Multibody that is simulated
   */
  AFScene::MultiBody* _multiBody;

  /*
   * Multibody state given by SoT
   */
  AFSoT::SoTMultiBody* _sotMultiBody; 

  //--------------------------------
};

#endif
