/*
 * =====================================================================================
 *
 *       Filename:  PdController.hh
 *
 *    Description:  A sadTorque Controller specialized in AMELIF and HRP2
 *
 *        Created:  03/31/2009 10:03:53 PM
 *
 *         Author:  Hak Sovannara (shak@laas.fr), 
 *
 * =====================================================================================
 */

#ifndef sadTorque_CONTROLLER_HH
#define sadTorque_CONTROLLER_HH

#include "sadAbstractController.hh"

#include "AFSoT/SoTMultiBody.hh" 

namespace AFScene
{
  class MultiBody;
}

/*! /brief basic sadTorque controller
 * Kp and Kd parameters of the controller
 * dt the time-step for derivation
 *
 * We assume all joints have 1 DOF so size(Kp) = size(Kd) = size(Tau)...
 * Initialization:
 *      setSize(SIZE);
 *      setTimeStep(DT);
 *      
 *      setMultiBody(MB, CFILE);
 *      setSoTHRP2(SOTHRP2);
 *
 * input: a desired configuration/velocity computed by the Stack of Tasks
 *        an actual configuration/velocity given by the AMELIF multibody
 *
 * output: a torque command for the robot
 */
class sadTorqueController : public sadAbstractController
{
public:

  sadTorqueController();
  sadTorqueController(int dim);

  ~sadTorqueController();

public: 
  static const double _EPSILON_STATIC;

public:
  /* 
   * Set the size of every vector (_q, _qdes, _Kp...)
   * it is the number of degree of freedom of the multibody
   * (initialize also everything to 0)
   * Must be called first;
   * */
  void setSize(int dimension);

  /*
   * Set the time step
   */
  void setTimeStep(double dt);

  /*
   * Compute the torque command to be applied
   * given the actual configuration, the reference control,
   * the time-step and the constants of the controller.
   *
   * Return  the torque control to be applied on the motors.
   */
  VNL::Vector<double> computeTorque();

  //----------AMELIF SPECIFIC PART---------

  void setFrictionStatic(const VNL::Vector<double>& );
  void setFrictionFluid(const VNL::Vector<double>& );

  void getFrictionStatic(VNL::Vector<double>& );
  void getFrictionFluid(VNL::Vector<double>& );

  /*
   * Update the desired position to be equal to the actual position
   * ie for initialization purpose
   */
  void update();

  /* 
   * Set the multibody which will be used, this multibody belongs to the
   * scene we want to simulate. It contains the actual position/speed
   * of the robot
   *
   * const char* params is a string representing a file 
   *                    which contains Kp and Kd constant
   * */
  bool parseConfigFile( const std::string& paramsFilename );
  void setMultiBody(AFScene::MultiBody* mb,const std::string& params);

  /*
   * Set the sotMultibody,
   * it contains the desired position/speed (computed in sot)
   */
  void setSotMultiBody(AFSoT::SoTMultiBody* sotMB);


  //--------------------------------

private:

  /*
   * Dimension of all vectors
   */
  int _dimension;

  /*
   * Actual/Mesured configuration/speed
   */
  VNL::Vector<double> _dq;

  /*
   * Controller parameters
   */
  VNL::Vector<double> _fstatic;
  VNL::Vector<double> _ffluid;

  /*
   * sadTorque command
   */
  VNL::Vector<double> _tau;

  /*
   * Timestep
   */
  double _dt;

  //----------AMELIF SPECIFIC PART---------
		
  /*
   * Multibody that is simulated
   */
  AFScene::MultiBody* _multiBody;

  /*
   * Multibody state given by SoT
   */
  AFSoT::SoTMultiBody* _sotMultiBody; 

  //--------------------------------
};

#endif
