// STL
#include <iostream>

// LIBC
#include <cstdlib>
#include <cassert>

// Headers
#include "CorbaConnection.h"

// Namespaces
using namespace std;


/* ------------------------------------------------------- */
/* --- ORB Singleton ------------------------------------- */
/* ------------------------------------------------------- */

CORBA::ORB_var TheORB::orb = 0x0;
string TheORB::ip = "127.0.0.1";
string TheORB::port = "2809";
string TheORB::ns = "NameService";

void TheORB::set_ns_ip(const char* ip)
{
  TheORB::ip = ip;
}

void TheORB::set_ns_port(const char* port)
{
  TheORB::port = port;
}

void TheORB::set_ns_name(const char* ns)
{
  TheORB::ns = ns;
}

CORBA::ORB_ptr TheORB::get()
{
  static TheORB instance;
  return instance.orb;
}

TheORB::TheORB()
{
  cerr << "Initialize ORB" << endl;

  int argc = 3;
  string arg2 = "NameService=corbaloc:iiop:"+ip+":"+port+"/"+ns;
  char* argv2 = strdup(arg2.c_str());
  char *argv[] = {"client","-ORBInitRef",argv2};

  try {
    orb = CORBA::ORB_init(argc, argv);
  }
  catch(...) {
    cerr << "Exception caught at ORB initialization!" << endl;
    free(argv2);
    exit(-1);
  }

  free(argv2);
}


/* ------------------------------------------------------- */
/* --- CORBA client -------------------------------------- */
/* ------------------------------------------------------- */

const int CorbaConnection::trials_before_connection = 5;

CorbaConnection::CorbaConnection():
  server(CORBA::Object::_nil()),
  server_id("DefaultServer"),
  trials(0),
  connection_ok(false)
{
  // first call to get_server() must attempt to connect
  trials = trials_before_connection;
}

void CorbaConnection::set_server_name(const char* name)
{
  server_id = name;
}

void CorbaConnection::set_server_kind(const char* kind)
{
  server_kind = kind;
}

CORBA::Object_ptr CorbaConnection::get_server()
{
  connection_ok = !CORBA::is_nil(server);

  if(!connection_ok)
  {
    if(trials >= trials_before_connection)
    {
      connection_ok = (try_corba_connection() == 0);
      trials = 0;
    }
    else { ++trials; }
  }

  return server;
}

CORBA::Object_ptr
CorbaConnection::get_object_reference(const char* serverId,
				      const char* serverKind)
{
  try {
    //-------------------------------------------------------------
    // Bind ORB object to name service object.
    //-------------------------------------------------------------
    cerr << "Resolve initial references..." << endl;

    CORBA::Object_var obj = TheORB::get()->resolve_initial_references("NameService");
    assert(!CORBA::is_nil(obj));

    //-------------------------------------------------------------
    // Narrow this to the naming context
    //-------------------------------------------------------------
    cerr << "Narrow to naming context..." << endl;

    CosNaming::NamingContext_var nc =
      CosNaming::NamingContext::_narrow(obj.in());
    assert(!CORBA::is_nil(nc.in()));

    //-------------------------------------------------------------
    // The "name text" put forth by CORBA server in name service.
    //-------------------------------------------------------------
    cerr << "Resolve server name: create CosName..." << endl;

    CosNaming::Name _corbaCosName;
    _corbaCosName.length(1);
    _corbaCosName[0].id = server_id.c_str();
    _corbaCosName[0].kind = server_kind.c_str();

    //-------------------------------------------------------------
    // Resolve "name text" identifier to an object reference.
    //-------------------------------------------------------------
    cerr << "Resolve naming context: return reference..." << endl;

    return nc->resolve(_corbaCosName);
  }
  catch(CORBA::COMM_FAILURE& ex) {
    cerr << "Caught system exception COMM_FAILURE -- unable to contact the "
	      << "object." << endl;
    throw DS_ServerConnectionException();
    return CORBA::Object::_nil();
  }
  catch(CORBA::SystemException& ) {
    cerr << "Caught a CORBA::SystemException." << endl;
    throw DS_SystemException();
    return CORBA::Object::_nil();
  }
  catch(CORBA::Exception& ) {
    cerr << "Caught CORBA::Exception." << endl;
    throw DS_Exception();
    return CORBA::Object::_nil();
  }
  catch(omniORB::fatalException& fe) {
    cerr << "Caught omniORB::fatalException:" << endl;
    cerr << "  file: " << fe.file() << endl;
    cerr << "  line: " << fe.line() << endl;
    cerr << "  mesg: " << fe.errmsg() << endl;
    throw DS_FatalException();
    return CORBA::Object::_nil();
  }
  catch(...) {
    cerr << "Caught unknown exception." << endl;
    throw DS_Exception();
    return CORBA::Object::_nil();
  }

  return CORBA::Object::_nil();
}

int CorbaConnection::try_corba_connection()
{
  server = get_object_reference(server_id.c_str(), server_kind.c_str());

  if(CORBA::is_nil(server))
    return 1;

  return 0;
}
