/* Libc */
#include <cassert>

/* STL */
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
using namespace std;

/* Matrix */
#include <MatrixAbstractLayer/boost.h>
namespace ml = maal::boost;

/* SoT */
#include "sot/sotDebug.h"

/* SotTeleopClient */
#include "CorbaConnection.h"

/* Server command C++ bindings. */
#include <ServerCommand.h>


/* ---------------------------------------------------------------- */
/* --- CONFIGURATION ---------------------------------------------- */
/* ---------------------------------------------------------------- */

char NS_IP[256] = "127.0.0.1";
char NS_PORT[256] = "2809";
char SOT_NAME[256] = "coshell";

void READ_CONFIG()
{
  ifstream in("config.txt");
  if(!in.is_open()) {
    return;
  }

  char dummy[256];
  in >> dummy >> NS_IP
     >> dummy >> NS_PORT
     >> dummy >> SOT_NAME;
}


/* ---------------------------------------------------------------- */
/* --- CLIENT ----------------------------------------------------- */
/* ---------------------------------------------------------------- */

class sotTeleopClient
{
public:

  sotTeleopClient();

  void setForceInput(const ml::Vector& f);
  void setHeadInput(const ml::Vector& yp);
  void setConnect(const ml::Vector& connect);
  void getPositionOutput(ml::Vector& p);

  void commandLine(string& cmd, istream& args);

private:

  CorbaConnection connection_to_sot;
  SOT_Server_Command_ptr server_sot;

  long forceRSOUT;
  long forceLSOUT;
  long headSOUT;
  long gripRSOUT;
  long connectSOUT;
  long posRSIN;
  long posLSIN;
};


/* ---------------------------------------------------------------- */
/* --- HELPERS ---------------------------------------------------- */
/* ---------------------------------------------------------------- */

namespace
{
  SOT_Server_Command::CharSeq&
  setSigName( const string& buffer,
	      SOT_Server_Command::CharSeq& signame,
	      const string& IO )
  {
    const string sigref = "Teleop-Distant::"+IO+"put(vector)::";
    string sigfullname = sigref + buffer;
    const unsigned nbCar = sigfullname.length();
    signame.length(nbCar);
    for( int i=0;i<nbCar;++i ){ signame[i]=sigfullname[i]; }
    return signame;
  }

  SOT_Server_Command::DoubleSeq&
  setInitVector( SOT_Server_Command::DoubleSeq &force,
		 const unsigned int & dim )
  {
    force.length(dim);
    for( unsigned i=0;i<dim;++i ) force[i]=0;
    return force;
  }

  string sendSotCommand( SOT_Server_Command_ptr serverPtr,
			 const std::string& cmd )
  {
    SOT_Server_Command::CharSeq signame; 
    signame.length( cmd.size() ); 
    for( unsigned int i=0;i<cmd.size();++i ) signame[i]=cmd[i];
    SOT_Server_Command::StringStreamer_var outputShell;
    serverPtr->runAndRead( signame,outputShell );

    std::ostringstream ans; 
    for( unsigned int i=0;i<outputShell->length();i++ )   
    { ans << outputShell[i]; }
    usleep( 1000*20 );
    return ans.str();
  }
}

#define INIT_SOUT(name,dim) \
  name \
     = server_sot->createOutputVectorSignal( setSigName( #name,signame,"out" ) ); \
  server_sot->writeOutputVectorSignal( name,setInitVector(initVector,dim) );

#define INIT_SIN(name) \
  name \
     = server_sot->createInputVectorSignal( setSigName( #name,signame,"in" ) ); 

#define ML_TO_CORBA(name, signame, dim)			      \
  SOT_Server_Command::DoubleSeq c##name; c##name.length(dim); \
  for( int i=0; i<dim; ++i ){ c##name[i] = name(i); }	      \
  server_sot->writeOutputVectorSignal(signame, c##name);

#define CORBA_TO_ML(name, signame, dim)			\
  SOT_Server_Command::DoubleSeq_var c##name;		\
  server_sot->readInputVectorSignal(signame, c##name);	\
  if(c##name->length() != dim) {			\
    sotERROR << "Wrong signal dimension: "		\
	     << c##name->length() << "." << endl;	\
  }							\
  for( int i=0; i<dim; ++i ){ name(i) = c##name[i];  }


/* ---------------------------------------------------------------- */
/* --- CLIENT IMPLEMENTATION -------------------------------------- */
/* ---------------------------------------------------------------- */

sotTeleopClient::sotTeleopClient()
{
  TheORB::set_ns_ip(NS_IP);
  TheORB::set_ns_port(NS_PORT);

  initServer<SOT_Server_Command_ptr, SOT_Server_Command>(connection_to_sot, server_sot, SOT_NAME);

  SOT_Server_Command::CharSeq signame;
  SOT_Server_Command::DoubleSeq initVector;

  try {
    INIT_SOUT(forceRSOUT,6);
    INIT_SOUT(forceLSOUT,6);
    INIT_SOUT(headSOUT,2);
    INIT_SOUT(gripRSOUT,1);
    INIT_SOUT(connectSOUT,1);
      
    INIT_SIN(posRSIN);
    INIT_SIN(posLSIN);
  }
  catch(...) {
    sotERROR <<"Signals already exist." <<std::endl;
  }

  sendSotCommand(server_sot, "run ${CMAKE_INSTALL_PREFIX}/script/teleoperation/initConnec");
}

void sotTeleopClient::setForceInput(const ml::Vector& f)
{
  ML_TO_CORBA(f, forceRSOUT, 6);
}

void sotTeleopClient::setHeadInput(const ml::Vector& yp)
{
  ML_TO_CORBA(yp, headSOUT, 2);
}

void sotTeleopClient::setConnect(const ml::Vector& connect)
{
  ML_TO_CORBA(connect, connectSOUT, 1);
}

void sotTeleopClient::getPositionOutput(ml::Vector& p)
{
  CORBA_TO_ML(p, posRSIN, 7);
}

void sotTeleopClient::commandLine(string& cmd, istream& args)
{
  if(cmd == "set_force") {
    ml::Vector f(6);
    if( !(args>>f(0)) || !(args>>f(1)) || !(args>>f(2)) || !(args>>f(3)) || !(args>>f(4)) || !(args>>f(5)) ) {
      cerr<<cmd<<" usage: "<<cmd<<" fx fy fz tx ty tz"<<endl;
      cerr<<"command ignored"<<endl;
      return;
    }
    setForceInput(f);
  }
  else if(cmd == "set_head_yp") {
    ml::Vector yp(2);
    if( !(args>>yp(0)) || !(args>>yp(1)) ) {
      cerr<<cmd<<" usage: "<<cmd<<" yaw pitch"<<endl;
      cerr<<"command ignored"<<endl;
      return;
    }
    setHeadInput(yp);
  }
  else if(cmd == "connect") {
    ml::Vector connect(1);
    if( !(args>>connect(0)) ) {
      cerr<<cmd<<" usage: "<<cmd<<" {1|0}"<<endl;
      cerr<<"command ignored"<<endl;
      return;
    }
    setConnect(connect);
  }
  else {
    cerr<<"unknown command... ignored"<<endl;
    return;
  }
}


/* ---------------------------------------------------------------- */
/* --- MAIN ------------------------------------------------------- */
/* ---------------------------------------------------------------- */

int main()
{
  READ_CONFIG();

  sotTeleopClient theClient;

  string line, cmd;
  stringstream ss;

  while(1) {
    cout << endl << "> ";

    getline(cin, line);
    if(line.empty()){ continue; }
    if(line == "exit"){ break; }

    size_t start = line.find_first_not_of(" ");
    if(start != string::npos){ line.erase(0, start); }

    ss.clear(); ss.str(line);
    cmd.clear();  ss >> cmd;

    theClient.commandLine(cmd, ss);
  }

  return 0;
}
