/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotNextStep.h
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <time.h>
#ifndef WIN32
#include <sys/time.h>
#else
#include <MatrixAbstractLayer/boost.h>
#include <sot/sotUtilsWindows.h>
#include <Winsock2.h>
#endif /*WIN32*/

#include <sot/sotNextStep.h>
#include <sot/sotDebug.h>
#include <sot/sotFactory.h>
#include <sot/sotMacrosSignal.h>
#include <sot/sotExceptionPatternGenerator.h>


SOT_FACTORY_ENTITY_PLUGIN(sotNextStep,"NextStep");
 


/* --- CONSTRUCT ------------------------------------------------------------- */
/* --- CONSTRUCT ------------------------------------------------------------- */
/* --- CONSTRUCT ------------------------------------------------------------- */

const ml::Vector sotNextStep::VECTOR_NULL_ARG;
const unsigned int sotNextStep::PERIOD_DEFAULT = 160; // 160iter=800ms
const double sotNextStep::ZERO_STEP_POSITION_DEFAULT = .19;

sotNextStep::
sotNextStep( const std::string & name ) 
  :sotEntity(name)
  
  ,footPrintList()
  ,period(PERIOD_DEFAULT)
  ,timeLastIntroduction( 0 )

  ,state( STATE_STOPED )

  ,zeroStepPosition( ZERO_STEP_POSITION_DEFAULT )

  ,ref0Mrf()
  ,ref0Mlf()
  ,twoHandObserver( name )

  ,referencePositionLeftSIN( NULL,"sotNextStep("+name+")::input(vector)::posrefleft" )
  ,referencePositionRightSIN( NULL,"sotNextStep("+name+")::input(vector)::posrefright" )

  ,contactFootSIN( NULL,"sotNextStep("+name+")::input(uint)::contactfoot" )
  ,triggerSOUT( "sotNextStep("+name+")::input(dummy)::trigger" )
{
  sotDEBUGIN(5);
  
  triggerSOUT.setFunction( boost::bind(&sotNextStep::triggerCall,this,_1,_2) );

  signalRegistration( referencePositionLeftSIN<<referencePositionRightSIN
		      <<contactFootSIN<<triggerSOUT );
  signalRegistration( twoHandObserver.getSignals() );

  referencePositionLeftSIN.plug( &twoHandObserver.referencePositionLeftSOUT );
  referencePositionRightSIN.plug( &twoHandObserver.referencePositionRightSOUT );

  sotDEBUGOUT(5);
}


sotNextStep::
~sotNextStep( void )
{
  sotDEBUGIN(5);

  sotDEBUGOUT(5);
  return;
}


/* --- FUNCTIONS ------------------------------------------------------------ */
/* --- FUNCTIONS ------------------------------------------------------------ */
/* --- FUNCTIONS ------------------------------------------------------------ */

void sotNextStep::
nextStep( const int & timeCurr )
{
  sotDEBUGIN(15);

  const unsigned& sfoot = contactFootSIN( timeCurr );
  const sotMatrixHomogeneous& wMlf = twoHandObserver.leftFootPositionSIN.access( timeCurr );
  const sotMatrixHomogeneous& wMrf = twoHandObserver.rightFootPositionSIN.access( timeCurr );

  sotMatrixHomogeneous ffMref;
  sotMatrixHomogeneous ref0Mff;
  sotMatrixHomogeneous sfMff;
  if( sfoot != 1 ) // --- left foot support ---
  {
    ffMref = referencePositionRightSIN.access( timeCurr );
    ref0Mff = ref0Mrf;
    sotMatrixHomogeneous sfMw; wMlf.inverse(sfMw); sfMw.multiply(wMrf, sfMff);
  }
  else // -- right foot support ---
  {
    ffMref = referencePositionLeftSIN.access( timeCurr );
    ref0Mff = ref0Mlf;
    sotMatrixHomogeneous sfMw; wMrf.inverse(sfMw); sfMw.multiply(wMlf, sfMff);
  }

  sotMatrixHomogeneous ref0Mref; ref0Mff.multiply(ffMref, ref0Mref);

  sotMatrixHomogeneous Pt = ref0Mff;
  for( int i=0;i<3;++i ){ Pt(i,3) = 0; }
  sotMatrixHomogeneous P; Pt.transpose(P);
  sotMatrixHomogeneous tmp; ref0Mref.multiply(Pt, tmp);
  sotMatrixHomogeneous ffMffd; P.multiply(tmp, ffMffd);

  sotMatrixHomogeneous sfMffd; sfMff.multiply(ffMffd, sfMffd);

  Pt = sfMffd;
  for( int i=0;i<3;++i ){ Pt(i,3) = 0; }
  Pt.transpose(P);

  sfMffd.multiply(Pt, tmp);
  sotMatrixHomogeneous Href; P.multiply(tmp, Href);

  sotMatrixRotation R; Href.extract(R);
  sotVectorRollPitchYaw rpy; rpy.fromMatrix(R);

  sotFootPrint newStep;
  newStep.contact = CONTACT_RIGHT_FOOT;
  newStep.x = Href(0, 3);
  newStep.y = Href(1, 3);
  newStep.theta = rpy(2) * 180 / 3.14159256;

  newStep.introductionTime = timeCurr;

  footPrintList.push_back( newStep );
  footPrintList.pop_front();

  sotDEBUGOUT(15);
}


void sotNextStep::thisIsZero()
{
  sotDEBUGIN(15);

  const sotMatrixHomogeneous& sfMref_r = referencePositionRightSIN.accessCopy();
  sfMref_r.inverse(ref0Mrf);

  const sotMatrixHomogeneous& sfMref_l = referencePositionLeftSIN.accessCopy();
  sfMref_l.inverse(ref0Mlf);

  sotDEBUGOUT(15);
}


void sotNextStep::
starter( const int & timeCurr )
{
  sotDEBUGIN(15); 

  footPrintList.clear();
  sotFootPrint initSteps[4];

  initSteps[0].contact = CONTACT_RIGHT_FOOT;
  initSteps[0].x = 0;
  initSteps[0].y = - zeroStepPosition/2;
  initSteps[0].theta = 0;
  initSteps[0].introductionTime = -1;
  footPrintList.push_back( initSteps[0] );
  introductionCallBack( -1 );

  initSteps[1].contact = CONTACT_LEFT_FOOT;
  initSteps[1].x = 0;
  initSteps[1].y = + zeroStepPosition;
  initSteps[1].theta = 0;
  initSteps[1].introductionTime = -1;
  footPrintList.push_back( initSteps[1] );
  introductionCallBack( -1 );

  initSteps[2].contact = CONTACT_RIGHT_FOOT;
  initSteps[2].x = 0;
  initSteps[2].y = - zeroStepPosition;
  initSteps[2].theta = 0;
  initSteps[2].introductionTime = -1;
  footPrintList.push_back( initSteps[2] );
  introductionCallBack( -1 );

  initSteps[3].contact = CONTACT_LEFT_FOOT;
  initSteps[3].x = 0;
  initSteps[3].y = + zeroStepPosition;
  initSteps[3].theta = 0;
  initSteps[3].introductionTime = -1;
  footPrintList.push_back( initSteps[3] );
  introductionCallBack( -1 );

  timeLastIntroduction = timeCurr - period+1;
  if( verbose ) (*verbose) << "NextStep started." << std::endl;

  sotDEBUGOUT(15); 
  return;
}

void sotNextStep::
stoper( const int & timeCurr )
{
  sotDEBUGIN(15); 

  sotDEBUGOUT(15); 
  return;
}


/* --- SIGNALS -------------------------------------------------------------- */
/* --- SIGNALS -------------------------------------------------------------- */
/* --- SIGNALS -------------------------------------------------------------- */

int& sotNextStep::
triggerCall( int& dummy,int timeCurrent )
{

  sotDEBUGIN(45);

  switch( state )
    {
    case STATE_STOPED: break;
    case STATE_STARTED: 
      {
	int nextIntoductionTime = timeLastIntroduction+period;
	if( nextIntoductionTime<=timeCurrent ) 
	  {
	    nextStep( timeCurrent );
	    if( NULL!=verbose )
	      {
		sotFootPrint & lastStep =  footPrintList.back();
		(*verbose) << "<T=" << timeCurrent << "> Introduced a new step: "; 
		switch( lastStep.contact )
		  {
		  case CONTACT_LEFT_FOOT: (*verbose) << "LF " ; break;
		  case CONTACT_RIGHT_FOOT: (*verbose) << "RF " ; break;
		  }
		(*verbose) << lastStep.x << "," << lastStep.y << "," 
			   << lastStep.theta << std::endl;
	      }
	    introductionCallBack( timeCurrent );
	    timeLastIntroduction=timeCurrent;
	  }
	break;
      }
    case STATE_STARTING:
      {
	starter( timeCurrent );
	break; 
      }
    case STATE_STOPING:
      {
	stoper( timeCurrent );
	break; 
      }
    };

  sotDEBUGOUT(45);

  return dummy;
}


/* --- PARAMS --------------------------------------------------------------- */
/* --- PARAMS --------------------------------------------------------------- */
/* --- PARAMS --------------------------------------------------------------- */

void sotNextStep::
display( std::ostream& os ) const
{
  os << "sotNextStep <" << getName() <<">:" << std::endl;
  for( std::deque< sotFootPrint >::const_iterator iter = footPrintList.begin();
       iter!=footPrintList.end();++iter )
    {
      os << "<time=" << iter->introductionTime << "> " ;
      switch( iter->contact )
	{
	case CONTACT_LEFT_FOOT: os << "LF " ; break;
	case CONTACT_RIGHT_FOOT: os << "RF " ; break;
	}
      os << "(" << iter->x << "," << iter->y << "," << iter->theta << ")" << std::endl;
    }
}


void sotNextStep::
commandLine( const std::string& cmdLine,
	     std::istringstream& cmdArgs,
	     std::ostream& os )
{
  

  if( cmdLine == "help" )
    {
      os << "NextStep: " << std::endl
	 << " - verbose [OFF]" << std::endl
	 << " - state [{start|stop}] \t get/set the stepper state. " << std::endl
	 << " - yZeroStep [<value>] \t get/set the Y default position." << std::endl
	 << " - thisIsZero {record|disp}" << std::endl
	 << std::endl;
    }
  else if( cmdLine == "state" ) 
    {
      cmdArgs >> std::ws;
      if( cmdArgs.good() )
	{
	  std::string statearg; cmdArgs >> statearg;
	  if( statearg == "start" ) { state = STATE_STARTING; }
	  else if( statearg == "stop" ) { state = STATE_STOPING; }
	}
      else 
	{
	  os <<"state = ";
	  switch( state ) 
	    {
	    case STATE_STARTING: os << "starting"; break;
	    case STATE_STOPING: os << "stoping"; break;
	    case STATE_STARTED: os << "started"; break;
	    case STATE_STOPED: os << "stoped"; break;
	    default: os <<"error"; break;
	    }
	  os << std::endl;
	}
    }
  else if( cmdLine == "yZeroStep" )
    {
      cmdArgs >> std::ws; if( cmdArgs.good() )
	{ cmdArgs >> zeroStepPosition; }
      else { os << "yzero = " << zeroStepPosition; }
    }
  else if( cmdLine == "thisIsZero" )
    {
      std::string arg; cmdArgs >> arg; 
      if( arg == "disp_left" ) { os << "zero_left = " << ref0Mlf; }
      else if( arg == "disp_right" ) { os << "zero_right = " << ref0Mrf; }
      else if( arg == "record" ) { thisIsZero(); }
    }
  else if( cmdLine == "verbose" ) 
    {
      cmdArgs >> std::ws; std::string offarg;
      if( (cmdArgs.good())&&(cmdArgs>>offarg,offarg=="OFF") ) { verbose =  NULL; }
      else { verbose = &os; }
    }
  else { sotEntity::commandLine( cmdLine,cmdArgs,os); }
}


/* --- TWO HAND -------------------------------------------------------------- */
/* --- TWO HAND -------------------------------------------------------------- */
/* --- TWO HAND -------------------------------------------------------------- */

sotNextStepTwoHandObserver::
sotNextStepTwoHandObserver( const std::string & name )
  :referencePositionLeftSIN( NULL,"NextStepTwoHandObserver("+name+")::input(vector)::positionLeft" )
  ,referenceVelocityLeftSIN( NULL,"NextStepTwoHandObserver("+name+")::input(vector)::velocityLeft" )
  ,referenceAccelerationLeftSIN( NULL,"NextStepTwoHandObserver("+name+")::input(vector)::accelerationLeft" )
  ,leftFootPositionSIN( NULL,"sotNextStepTwoHandObserver("+name+")::input(matrixhomo)::leftfoot" )

  ,referencePositionRightSIN( NULL,"NextStepTwoHandObserver("+name+")::input(vector)::positionRight" )
  ,referenceVelocityRightSIN( NULL,"NextStepTwoHandObserver("+name+")::input(vector)::velocityRight" )
  ,referenceAccelerationRightSIN( NULL,"NextStepTwoHandObserver("+name+")::input(vector)::accelerationRight" )
  ,rightFootPositionSIN( NULL,"sotNextStepTwoHandObserver("+name+")::input(matrixhomo)::rightfoot" )  
  
  ,referencePositionLeftSOUT( boost::bind(&sotNextStepTwoHandObserver::computeReferencePositionLeft,this,_1,_2),
			      leftFootPositionSIN<<referencePositionLeftSIN<<referencePositionRightSIN,
			      "sotNextStepTwoHandObserver("+name+")::output(vector)::position2handLeft" ) 
  ,referencePositionRightSOUT( boost::bind(&sotNextStepTwoHandObserver::computeReferencePositionRight,this,_1,_2),
			       rightFootPositionSIN<<referencePositionLeftSIN<<referencePositionRightSIN,
			       "sotNextStepTwoHandObserver("+name+")::output(vector)::position2handRight" ) 

  ,referenceVelocitySOUT( SOT_MEMBER_SIGNAL_2( sotNextStepTwoHandObserver::computeReferenceVelocity,
					       referenceVelocityLeftSIN,ml::Vector,
					       referenceVelocityRightSIN,ml::Vector),
			  "NextStepTwoHandObserver("+name+")::output(vector)::velocity2Hand" )
  ,referenceAccelerationSOUT( SOT_MEMBER_SIGNAL_2( sotNextStepTwoHandObserver::computeReferenceAcceleration,
						   referenceAccelerationLeftSIN,ml::Vector,
						   referenceAccelerationRightSIN,ml::Vector),
			      "NextStepTwoHandObserver("+name+")::output(vector)::acceleration2Hand" )
{ sotDEBUGINOUT(25); }


sotSignalArray<int> sotNextStepTwoHandObserver::
getSignals( void )
{
  return (referencePositionLeftSIN << referenceVelocityLeftSIN 
	  << referenceAccelerationLeftSIN
	  << leftFootPositionSIN
	  << referencePositionRightSIN << referenceVelocityRightSIN 
	  << referenceAccelerationRightSIN
	  << rightFootPositionSIN
	  << referencePositionLeftSOUT << referencePositionRightSOUT
	  << referenceVelocitySOUT << referenceAccelerationSOUT);


}
sotNextStepTwoHandObserver::
operator sotSignalArray<int> ()
{

  return (referencePositionLeftSIN << referenceVelocityLeftSIN 
	  << referenceAccelerationLeftSIN
	  << leftFootPositionSIN
	  << referencePositionRightSIN << referenceVelocityRightSIN 
	  << referenceAccelerationRightSIN
	  << rightFootPositionSIN
	  << referencePositionLeftSOUT << referencePositionRightSOUT
	  << referenceVelocitySOUT << referenceAccelerationSOUT);
}

sotMatrixHomogeneous& sotNextStepTwoHandObserver::
computeRefPos( sotMatrixHomogeneous& res,int timeCurr,const sotMatrixHomogeneous& wMsf )
{
  sotDEBUGIN(15);

#define RIGHT_HAND_REFERENCE 0
#if RIGHT_HAND_REFERENCE

  const sotMatrixHomogeneous& wMrh = referencePositionRightSIN( timeCurr );  
  sotMatrixHomogeneous sfMw; wMsf.inverse(sfMw);
  sfMw.multiply(wMrh, res);

#else

  const sotMatrixHomogeneous& wMlh = referencePositionLeftSIN( timeCurr );
  const sotMatrixHomogeneous& wMrh = referencePositionRightSIN( timeCurr );  

  sotMatrixHomogeneous sfMw; wMsf.inverse(sfMw);
  sotMatrixHomogeneous sfMlh; sfMw.multiply(wMlh, sfMlh);
  sotMatrixHomogeneous sfMrh; sfMw.multiply(wMrh, sfMrh);

  sotMatrixRotation R;
  sotVectorRollPitchYaw rpy;

  ml::Vector prh(3); sfMrh.extract(prh);
  sfMrh.extract(R);
  sotVectorRollPitchYaw rpy_rh; rpy_rh.fromMatrix(R);

  ml::Vector plh(3); sfMlh.extract(plh);
  sfMlh.extract(R);
  sotVectorRollPitchYaw rpy_lh; rpy_lh.fromMatrix(R);

  ml::Vector p = .5 * (plh + prh);
  for(int i = 0; i < 3; ++i){ rpy(i) = 0.5 * (rpy_rh(i) + rpy_lh(i)); }

  rpy.toMatrix(R);
  res.buildFrom(R, p);

#endif
  
  sotDEBUGOUT(15);
  return res;
}

sotMatrixHomogeneous& sotNextStepTwoHandObserver::
computeReferencePositionLeft( sotMatrixHomogeneous& res,int timeCurr )
{
  sotDEBUGIN(15);

  const sotMatrixHomogeneous& wMsf = leftFootPositionSIN( timeCurr );

  sotDEBUGOUT(15);
  return computeRefPos( res,timeCurr,wMsf );
}




sotMatrixHomogeneous& sotNextStepTwoHandObserver::
computeReferencePositionRight( sotMatrixHomogeneous& res,int timeCurr )
{
  sotDEBUGIN(15);

  const sotMatrixHomogeneous& wMsf = rightFootPositionSIN( timeCurr );

  sotDEBUGOUT(15);
  return computeRefPos( res,timeCurr,wMsf );
}


ml::Vector& sotNextStepTwoHandObserver::
computeReferenceVelocity( const ml::Vector& right,
			  const ml::Vector& left,
			  ml::Vector& res )
{
  sotDEBUGIN(15);

  /* TODO */
  
  sotDEBUGOUT(15);
  return res;
}


ml::Vector& sotNextStepTwoHandObserver::
computeReferenceAcceleration( const ml::Vector& right,
			      const ml::Vector& left,
			      ml::Vector& res )
{
  sotDEBUGIN(15);

  /* TODO */
  
  sotDEBUGOUT(15);
  return res;
}


