/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotNextStep.h
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#ifndef __SOT_SOTNEXTSTEP_H__
#define __SOT_SOTNEXTSTEP_H__


/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* Matrix */
#include <MatrixAbstractLayer/boost.h>
namespace ml = maal::boost;

/* SOT */
#include <sot/sotEntity.h>
#include <sot/sotSignalPtr.h>
#include <sot/sotSignalTimeDependant.h>
#include <sot/sotMatrixHomogeneous.h>
#include <sot/sotVectorRollPitchYaw.h>
#include <sot/sotMatrixRotation.h>

/* STD */
#include <string>
#include <deque>


/* --------------------------------------------------------------------- */
/* --- API ------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

#if defined (WIN32) 
#  if defined (sotNextStep_EXPORTS) 
#    define SOTNEXTSTEP_EXPORT __declspec(dllexport)
#  else  
#    define SOTNEXTSTEP_EXPORT __declspec(dllimport)
#  endif 
#else
#  define SOTNEXTSTEP_EXPORT
#endif


/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */


/* --- Two Hand Observer ----------------------------------------------- */

/*!
 * Computes a reference frame in either left or right foot
 * coordinates, based on the position of the hands. Used by the
 * stepper to compute the reference foot prints to send to
 * the Pattern Generator.
*/
class SOTNEXTSTEP_EXPORT sotNextStepTwoHandObserver
{
 public:

  sotSignalPtr< sotMatrixHomogeneous,int > referencePositionLeftSIN; 
  sotSignalPtr< ml::Vector,int > referenceVelocityLeftSIN; 
  sotSignalPtr< ml::Vector,int > referenceAccelerationLeftSIN; 
  sotSignalPtr< sotMatrixHomogeneous,int > leftFootPositionSIN;

  sotSignalPtr< sotMatrixHomogeneous,int > referencePositionRightSIN; 
  sotSignalPtr< ml::Vector,int > referenceVelocityRightSIN; 
  sotSignalPtr< ml::Vector,int > referenceAccelerationRightSIN; 
  sotSignalPtr< sotMatrixHomogeneous,int > rightFootPositionSIN;

  sotSignalTimeDependant< sotMatrixHomogeneous,int > referencePositionLeftSOUT; 
  sotSignalTimeDependant< sotMatrixHomogeneous,int > referencePositionRightSOUT; 
  sotSignalTimeDependant< ml::Vector,int > referenceVelocitySOUT; 
  sotSignalTimeDependant< ml::Vector,int > referenceAccelerationSOUT; 

 public:

  sotNextStepTwoHandObserver( const std::string & name );

  sotMatrixHomogeneous& computeReferencePositionLeft( sotMatrixHomogeneous& res,int timeCurr );
  sotMatrixHomogeneous& computeReferencePositionRight( sotMatrixHomogeneous& res,int timeCurr );
  ml::Vector& computeReferenceVelocity( const ml::Vector& right,const ml::Vector& left,ml::Vector& res );
  ml::Vector& computeReferenceAcceleration( const ml::Vector& right,const ml::Vector& left,ml::Vector& res );

  sotSignalArray<int> getSignals( void );
  operator sotSignalArray<int> ();

 private:

  sotMatrixHomogeneous& computeRefPos( sotMatrixHomogeneous& res,int timeCurr,const sotMatrixHomogeneous& wMsf );
};


/* --- Next Step (stepper) --------------------------------------------- */

/*!
 * Uses the reference frame computed by the observer to
 * compute foot print references that are sent to the
 * Pattern Generator. The current implementation compares
 * the position of the reference frame to a 'zero' position
 * recorded using the command stepper.thisIsZero record.
 */
class SOTNEXTSTEP_EXPORT sotNextStep
  : public sotEntity
{
 public:

  static const std::string CLASS_NAME;
  virtual const std::string& getClassName( void ) const { return CLASS_NAME; }

 protected: /* --- FOOT PRINT LIST --- */

  enum sotContactName
  {
    CONTACT_LEFT_FOOT
    ,CONTACT_RIGHT_FOOT
  };
  class sotFootPrint
  {
  public:
    double x,y,theta;
    sotContactName contact;
    int introductionTime;
  };
  std::deque< sotFootPrint > footPrintList;

 protected: /* --- INTRODUCTION PERIOD --- */

  unsigned int period;
  static const unsigned int PERIOD_DEFAULT;
  int timeLastIntroduction;

 protected: /* --- STATE --- */

  enum sotSteppingState
  {
    STATE_STARTING   //! Introducing 4 steps then switches to STATE_STARTED.
    ,STATE_STOPING   //! Running but stop requested: introduce a last step and stop.
    ,STATE_STARTED   //! Running, simply introduce steps.
    ,STATE_STOPED    //! Nothing to do, cannot introduce steps in the FIFO
  };
  sotSteppingState state;

 protected: /* --- STEPPING --- */

  double zeroStepPosition;
  const static double ZERO_STEP_POSITION_DEFAULT; // = .19

 protected: /* --- REFERENCE FRAME --- */

  sotMatrixHomogeneous ref0Mrf;
  sotMatrixHomogeneous ref0Mlf;
  sotNextStepTwoHandObserver twoHandObserver;

  void thisIsZero();

 protected: /* --- DEBUG --- */

  std::ostream* verbose;


 public: /* --- CONSTRUCTION --- */
  sotNextStep( const std::string& name );
  virtual ~sotNextStep( void );

 public: /* --- SIGNAL --- */

  sotSignalPtr< sotMatrixHomogeneous,int > referencePositionLeftSIN; 
  sotSignalPtr< sotMatrixHomogeneous,int > referencePositionRightSIN; 
  sotSignalPtr< unsigned,int > contactFootSIN;

  sotSignal< int,int > triggerSOUT; 


 public: /* --- FUNCTIONS --- */

  static const ml::Vector VECTOR_NULL_ARG;

  virtual void nextStep( const int & timeCurr );
  virtual void starter( const int & timeCurr );
  virtual void stoper( const int & timeCurr );
		 
  virtual void introductionCallBack( const int & timeCurr ) {};

  int& triggerCall( int& dummy,int timeCurr );

 public: /* --- PARAMS --- */

  virtual void display( std::ostream& os ) const; 
  virtual void commandLine( const std::string& cmdLine,
			    std::istringstream& cmdArgs,
			    std::ostream& os );
};


#endif // #ifndef __SOT_SOTNEXTSTEP_H__

