/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotDynamic.h
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/



#ifndef __SOT_DYNAMIC_H__
#define __SOT_DYNAMIC_H__

/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* Matrix */
#include <MatrixAbstractLayer/boost.h>
namespace ml = maal::boost;

/* JRL dynamic */
#include <dynamicsJRLJapan/Joint.h>
#include <dynamicsJRLJapan/HumanoidDynamicMultiBody.h>
namespace djj = dynamicsJRLJapan;

/* SOT */
#include <sot/sotFlags.h>
#include <sot/sotEntity.h>
#include <sot/sotPool.h>
#include <sot/sotSignalPtr.h>
#include <sot/sotSignalTimeDependant.h>
#include <sot/sotExceptionDynamic.h>
#include <sot/sotMatrixHomogeneous.h>
#include <sot/sotMatrixInertia.h>

/* STD */
#include <string>

/* --------------------------------------------------------------------- */
/* --- API ------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

#if defined (WIN32) 
#  if defined (sotDynamic_EXPORTS)
#    define SOTDYNAMIC_EXPORT __declspec(dllexport)
#  else  
#    define SOTDYNAMIC_EXPORT __declspec(dllimport)
#  endif 
#else
#  define SOTDYNAMIC_EXPORT
#endif

/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/*! @ingroup signals
  \brief This class provides an inverse dynamic model of the robot.
  More precisely it wraps the newton euler algorithm implemented
  by the dynamicsJRLJapan library to make it accessible in the stack of tasks.
  The robot is described by a VRML file.
*/
  
class SOTDYNAMIC_EXPORT sotDynamic
:public sotEntity
{
 public:
  static const std::string CLASS_NAME;

 protected:
 public:

  /*! \name Fields to access dynamicsJRLJapan Library 
    @{
   */

  /*! \brief Abstract pointer on the structure.
    Ultimately we should be able to use only the abstract
    interface and do not care about the implementation.
  */
  CjrlHumanoidDynamicRobot* aHDR;

  /*! \brief Pointer on the specific dynamicsJRLJapan 
    data structure */
  djj::HumanoidDynamicMultiBody* aHDMB;
  //djj::DynamicMultiBody * aDMB;

  sotMatrixInertia inertia;
  int debugInertia;

  /*! \brief Fields to access the humanoid model 
   @{ */

  /*! \brief Directory where the VRML humanoid model is stored */
  std::string vrmlDirectory;
  /*! \brief Name of the root's robot model file */
  std::string vrmlMainFile;
  /*! \brief Name of the name specifying which end-effector is
    the head, the feet and so on... */
  std::string xmlSpecificityFile;
  /*! \brief Name of the name specifying which end-effector is
    the head, the feet and so on... */
  std::string xmlRankFile;
  /*! @} */

  /*! @} */
  bool init;
  std::list< sotSignalAbstract<int>*  > genericSignalRefs;

 public: /* --- CONSTRUCTION --- */

  sotDynamic( const std::string& name, bool build=true );
  virtual ~sotDynamic( void );

 public: /* --- MODEL CREATION --- */
 
  virtual void buildModel( void );

  void setVrmlDirectory( const std::string& filename );
  void setVrmlMainFile( const std::string& filename );
  void setXmlSpecificityFile( const std::string& filename );
  void setXmlRankFile( const std::string& filename );
  void parseConfigFiles( void );

 public: /* --- SIGNAL ACTIVATION --- */
  sotSignalTimeDependant< ml::Matrix,int > &
    createEndeffJacobianSignal( const std::string& signame,
				   const unsigned int & bodyRank );
  sotSignalTimeDependant< ml::Matrix,int > &
    createJacobianSignal( const std::string& signame,
			  const unsigned int & bodyRank );
  void destroyJacobianSignal( const std::string& signame );
  sotSignalTimeDependant< sotMatrixHomogeneous,int >&
    createPositionSignal( const std::string& signame,
			  const unsigned int & bodyRank );
  void destroyPositionSignal( const std::string& signame );
  sotSignalTimeDependant< ml::Vector,int >&
    createAccelerationSignal( const std::string& signame,
			     const unsigned int & bodyRank );
  void destroyAccelerationSignal( const std::string& signame );

  bool zmpActivation( void ) { return aHDMB->getComputeZMP(); }
  void zmpActivation( const bool& b ) { return aHDMB->setComputeZMP(b); }
  bool comActivation( void ) { return aHDMB->getComputeCoM(); }
  void comActivation( const bool& b ) { return aHDMB->setComputeCoM(b); }

 public: /* --- SIGNAL --- */

  sotSignalPtr<ml::Vector,int> jointPositionSIN;
  sotSignalPtr<ml::Vector,int> freeFlyerPositionSIN;
  sotSignalPtr<ml::Vector,int> jointVelocitySIN;
  sotSignalPtr<ml::Vector,int> freeFlyerVelocitySIN;
  sotSignalPtr<ml::Vector,int> jointAccelerationSIN;
  sotSignalPtr<ml::Vector,int> freeFlyerAccelerationSIN;

  // protected:
 public:
  typedef int Dummy;
  sotSignalTimeDependant<Dummy,int> firstSINTERN;
  sotSignalTimeDependant<Dummy,int> newtonEulerSINTERN;

  int& computeNewtonEuler( int& dummy,int time );
  int& initNewtonEuler( int& dummy,int time );

 public:
  sotSignalTimeDependant<ml::Vector,int> zmpSOUT;
  sotSignalTimeDependant<ml::Matrix,int> JcomSOUT;
  sotSignalTimeDependant<ml::Vector,int> comSOUT;
  sotSignalTimeDependant<ml::Matrix,int> inertiaSOUT;

  sotSignalTimeDependant<ml::Matrix,int>& jacobiansSOUT( const std::string& name );
  sotSignalTimeDependant<sotMatrixHomogeneous,int>& positionsSOUT( const std::string& name );
  sotSignalTimeDependant<ml::Vector,int>& accelerationsSOUT( const std::string& name );

  sotSignalTimeDependant<double,int> footHeightSOUT;
  sotSignalTimeDependant<ml::Vector,int> upperJlSOUT;
  sotSignalTimeDependant<ml::Vector,int> lowerJlSOUT;

  sotSignal<ml::Vector,int> inertiaRotorSOUT;
  sotSignal<ml::Vector,int> gearRatioSOUT;
  sotSignalTimeDependant<ml::Matrix,int> inertiaRealSOUT;
  sotSignalTimeDependant<ml::Vector,int> MomentaSOUT;
  sotSignalTimeDependant<ml::Vector,int> AngularMomentumSOUT;
  sotSignalTimeDependant<ml::Matrix,int> JAngularMomentumSOUT;

 protected:
  ml::Vector& computeZmp( ml::Vector& res,int time );
  ml::Vector& computeMomenta( ml::Vector &res, int time);
  ml::Vector& computeAngularMomentum( ml::Vector &res, int time);
  ml::Matrix& computeJcom( ml::Matrix& res,int time );
  ml::Vector& computeCom( ml::Vector& res,int time );
  ml::Matrix& computeInertia( ml::Matrix& res,int time );
  ml::Matrix& computeInertiaReal( ml::Matrix& res,int time );
  ml::Matrix& computeJAngularMomentum( ml::Matrix& res,int time );
  double& computeFootHeight( double& res,int time );
  
  ml::Matrix& computeGenericJacobian( djj::Joint* j,ml::Matrix& res,int time );
  ml::Matrix& computeGenericEndeffJacobian( djj::Joint* j,ml::Matrix& res,int time );
  sotMatrixHomogeneous& computeGenericPosition( djj::Joint* j,sotMatrixHomogeneous& res,int time );
  ml::Vector& computeGenericAcceleration( djj::Joint* j,ml::Vector& res,int time );

  ml::Vector& getUpperJointLimits( ml::Vector& res,const int& time );
  ml::Vector& getLowerJointLimits( ml::Vector& res,const int& time );


 public: /* --- PARAMS --- */
  virtual void commandLine( const std::string& cmdLine,
			    std::istringstream& cmdArgs,
			    std::ostream& os );
    

};





#endif // #ifndef __SOT_DYNAMIC_H__

