/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotIntegratorForce.h
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <sot/sotIntegratorForce.h>
#include <sot/sotDebug.h>
#include <sot/sotFactory.h>

SOT_FACTORY_ENTITY_PLUGIN(sotIntegratorForce,"IntegratorForce");

const double sotIntegratorForce:: TIME_STEP_DEFAULT = 5e-3;

sotIntegratorForce::
sotIntegratorForce( const std::string & name ) 
  :sotEntity(name)
   ,timeStep( TIME_STEP_DEFAULT )
   ,forceSIN(NULL,"sotIntegratorForce("+name+")::input(vector)::force")
   ,massInverseSIN(NULL,"sotIntegratorForce("+name+")::input(matrix)::massInverse")
   ,frictionSIN(NULL,"sotIntegratorForce("+name+")::input(matrix)::friction")
   ,velocityPrecSIN(NULL,"sotIntegratorForce("+name+")::input(matrix)::vprec")
  
   ,velocityDerivativeSOUT
  ( boost::bind(&sotIntegratorForce::computeDerivative,this,_1,_2),
    velocityPrecSIN<<forceSIN<<massInverseSIN<<frictionSIN,
    "sotIntegratorForce("+name+")::output(Vector)::velocityDerivative" )
   ,velocitySOUT( boost::bind(&sotIntegratorForce::computeIntegral,this,_1,_2),
		 velocityPrecSIN<<velocityDerivativeSOUT,
		 "sotIntegratorForce("+name+")::output(Vector)::velocity" )

   ,massSIN(NULL,"sotIntegratorForce("+name+")::input(matrix)::mass")
  ,massInverseSOUT( boost::bind(&sotIntegratorForce::computeMassInverse,this,_1,_2),
		    massSIN,
		    "sotIntegratorForce("+name+")::input(matrix)::massInverseOUT")
{
  sotDEBUGIN(5);
  
  signalRegistration( forceSIN                   << massInverseSIN
		      << frictionSIN             << velocityPrecSIN
		      << velocityDerivativeSOUT  << velocitySOUT 
		      << massInverseSOUT         << massSIN );

  massInverseSIN.plug( &massInverseSOUT );

  sotDEBUGOUT(5);
}


sotIntegratorForce::
~sotIntegratorForce( void )
{
  sotDEBUGIN(5);

  sotDEBUGOUT(5);
  return;
}

/* --- SIGNALS -------------------------------------------------------------- */
/* --- SIGNALS -------------------------------------------------------------- */
/* --- SIGNALS -------------------------------------------------------------- */

/* The derivative of the signal is such that: M v_dot + B v = f. We deduce:
 * v_dot =  M^-1 (f - Bv)
 */
ml::Vector& sotIntegratorForce::
computeDerivative( ml::Vector& res,
		   const int& time )
{
  sotDEBUGIN(15);

  const ml::Vector & force = forceSIN( time );
  const ml::Matrix & massInverse = massInverseSIN( time );
  const ml::Matrix & friction = frictionSIN( time );

  sotDEBUG(15) << "force = " << force << std::endl;

  ml::Vector f_bv( force.size() );

  if( velocityPrecSIN )
    { 
      const ml::Vector & vel = velocityPrecSIN( time );
      sotDEBUG(15) << "vel = " << vel << std::endl;
      friction .multiply( vel,f_bv ); f_bv *= -1; 
    } else { f_bv.fill(0) ; } // vel is not set yet.
      
  f_bv+=force;
  massInverse.multiply( f_bv, res );
  
  sotDEBUGOUT(15);
  return res;
}

ml::Vector& sotIntegratorForce::
computeIntegral( ml::Vector& res,
		 const int& time )
{
  sotDEBUGIN(15);

  const ml::Vector & dvel =velocityDerivativeSOUT( time );
  res = dvel;
  res *= timeStep;
  if( velocityPrecSIN )
    {
      const ml::Vector & vel = velocityPrecSIN( time );
      res += vel; 
    } 
  else { /* vprec not set yet. */ }
  velocityPrecSIN = res ;

  sotDEBUGOUT(15);
  return res;
}

ml::Matrix& sotIntegratorForce::
computeMassInverse( ml::Matrix& res,
		    const int& time )
{
  sotDEBUGIN(15);

  const ml::Matrix & mass = massSIN( time );
  mass.inverse( res );

  sotDEBUGOUT(15);
  return res;
}


/* --- PARAMS --------------------------------------------------------------- */
/* --- PARAMS --------------------------------------------------------------- */
/* --- PARAMS --------------------------------------------------------------- */
void sotIntegratorForce::
commandLine( const std::string& cmdLine,
	     std::istringstream& cmdArgs,
	     std::ostream& os )
{
  sotDEBUG(25) << "Cmd " << cmdLine <<std::endl;

  if( cmdLine == "help" )
    {
      os << "IntegratorForce: " 
	 << "  - dt [<time>]: get/set the timestep value. " << std::endl;
    }
  else if( cmdLine == "dt" )
    {
      cmdArgs >>std::ws; 
      if( cmdArgs.good() )
	{
	  double val; cmdArgs>>val; 
	  if( val >0. ) timeStep = val;
	}
      else { os << "TimeStep = " << timeStep << std::endl;}
    }
  else { sotEntity::commandLine( cmdLine,cmdArgs,os); }
}

