/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotIntegratorForceExact.h
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <sot/sotIntegratorForceExact.h>
#include <sot/sotDebug.h>
#include <sot/sotFactory.h>
#include <sot/sotExceptionDynamic.h>

SOT_FACTORY_ENTITY_PLUGIN(sotIntegratorForceExact,"IntegratorForceExact");

sotIntegratorForceExact::
sotIntegratorForceExact( const std::string & name ) 
  :sotIntegratorForce(name)
{
  sotDEBUGIN(5);
  
  velocitySOUT.
    setFunction( boost::bind(&sotIntegratorForceExact::computeVelocityExact,
			     this,_1,_2));
  velocitySOUT.removeDependancy( velocityDerivativeSOUT );

  sotDEBUGOUT(5);
}


sotIntegratorForceExact::
~sotIntegratorForceExact( void )
{
  sotDEBUGIN(5);

  sotDEBUGOUT(5);
  return;
}

/* --- EIGEN VALUE ---------------------------------------------------------- */
/* --- EIGEN VALUE ---------------------------------------------------------- */
/* --- EIGEN VALUE ---------------------------------------------------------- */

#include <boost/numeric/bindings/lapack/geev.hpp>

static void eigenDecomp( const ml::Matrix& M,
			 ml::Matrix& P,
			 ml::Vector& eig )
{
  unsigned int SIZE = M.nbCols();
  ublas::matrix<double, ublas::column_major> Y=M.matrix;
  ublas::vector< std::complex<double> > evals(SIZE);
  ublas::matrix<double, ublas::column_major> vl(SIZE,SIZE);
  ublas::matrix<double, ublas::column_major> vr(SIZE,SIZE);
  
  const int errCode = lapack::geev(Y, evals, &vl, &vr, lapack::optimal_workspace());
  if( errCode<0 )
    { SOT_THROW sotExceptionDynamic( sotExceptionDynamic::INTEGRATION,
				     "Invalid argument to geev","" ); }
  else if( errCode>0 )
    { SOT_THROW sotExceptionDynamic( sotExceptionDynamic::INTEGRATION,
				     "No convergence for given matrix",""); }
  
  P.resize(SIZE,SIZE);   eig.resize(SIZE);
  for( unsigned int i=0;i<SIZE;++i )
    {
      for( unsigned int j=0;j<SIZE;++j ){ P(i,j)=vr(i,j); }
      eig(i)=evals(i).real();
      if( fabsf(evals(i).imag())>1e-5 ) 
	{ 
	  SOT_THROW sotExceptionDynamic( sotExceptionDynamic::INTEGRATION,
					 "Error imaginary part not null. ",
					 "(at position %d: %f)",i,evals(i).imag() );
	}
    }
  return ;
}

static void expMatrix( const ml::Matrix& MiB,
		       ml::Matrix& Mexp )
{
  unsigned int SIZE = MiB.nbCols();

  ml::Matrix Pmib(MiB.nbCols(),MiB.nbCols());
  ml::Vector eig_mib(MiB.nbCols());
  eigenDecomp( MiB,Pmib,eig_mib );
  sotDEBUG(45) << "V = " << Pmib;
  sotDEBUG(45) << "d = " << eig_mib;

  ml::Matrix Pinv(SIZE,SIZE); Pmib.inverse(Pinv);
  sotDEBUG(45) << "Vinv = " << Pinv;

  Mexp.resize(SIZE,SIZE);
  for( unsigned int i=0;i<SIZE;++i )
    for( unsigned int j=0;j<SIZE;++j )
      Pmib(i,j)*= exp( eig_mib(j) );
  Pmib.multiply(Pinv,Mexp);
  
  sotDEBUG(45) << "expMiB = " << Mexp;
  return ;
}

/* --- SIGNALS -------------------------------------------------------------- */
/* --- SIGNALS -------------------------------------------------------------- */
/* --- SIGNALS -------------------------------------------------------------- */

/* The derivative of the signal is such that: M v_dot + B v = f. We deduce:
 * v_dot =  M^-1 (f - Bv)
 * Using Exact method: 
 * dv = exp( M^-1.B.t) ( v0-B^-1.f ) + B^-1.f
 */

ml::Vector& sotIntegratorForceExact::
computeVelocityExact( ml::Vector& res,
		      const int& time )
{
  sotDEBUGIN(15);

  const ml::Vector & force = forceSIN( time );
  const ml::Matrix & massInverse = massInverseSIN( time );
  const ml::Matrix & friction = frictionSIN( time );
  unsigned int nf = force.size(), nv = friction.nbCols();
  res.resize(nv); res.fill(0);

  if(! velocityPrecSIN )
    { 
      ml::Vector zero( nv ); zero.fill(0);
      velocityPrecSIN = zero;
    } 
  const ml::Vector & vel = velocityPrecSIN( time );
  double & dt = this->sotIntegratorForce::timeStep; // this is &

  sotDEBUG(15) << "force = " << force;
  sotDEBUG(15) << "vel = " << vel;
  sotDEBUG(25) << "Mi = " << massInverse;
  sotDEBUG(25) << "B = " << friction;
  sotDEBUG(25) << "dt = " << dt << std::endl;

  ml::Matrix MiB( nv,nv );
  massInverse.multiply(friction,MiB);
  sotDEBUG(25) << "MiB = " << MiB;
  
  ml::Matrix MiBexp( nv,nv );
  MiB*=-dt; expMatrix(MiB,MiBexp);
  sotDEBUG(25) << "expMiB = " << MiBexp;

  ml::Matrix Binv( nv,nv ); friction.inverse(Binv);
  ml::Vector Bif( nf ); Binv.multiply( force,Bif );
  sotDEBUG(25) << "Binv = " << Binv;
  sotDEBUG(25) << "Bif = " << Bif;
  
  ml::Vector v0_bif = vel;
  v0_bif -= Bif;
  sotDEBUG(25) << "Kst = " << v0_bif;
  
  res.resize( MiBexp.nbRows() );
  MiBexp.multiply( v0_bif,res );

  res += Bif;
  velocityPrecSIN = res ;
  sotDEBUG(25) << "vfin = " << res;
 

  sotDEBUGOUT(15);
  return res;
}


/* --- PARAMS --------------------------------------------------------------- */
/* --- PARAMS --------------------------------------------------------------- */
/* --- PARAMS --------------------------------------------------------------- */
// void sotIntegratorForceExact::
// commandLine( const std::string& cmdLine,
// 	     std::istringstream& cmdArgs,
// 	     std::ostream& os )
// {
//   sotDEBUG(25) << "Cmd " << cmdLine <<std::endl;

//   if( cmdLine == "help" )
//     {
//       os << "IntegratorForceExact: " << std::endl;
//     }
//   else { sotIntegratorForce::commandLine( cmdLine,cmdArgs,os); }
// }

