/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotEntity.cpp
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/*! Local framework includes */
#include <sot/sotEntity.h>
#include <sot/sotPool.h>
#include <sot/sotDebug.h>

/*! System includes */
#include <stdlib.h>
#include <sstream>


using namespace std;

const std::string sotEntity::CLASS_NAME = "Entity";


void sotEntity::
entityRegistration( void ) 
{
  
  //sotPool.entity[name] = this;
  sotPool.registerEntity(name,this);
}

void sotEntity::
entityDeregistration( void ) 
{
  sotPool.deregisterEntity(name);
}

sotEntity::
sotEntity( const string& name__ )
  : name(name__) 
{
  sotDEBUG(15) << "New entity <"<<name__<<">"<<endl;
  if( name.length()==0 ) 
    {
      stringstream oss; oss << rand();
      name = CLASS_NAME; 
      name+="::";
      name+=oss.str();
    }

  entityRegistration();
}

sotEntity::
~sotEntity( void )
{
  sotDEBUG(25) << "# In (" << name << " { " << endl;
  entityDeregistration();
  sotDEBUGOUT(25);
}

/* -------------------------------------------------------------------------- */
/* --- SIGNALS -------------------------------------------------------------- */
/* -------------------------------------------------------------------------- */
void sotEntity::
signalRegistration( const sotSignalArray<int>& signals )
{
  for( unsigned int i=0;i<signals.getSize();++i ) 
    {
      sotSignalAbstract<int>& sig = signals[i];
      //const string& signame = sig.getName();
      istringstream iss( sig.getName() );
      const int SIZE = 128;
      char buffer[SIZE];
      while( iss.good() )
	{ iss.getline(buffer,SIZE,':'); }
      const string& signame( buffer );

      SignalMap::iterator sigkey = signalMap.find(signame);
      if( sigkey != signalMap.end() ) // key does exist
	{
	  sotERRORF( "Key %s already exist in the signalMap.",signame.c_str() );
	  if( sigkey->second!=&sig )
	    {
	      throw sotExceptionFactory( sotExceptionFactory::SIGNAL_CONFLICT,
					 "Another signal already defined with the same name. ",
					 "Signame is <%s>.",signame.c_str() );
	    }
	}
      else
	{
	  sotDEBUG(10) << "Register signal <"<< signame << "> for entity <"
		        << getName() << "> ."<<endl;
	  signalMap[signame] = &sig;
	}
    }
}


void sotEntity::
signalDeregistration( const std::string& signame )
{
  SignalMap::iterator sigkey = signalMap.find(signame);
  if( sigkey == signalMap.end() ) // key does not exist
    {
      sotERRORF( "Key %s does not exist in the signalMap.",signame.c_str() );
      throw sotExceptionFactory( sotExceptionFactory::UNREFERED_SIGNAL,
				 "No signal defined with the given name. ",
				 " (while erasing <%s>).",signame.c_str() );
    }
  else
    {
      sotDEBUG(10) << "Deregister signal <"<< signame << "> for entity <"
		   << getName() << "> ."<<endl;
      signalMap.erase(signame);
    }
}


#define __SOT_ENTITY_GET_SIGNAL__(ITER_TYPE) \
  SignalMap::ITER_TYPE sigkey = signalMap.find(signame); \
  if( sigkey == signalMap.end() ) /* key does NOT exist */ \
    { \
      throw sotExceptionFactory( sotExceptionFactory::UNREFERED_SIGNAL,\
				 "The requested signal is not registered",\
				 ": %s",signame.c_str() );\
    }\
  return *(sigkey ->second) ;



sotSignalAbstract<int>& sotEntity::
getSignal( const string & signame )
{
  __SOT_ENTITY_GET_SIGNAL__(iterator);
}

const sotSignalAbstract<int>& sotEntity::
getSignal( const string & signame ) const
{
  __SOT_ENTITY_GET_SIGNAL__(const_iterator);
}
  

std::ostream& sotEntity::
displaySignalList( std::ostream& os ) const
{
  os << "--- <" << getName() << "> signal list: "<<endl;
  const SignalMap::const_iterator iterend=signalMap.end();
  for( SignalMap::const_iterator iter = signalMap.begin();iterend!=iter;++iter )
    {
      os << "    "; if( (++iter)--==iterend ) os << "`"; else os <<"|";
      os << "-- <" << *(iter->second) << endl;
    }
  return os;
}

std::ostream& sotEntity::
writeGraph( std::ostream& os ) const
{
  const SignalMap::const_iterator iterend=signalMap.end();
  for( SignalMap::const_iterator iter = signalMap.begin();iterend!=iter;++iter )
    {
      
      (*(iter->second)).writeGraph(os);
    }
  return os;
}

std::ostream& sotEntity::
writeCompletionList( std::ostream& os ) const
{
  const SignalMap::const_iterator iterend=signalMap.end();
  for( SignalMap::const_iterator iter = signalMap.begin();iterend!=iter;++iter )
    {
      
      os << getName() << "." << (*(iter->second)).shortName() << std::endl;
    }

  os << getCommandList() << std::endl;
  return os;
}

void sotEntity::
display( std::ostream& os ) const
{
  os<<this->getClassName()<<": "<<name; 
}

std::ostream& operator<< (std::ostream& os, const sotEntity& ent )
{
  ent.display(os);
  return os;
}

/* --- PARAMS --------------------------------------------------------------- */
/* --- PARAMS --------------------------------------------------------------- */
/* --- PARAMS --------------------------------------------------------------- */

static std::string SOTENTITY_COMMAND_LIST = "print\nsignals\nsignalDep";
const std::string& sotEntity::
getCommandList( void ) const
{
  return SOTENTITY_COMMAND_LIST; 
}

void sotEntity::
commandLine( const std::string& cmdLine,std::istringstream& cmdArgs,std::ostream& os )
{
  if( cmdLine == "help" )
    {
      os << "Entity : " << std::endl
	 << "  - print\t\t\tWhat d'you think?"<<endl
	 << "  - signals\t\t\tDisplay the signals list."<<endl
	 << "  - signalDep <signame> \tDisplay the dependancy graph for signal signame."<<endl;
    }
  else if( cmdLine == "print")
    {
      os << *this << std::endl;
    }
  else if( cmdLine == "signals")
    {      displaySignalList(os);    }
  else if( cmdLine == "signalDep")
    {    
      string sig; cmdArgs>>sig; 
      cmdArgs >>  ws; int depth=-1; 
      if( cmdArgs.good() ) { cmdArgs >> depth; }
      getSignal(sig) .displayDependancies( os,depth ); os<<endl;
    }
  else 
    {
      try{
	sotSignalAbstract<int> & sig = getSignal( cmdLine );

	int time; cmdArgs>>ws;
	if( cmdArgs.good() )
	  {cmdArgs >> time;} else {time=0;}
	sig.recompute( time );
      
	os << cmdLine << " = "; sig.get( os );
      } catch( sotExceptionFactory& e ) {
	switch( e.getCode() )
	  {
	  case sotExceptionFactory::UNREFERED_SIGNAL: 
	    SOT_THROW sotExceptionFactory( sotExceptionFactory::UNREFERED_FUNCTION,
					   "The requested function/signal :","<%s> is "
					   "not registered.",cmdLine.c_str() );
	    break;
	  default:
	    throw; 
	  }
      } catch( ... ) { throw;  }
    }
  
}
