/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotInterpretor.cpp
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* SOT */
#include <sot/sotInterpretor.h>
#include <sot/sotPluginLoader.h>
#include <sot/sotDebug.h>

/* --- STD --- */
using namespace std;

/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */

const std::string sotInterpretor::PROMPT_DEFAULT = "> ";

sotInterpretor::
sotInterpretor( sotPluginLoader* dl__ )
  : dlPtr(dl__),initDone(false)//,prompt( PROMPT_DEFAULT )
{
  registerFunction("plug",boost::bind(&sotInterpretor::cmdPlug,this,_1,_2,_3));
  registerFunction("new",boost::bind(&sotInterpretor::cmdNew,this,_1,_2,_3));
  registerFunction("destroy",boost::bind(&sotInterpretor::cmdDestroy,this,_1,_2,_3));
  registerFunction("run",boost::bind(&sotInterpretor::cmdRun,this,_1,_2,_3));
  registerFunction("loadPlugin",boost::bind(&sotInterpretor::cmdLoadPlugin,this,_1,_2,_3));
  registerFunction("unloadPlugin",boost::bind(&sotInterpretor::cmdUnloadPlugin,this,_1,_2,_3));
  registerFunction("help",boost::bind(&sotInterpretor::cmdHelp,this,_1,_2,_3));
  registerFunction("set",boost::bind(&sotInterpretor::cmdSetSignal,this,_1,_2,_3));
  registerFunction("get",boost::bind(&sotInterpretor::cmdGetSignal,this,_1,_2,_3));
  registerFunction("compute",boost::bind(&sotInterpretor::cmdComputeSignal,this,_1,_2,_3));
  prompt = PROMPT_DEFAULT;
  initDone = true;
}

void sotInterpretor::
registerFunction( const string& funname, 
		  const sotInterpretor::ShellBasicFunction& fun )
{
  if( initDone ) {sotDEBUG(15) << "Register " << funname << std::endl;}
  FunctionMap::iterator funkey = functionMap.find(funname);
  if( funkey != functionMap.end() ) // key does exist
    {
      if( initDone )
	{
	  sotDEBUG(15) << "!! Another function already defined with the same name. "
		       << "Funname is" <<funname.c_str() << endl;
	}
      throw sotExceptionFactory( sotExceptionFactory::FUNCTION_CONFLICT,
				 "Another function already defined with the same name. ",
				 "Funname is <%s>.",funname.c_str() );
    }
  else
    {
//       sotDEBUG(10) << "Register function <"<< funname 
// 		   << "> in the shell interpretor." <<endl;
      functionMap[funname] = fun;
    }
}



bool sotInterpretor::
deregisterFunction( const std::string& funname )
{
  return ( 0!=functionMap.erase(funname) );
}


/* --------------------------------------------------------------------- */
/* --- NEW ------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
#include <sot/sotFactory.h>
using namespace std;
#include <sot/sotEntity.h>
#include <sot/sotSignalAbstract.h>

void sotInterpretor::
cmdPlug( const std::string& cmdLine, istringstream& cmdArg, std::ostream& os )
{
  if( cmdLine == "help" ) 
    {
      os << "  - plug <obj1.sig1> <obj2.sig2>"
	 << "\t\tPlug sig1 (producer) on sig2 (consumer)." <<endl;
      return;
    }
  string ssig1,ssig2;
  cmdArg>>ssig1>>ssig2;
  
  string obj1,fun1;
  string obj2,fun2;
  istringstream str1(ssig1),str2(ssig2);
  if( (!objectNameParser(str1,obj1,fun1))||(!objectNameParser(str2,obj2,fun2)) )
    {
      SOT_THROW sotExceptionFactory( sotExceptionFactory::SYNTAX_ERROR,
				     "Plug function: syntax is plug OBJ1.SIG1 OBJ2.SIG2.",
				     "(while calling plug %s %s).",ssig1.c_str(),
				     ssig2.c_str() );
    }

  sotDEBUG(20) << "Get Ent1 <"<<obj1<<"> ."<<endl;
  sotEntity& ent1 = sotPool.getEntity(obj1);
  sotDEBUG(20) << "Get Sig1 <"<<fun1<<"> ."<<endl;
  sotSignalAbstract<int> &sig1 = ent1.getSignal(fun1);

  sotDEBUG(20) << "Get Ent2 <"<<obj2<<"> ."<<endl;
  sotEntity& ent2 = sotPool.getEntity(obj2);
  sotDEBUG(20) << "Get Sig2 <"<<fun2<<"> ."<<endl;
  sotSignalAbstract<int> &sig2 = ent2.getSignal(fun2);

  sotDEBUG(25) << "Plug..."<<endl;
  sig2.plug(&sig1);
}

void sotInterpretor::
cmdNew( const std::string& cmdLine, istringstream& cmdArg, std::ostream& os )
{
  if( cmdLine == "help" ) 
    {
      os << "  - new <class> <object>"
	 << "\t\t\tCreate a new object." <<endl;
      return;
    }
  string className;
  string objName;
  cmdArg >> className >>objName;
  sotDEBUG(15) << "New <" << className<<"> requested."<<endl;
  if( sotFactory.existEntity( className ) ) 
    {
      sotDEBUG(15) << "New entity<"<<className<<"> " <<objName<<std::endl;
      sotFactory.newEntity(className,objName);
    }
  else if( sotFactory.existFeature( className ) ) 
    {
      sotDEBUG(15) << "New feature<"<<className<<"> " <<objName<<std::endl;
      sotFactory.newFeature(className,objName);
    }
  else if( sotFactory.existTask( className ) ) 
    {
      sotDEBUG(15) << "New Task<"<<className<<"> " <<objName<<std::endl;
     sotFactory.newTask(className,objName);
    }
  else os << "  !! Class <" << className << "> does not exist."<<endl;
}
 

void sotInterpretor::
cmdDestroy( const std::string& cmdLine, istringstream& cmdArg, std::ostream& os )
{
  if( cmdLine == "help" ) 
    {
      os << "  - destroy <object>"
	 << "\t\t\tDestroy an object." <<endl;
      return;
    }
  string objName;  cmdArg >> objName;
  sotDEBUG(15) << "Destroy <" << objName <<"> requested."<<endl;
  delete &( sotPool.getEntity( objName ) );

}
 


void sotInterpretor::
cmdLoadPlugin( const std::string& cmdLine, std::istringstream& cmdArg, std::ostream& os )
{
  if( cmdLine == "help" ) 
    {
      os << "  - loadPlugin <file.so> <directory>"
	 << "\t\tLoad the plugin from the specified directory." <<endl;
      return;
    }
  if( NULL!=dlPtr )
    {
     string pluginName,directory;
      cmdArg >> pluginName;
      cmdArg >> directory;
      if( directory.length() != 0 ) dlPtr->setDirectory( directory );
      dlPtr ->addPlugin( pluginName );
     
      try{
	sotDEBUG(15) << "Try to load  " << pluginName<< endl; 
	sotDEBUG(25)<<"sotShell.dlPtr ="<< this->dlPtr <<endl; 
	dlPtr->loadPlugins();
	sotDEBUG(25)<<"sotShell.dlPtr ="<< this->dlPtr <<endl; 
      }catch( sotExceptionAbstract& e ) { sotDEBUG(5) << e << endl; throw e; }
    }
  else { os << "!!  Dynamic loading functionalities not accessible through the shell." <<endl; }
}

void sotInterpretor::
cmdUnloadPlugin( const std::string& cmdLine, std::istringstream& cmdArg,
		 std::ostream& os )
{
  if( cmdLine == "help" ) 
    {
      os << "  - unloadPlugin <path/file.so>"
	 << "\t\tUnload the plugin." <<endl;
      return;
    }

  sotDEBUGIN(15);
  if( NULL!=dlPtr )
    {
      string pluginName;
      cmdArg >> pluginName;
      try{
	sotDEBUG(25) << "Try short name " << pluginName << endl;
	const std::string& fullname = dlPtr->searchPlugin(pluginName);
	sotDEBUG(25) << "Full name " << fullname << endl;
	dlPtr->unloadPlugin(fullname);
      }
      catch(...) 
	{
	  sotDEBUG(25) << "Full name " << pluginName << endl;
	  dlPtr->unloadPlugin(pluginName);
	}
    }
  else { os << "!!  Dynamic loading functionalities not accessible through the shell." <<endl; }

  sotDEBUGOUT(15);
}


void sotInterpretor::
cmdHelp( const std::string& cmdLine, std::istringstream& cmdArg, std::ostream& os )
{
  sotDEBUGIN(25);
  os << "Help" <<endl;

  std::string procname; bool personalizedHelp = false;
  cmdArg >> ws; 
  if( cmdArg.good() ) 
    {
      const unsigned int gc = cmdArg.tellg(); 
      cmdArg >> procname; 
      cmdArg.seekg(gc); cmdArg.clear();
      personalizedHelp = true; 
      sotDEBUG(15)<< "Personalized help on <"<< procname<<">"<<gc<<endl;
    }

  bool procfund = !personalizedHelp;
  for( FunctionMap::const_iterator iter=functionMap.begin();
       iter!=functionMap.end();++iter )
    {
      if( iter->first!="help" ) 
	if( (! personalizedHelp) || ( procname==iter->first ) )
	  {  iter->second(cmdLine,cmdArg,os); procfund=true;  }
    }
  if(! procfund ) { os<<" *** Procedure <"<<procname<<"> not found."<<endl; }
  sotDEBUGOUT(25);
}


void sotInterpretor::
cmdRun( const std::string& cmdLine, std::istringstream& cmdArg, std::ostream& os )
{
  if( cmdLine == "help" )
    {
      os << "  - run <script.txt>\t\t\t\tRun the script." <<endl;  
      return;
    }
     
  sotDEBUGIN(15);

  string filename; cmdArg>>filename;
  sotDEBUG(25) << "Script <" <<filename<<">"<<endl;

  ifstream script( filename.c_str(),ios::in );
  if(! script.is_open() )
    {
      SOT_THROW sotExceptionFactory( sotExceptionFactory::READ_FILE,
				     "File is not open."," (while reading <%s>).",
				     filename.c_str() );
    }

  const int SIZE = 16384;
  char line[SIZE];int lineIdx;
  string name; 
  try{
    for( lineIdx=1;;lineIdx++ )
      {
	sotDEBUGIN(15);
	
	script >> name;
	script.getline(line,SIZE);
	if(! script.good() ) break;
	
	sotDEBUG(25) << "Run <"<<name<<"> with args <"<<line<<">"<<endl;
	istringstream iss(line);
	cmd( name,iss,os );
	
	sotDEBUGOUT(15);
      }
  } catch( sotExceptionAbstract& exc ) {
    std::string& msg = (std::string&)exc.getStringMessage();
    std::ostringstream oss; 
    oss <<" (in line " << lineIdx <<" of file <" << filename << ">)";
    msg = msg + oss.str();
    throw exc;
  }

  sotDEBUGOUT(15);

}


void sotInterpretor::
cmdSetSignal( const std::string& cmdLine, std::istringstream& cmdArg, std::ostream& os )
{
  if( cmdLine == "help" )
    {
      os << "  - set <obj.signal> <value>\t\t\tSet the signal to the constant value." <<endl;  
      return;
    }
     
  sotDEBUGIN(15);

  string objname,signame; 
  objectNameParser(cmdArg,objname,signame);
  sotEntity& obj = sotPool.getEntity(objname);
  sotSignalAbstract<int>& sig = obj.getSignal( signame );

  cmdArg >> ws;
  sig.set( cmdArg );

  sotDEBUGOUT(15);

}

void sotInterpretor::
cmdGetSignal( const std::string& cmdLine, std::istringstream& cmdArg, std::ostream& os )
{
  if( cmdLine == "help" )
    {
      os << "  - get <obj.signal> <value>\t\t\tGet the signal to the constant value." <<endl;  
      return;
    }
     
  sotDEBUGIN(15);

  string objname,signame;
  objectNameParser(cmdArg,objname,signame);
  sotEntity& obj = sotPool.getEntity(objname);
  sotSignalAbstract<int>& sig = obj.getSignal( signame );

  os << signame << " = "; sig.get( os );

  sotDEBUGOUT(15);

}

void sotInterpretor::
cmdComputeSignal( const std::string& cmdLine, std::istringstream& cmdArg, std::ostream& os )
{
  if( cmdLine == "help" )
    {
      os << "  - compute <obj.sig> <time>\t\t\tRecompute <sig> at time <time>  " <<endl;  
      return;
    }
     
  sotDEBUGIN(15);
  
  string objname,signame;
  objectNameParser(cmdArg,objname,signame);
  sotEntity& obj = sotPool.getEntity(objname);
  sotSignalAbstract<int>& sig = obj.getSignal( signame );

  int time; cmdArg >> std::ws;
  if( cmdArg.good() )
    {cmdArg >> time;} else {time=0;}
  sig.recompute( time );

  sotDEBUGOUT(15);

}


/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */


void sotInterpretor::
cmd( const std::string& cmdLine, istringstream& cmdArg, std::ostream& os )
{
  istringstream cmdparse(cmdLine.c_str());
  string obj,fun;
  if (cmdLine.find_first_not_of(" ") == string::npos)
  {}
  else if( objectNameParser( cmdparse,obj,fun ) )
    {
   sotDEBUG(15) << "Object <" << obj<< "> function <"<<fun<<">"<<endl;
   sotPool.commandLine( obj,fun,cmdArg,os );
  }
  else 
    {
     sotDEBUG(15) << "Function <" << cmdLine <<">"<< endl;
      FunctionMap::iterator funPtr = functionMap .find( cmdLine );
      if( funPtr == functionMap.end() )
	{
	  SOT_THROW sotExceptionFactory( sotExceptionFactory::UNREFERED_FUNCTION,
					 "Unknown function."," (while calling <%s>)",
					 cmdLine.c_str() );
	}
      funPtr->second(cmdLine,cmdArg,os);
    }

}



bool sotInterpretor::
objectNameParser( istringstream& cmdparse,std::string& objName,std::string& funName  )
{
  const int SIZE=128;
  char buffer[SIZE];
  cmdparse >> ws;
  cmdparse.getline( buffer,SIZE,'.' );
  if(! cmdparse.good() ) // The callback is not an object method
    return false;

  objName = buffer;
  //cmdparse.getline( buffer,SIZE );
  //funName = buffer;
  cmdparse >> funName;
  return true;
}




void sotInterpretor::
shell( std::istream& sin, std::ostream& sout, const std::string& promptUser )
{
  while( 1 )
    {
      if( promptUser.length() ) sout << promptUser; else sout << prompt;
      string cmdLine;
      const int SIZE = 16384;
      char cmdArgs[SIZE];
      sin >>skipws>> cmdLine ;
      sotDEBUG(15) << "Cmd <" <<cmdLine<<">"<<endl;
      if( cmdLine == "exit" ) break;
      if( sin.eof() ) break;

      sin.getline( cmdArgs,SIZE-1 );

      if( sin.gcount() >= SIZE-2 ) 
	{
	  sout << "!! Line size exceeded" << endl; 
	  do{ sin.getline( cmdArgs,SIZE-1 ); } while ( sin.gcount() >= SIZE-2 );
	  sout << cmdArgs << endl; 
	} 
      else
	{
	  sotDEBUG(15) << "Args <" <<cmdArgs<<">"<<endl;
      
	  istringstream args (cmdArgs);

	  try{ cmd(cmdLine,args,sout); }
	  catch( sotExceptionAbstract& e ) { sotDEBUG(1) << e; throw e; }
	  catch(...) { sotDEBUG(1) << "!! unknow!." <<endl; throw; }
	}
    }
}

sotShellFunctionRegisterer::
sotShellFunctionRegisterer( const std::string& funName,
			    const sotInterpretor::ShellBasicFunction& f)
{
  sotDEBUGIN(25);
  sotShell.registerFunction(funName,f);
  sotDEBUGOUT(25);
}


void sotInterpretor::writeCompletionList(std::ostream& os)
{
  for( FunctionMap::iterator iter=functionMap.begin();
       iter!=functionMap.end();iter++ )
    {
      const std::string & name = iter->first;
      os << name << std::endl;
    }



}


sotInterpretor sotShell;
