/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotPluginLoader.cpp
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* --- SOT --- */
#include <sot/sotPluginLoader.h>
#include <sot/sotDebug.h>

/* --- STD --- */
#include <fstream>
using namespace std;

#ifdef WIN32
#include <Windows.h>
#else 
#include <dlfcn.h> 
#endif

/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */


class sotPluginRefMap
{
public:

#ifdef WIN32
	typedef HMODULE sotPluginKey;
#else
	typedef void* sotPluginKey;
#endif

	typedef std::map< std::string,sotPluginKey> KeyMap;
	KeyMap keyMap;
};


sotPluginLoader::
sotPluginLoader( void )
{
  pluginRefs = new sotPluginRefMap();
}

sotPluginLoader::
~sotPluginLoader( void )
{
  delete pluginRefs;
}

const std::string& sotPluginLoader::
setDirectory( const std::string& n )
{ 
  return pluginDirectory = n; 
}
const std::string& sotPluginLoader::
getDirectory( void )
{
  return pluginDirectory; 
}

void sotPluginLoader::
loadPluginList( const std::string& configFileName, const std::string& dir )
{
  sotDEBUGIN(15);

  
  string name;
  //while (1)
  ifstream configFile( configFileName.c_str(),ios::in );
  if(! configFile.is_open())
    {
      SOT_THROW sotExceptionFactory( sotExceptionFactory::READ_FILE,
				     "File is not open."," (while reading <%s>).",
				     configFileName.c_str() );
    }

  for(;;)
    {
      configFile>>name;
      if(configFile.eof()) break;
      sotDEBUG(9)<<"Add <"<< name << "> to the list"<<endl; 
      addPlugin( name,dir);
    }

  sotDEBUGOUT(15);
}
 
void sotPluginLoader::
addPlugin( const std::string& name, const std::string& dir )
{
  sotDEBUGIN(15);
  if( dir.length() )
    pluginNames.push_back(  dir +"/"+name );
  else
    {      pluginNames.push_back(  pluginDirectory +"/"+name );    }
  sotDEBUGOUT(15);
}





/* -------------------------------------------------------------------------- */
/* --- DYNAMIC LOADER LIB --------------------------------------------------- */
/* -------------------------------------------------------------------------- */
void sotPluginLoader::
loadPlugins( void )
{
  sotDEBUGIN(15);

  for( list<string>::iterator iter = pluginNames.begin();
       iter!=pluginNames.end();++iter )
    {
      sotDEBUG(9)<<"Load <"<< *iter << "> plugin"<<endl;
#ifndef WIN32
	  sotPluginRefMap::sotPluginKey dlib = dlopen( iter->c_str(),RTLD_NOW|RTLD_GLOBAL); 
#else
	  sotPluginRefMap::sotPluginKey dlib = LoadLibrary ( iter->c_str()); 
#endif

      sotDEBUG(19)<<"Plugin <"<< *iter << "> loaded "<<endl; 

      if( NULL==dlib ) 
	{
	  pluginNames.erase(iter);
#ifndef WIN32
	  sotDEBUG(5) << "Failure while loading: " <<dlerror() <<endl;
	  SOT_THROW sotExceptionFactory( sotExceptionFactory::DYNAMIC_LOADING,
	  				 "Error while dlopen. ","<%s>",dlerror() );
#else
    // Retrieve the system error message for the last-error code
    LPTSTR pszMessage;
    DWORD dwLastError = GetLastError(); 
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        dwLastError,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR)&pszMessage,
        0, NULL );


	  sotDEBUG(5) << "Failure while loading: " << pszMessage <<endl;


	  SOT_THROW sotExceptionFactory( sotExceptionFactory::DYNAMIC_LOADING,
	  				 "Error while dlopen. ", pszMessage );

	  LocalFree(pszMessage);

#endif
	}
      
      loadedPluginNames[*iter] = (*iter);
	  pluginRefs->keyMap[*iter] = dlib;
    }
  pluginNames.clear();

  sotDEBUGOUT(15);
}

/* -------------------------------------------------------------------------- */
/* --- DYNAMIC UN-LOADER LIB ------------------------------------------------ */
/* -------------------------------------------------------------------------- */


void sotPluginLoader::
unloadPlugin( const std::string& plugname )
{
  sotDEBUGIN( 15 );

  sotPluginRefMap::KeyMap::iterator plugkey = pluginRefs->keyMap.find(plugname);
  if( plugkey==pluginRefs->keyMap.end() ) // key does exist
    {
      throw sotExceptionFactory( sotExceptionFactory::OBJECT_CONFLICT,
				 "Plugin not loaded",": <%s>.",plugname.c_str() );
    }

#ifndef WIN32
  const int errCode = dlclose(plugkey->second);
#else
  const int errCode = FreeLibrary(plugkey->second);
#endif
  if( errCode )
    {
#ifndef WIN32
      sotDEBUG(1) << "Error while unloading <" << plugname <<"> : " 
		  << dlerror() <<endl;
#else
		    // Retrieve the system error message for the last-error code
    LPTSTR pszMessage;
    DWORD dwLastError = GetLastError(); 
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        dwLastError,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR)&pszMessage,
        0, NULL );

		sotDEBUG(1) << "Error while unloading <" << plugname <<"> : " 
		  << pszMessage <<endl;

	  LocalFree(pszMessage);
#endif
    }

  //pluginRefs.erase( plugkey );

  sotDEBUGOUT( 15 );
}

const std::string& sotPluginLoader::
searchPlugin( const std::string& plugname )
{
  unsigned int refFound = 0;
  const std::string *plugFullName =0;
  for( sotPluginRefMap::KeyMap::iterator iter = pluginRefs->keyMap.begin();
       iter!=pluginRefs->keyMap.end();++iter )
    {
      const std::string &str = iter->first;
      size_t found = str.find_last_of("/\\");
      sotDEBUG(15) << " folder: " << str.substr(0,found) << std::endl;
      sotDEBUG(15) << " file: " << str.substr(found+1) << std::endl;
      if( str.substr(found+1)==plugname )
	{
	  refFound++;
	  plugFullName=&iter->first;
	}
    }

  if( 1!=refFound ) throw 1;

  return *plugFullName;
}
