/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotPool.cpp
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* --- SOT --- */
#include <sot/sotPool.h>
#include <sot/sotDebug.h>
#include <sot/sotEntity.h>
#include <list>

/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */

sotPoolStorage::
~sotPoolStorage( void )
{
  sotDEBUGIN(15);

  for( Entities::iterator iter=entity.begin();iter!=entity.end(); iter=entity.begin())
    {
      sotDEBUG(15) << "Delete \"" 
		   << (iter->first) <<"\""<<std::endl;
      delete ((sotEntity*)iter->second);
    }
  
  sotDEBUGOUT(15);
  return;
}



/* --------------------------------------------------------------------- */
void sotPoolStorage::
registerEntity( const std::string& entname,sotEntity* ent )
{
  Entities::iterator entkey = entity.find(entname);
  if( entkey != entity.end() ) // key does exist
    {
      throw sotExceptionFactory( sotExceptionFactory::OBJECT_CONFLICT,
				 "Another entity already defined with the same name. ",
				 "Entity name is <%s>.",entname.c_str() );
    }
  else
    {
      sotDEBUG(10) << "Register entity <"<< entname 
		   << "> in the pool." <<std::endl;
      entity[entname] = ent;
    }
}

void sotPoolStorage::
deregisterEntity( const std::string& entname )
{
  Entities::iterator entkey = entity.find(entname);
  if( entkey == entity.end() ) // key doesnot exist
    {
      throw sotExceptionFactory( sotExceptionFactory::OBJECT_CONFLICT,
				 "Entity not defined yet. ",
				 "Entity name is <%s>.",entname.c_str() );
    }
  else
    {
      sotDEBUG(10) << "Deregister entity <"<< entname 
		   << "> from the pool." <<std::endl;
      entity.erase( entkey );
    }
}

sotEntity& sotPoolStorage::
getEntity( const std::string& name )
{
  sotDEBUG(25) << "Get <" << name << ">"<<std::endl;
  Entities::iterator entPtr = entity .find( name );
  if( entPtr == entity.end() )
    {
      SOT_THROW sotExceptionFactory( sotExceptionFactory::UNREFERED_OBJECT,
				     "Unknown entity."," (while calling <%s>)",
				     name.c_str() );
    }
  else return *entPtr->second;
}

void sotPoolStorage::
clearPlugin( const std::string& name )
{
  sotDEBUGIN(5);
  std::list< sotEntity* > toDelete;
  
  for( Entities::iterator entPtr=entity.begin(); entPtr!=entity.end(); entPtr++ )
    {
      if( entPtr->second->getClassName() == name )
	{ toDelete.push_back( entPtr->second ); }
    }

  for( std::list< sotEntity* >::iterator iter=toDelete.begin(); 
       iter!=toDelete.end(); ++iter )
    {
      delete (sotEntity*) *iter;
    }
  sotDEBUGOUT(5);
  return;
}


/* --------------------------------------------------------------------- */
void sotPoolStorage::
registerTask( const std::string& entname,sotTaskAbstract* ent )
{
  Tasks::iterator entkey = task.find(entname);
  if( entkey != task.end() ) // key does exist
    {
      throw sotExceptionFactory( sotExceptionFactory::OBJECT_CONFLICT,
				 "Another task already defined with the same name. ",
				 "Task name is <%s>.",entname.c_str() );
    }
  else
    {
      sotDEBUG(10) << "Register task <"<< entname 
		   << "> in the pool." <<std::endl;
      task[entname] = ent;
    }
}

sotTaskAbstract& sotPoolStorage::
getTask( const std::string& name )
{
  Tasks::iterator entPtr = task .find( name );
  if( entPtr == task.end() )
    {
      SOT_THROW sotExceptionFactory( sotExceptionFactory::UNREFERED_OBJECT,
				     "Unknown task."," (while calling <%s>)",
				     name.c_str() );
    }
  return *entPtr->second;
}



/* --------------------------------------------------------------------- */
void sotPoolStorage::
registerFeature( const std::string& entname,sotFeatureAbstract* ent )
{
  Features::iterator entkey = feature.find(entname);
  if( entkey != feature.end() ) // key does exist
    {
      throw sotExceptionFactory( sotExceptionFactory::OBJECT_CONFLICT,
				 "Another feature already defined with the same name. ",
				 "Feature name is <%s>.",entname.c_str() );
    }
  else
    {
      sotDEBUG(10) << "Register feature <"<< entname 
		   << "> in the pool." <<std::endl;
      feature[entname] = ent;
    }
}

sotFeatureAbstract& sotPoolStorage::
getFeature( const std::string& name )
{
  Features::iterator entPtr = feature .find( name );
  if( entPtr == feature.end() )
    {
      SOT_THROW sotExceptionFactory( sotExceptionFactory::UNREFERED_OBJECT,
				     "Unknown feature."," (while calling <%s>)",
				     name.c_str() );
    }
  return *entPtr->second;
}




/* --------------------------------------------------------------------- */

#include <sot/sotEntity.h>
#include <sot/sotTaskAbstract.h>
#include <sot/sotFeatureAbstract.h>

#ifdef WIN32
#include <time.h>
#endif /*WIN32*/

void sotPoolStorage::
writeGraph(const std::string &aFileName)
{
  size_t IdxPointFound = aFileName.rfind(".");
  std::string tmp1 = aFileName.substr(0,IdxPointFound);
  size_t IdxSeparatorFound = aFileName.rfind("/");
  std::string GenericName;
  if (IdxSeparatorFound!=std::string::npos)
    GenericName = tmp1.substr(IdxSeparatorFound,tmp1.length());
  else 
    GenericName = tmp1;

  /* Reading local time */
  time_t ltime;
  ltime = time(NULL);
  struct tm ltimeformatted;
#ifdef WIN32
  localtime_s(&ltimeformatted,&ltime);
#else
  localtime_r(&ltime,&ltimeformatted);
#endif /*WIN32*/
  
  /* Opening the file and writing the first comment. */
  std::ofstream GraphFile;
  GraphFile.open((char *)aFileName.c_str(),std::ofstream::out);
  GraphFile << "/* This graph has been automatically generated. " << std::endl;
  GraphFile << "   " << 1900+ltimeformatted.tm_year 
	    << " Month: " << 1+ltimeformatted.tm_mon
	    << " Day: " << ltimeformatted.tm_mday
	    << " Time: " << ltimeformatted.tm_hour
	    << ":" << ltimeformatted.tm_min;
  GraphFile << " */" << std::endl;
  GraphFile << "digraph " << GenericName << " { ";
  GraphFile << "\t graph [ label=\"" << GenericName << "\" bgcolor = white rankdir=LR ]" << std::endl
	    << "\t node [ fontcolor = black, color = black, fillcolor = gold1, style=filled, shape=box ] ; " << std::endl;
  GraphFile << "\tsubgraph cluster_Tasks { " << std::endl;
  GraphFile << "\t\t color=blue; label=\"Tasks\";" << std::endl;
  
  for( Tasks::iterator iter=task.begin();
       iter!=task.end();iter++ )
    {
      sotTaskAbstract* ent = iter->second;
      GraphFile << "\t\t" << ent->getName() 
		<<" [ label = \"" << ent->getName() << "\" ," << std::endl 
		<<"\t\t   fontcolor = black, color = black, fillcolor = magenta, style=filled, shape=box ]" << std::endl; 
	  
    }
  
  GraphFile << "\t} " << std::endl;

  for( Entities::iterator iter=entity.begin();
       iter!=entity.end();iter++ )
    {
      sotEntity* ent = iter->second;
      /* GraphFile << ent->getName() 
		<<" [ label = \"" << ent->getName() << "\" ," << std::endl 
		<<"   fontcolor = black, color = black, fillcolor=cyan, style=filled, shape=box ]" << std::endl; */
      ent->writeGraph(GraphFile);
    }

  
  GraphFile << "}"<< std::endl;
  
  GraphFile.close();
}

void sotPoolStorage::
writeCompletionList(std::ostream& os)
{
  for( Entities::iterator iter=entity.begin();
       iter!=entity.end();iter++ )
    {
      sotEntity* ent = iter->second;
      ent->writeCompletionList(os);
    }


}


void sotPoolStorage::
commandLine( const std::string& objectName,const std::string& functionName,
	     std::istringstream& cmdArg, std::ostream& os )
{
  sotDEBUG(15) << "Object <" << objectName<< "> function <"
	       <<functionName<<">"<<std::endl;

  if( objectName=="pool" )
    {
      if( functionName=="help" )
	{
	  os <<"Pool: " << std::endl
	     <<"  - list" << std::endl
	     <<"  - listFeature" << std::endl
	     <<"  - listTask" << std::endl
	     << " - writegraph FileName" << std::endl;
	}
      else if( functionName=="list" )
	{
	  for( Entities::iterator iter=entity.begin();
	       iter!=entity.end();iter++ )
	    {
	      sotEntity* ent = iter->second;
	      os << ent->getName() 
		 <<" (" << ent->getClassName() << ")" << std::endl;
	    }

	}
      else if( functionName=="listTask" )
	{
	  for( Tasks::iterator iter=task.begin();
	       iter!=task.end();iter++ )
	    {
	      sotTaskAbstract* ent = iter->second;
	      os << ent->getName() 
		 <<" (" << ent->getClassName() << ")" << std::endl;
	    }

	}
      else if( functionName=="listFeature" )
	{
	  for( Features::iterator iter=feature.begin();
	       iter!=feature.end();iter++ )
	    {
	      sotFeatureAbstract* ent = iter->second;
	      os << ent->getName() 
		 <<" (" << ent->getClassName() << ")" << std::endl;
	    }

	}
      else if (functionName=="writegraph")
	{
	  std::string aFileName;
	  cmdArg >> aFileName;
	  writeGraph(aFileName);
	}

    }
  else
    {
      sotEntity& ent = getEntity(objectName);
      ent.commandLine(functionName,cmdArg,os);
    }
}



#include <sot/sotInterpretor.h>

sotSignalAbstract<int>& 
sotPoolStorage::
getSignal( std::istringstream& sigpath )
{
  std::string objname,signame;
  if(! sotInterpretor::objectNameParser( sigpath,objname,signame ) )
    { SOT_THROW sotExceptionFactory( sotExceptionFactory::UNREFERED_SIGNAL,
				     "Parse error in signal name" ); }

  sotEntity& ent = getEntity( objname );
  return ent.getSignal( signame );
}


sotPoolStorage sotPool;

