/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotFeaturePoint6d.cpp
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* --- SOT --- */
#include <sot/sotDebug.h>
#include <sot/sotFeaturePoint6d.h>
#include <sot/sotExceptionFeature.h>

#include <sot/sotMatrixHomogeneous.h>
#include <sot/sotMatrixRotation.h>
#include <sot/sotVectorUTheta.h>

using namespace std;

#include <sot/sotFactory.h>
SOT_FACTORY_FEATURE_PLUGIN(sotFeaturePoint6d,"FeaturePoint6d");

/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */

      
const sotFeaturePoint6d::ComputationFrameType sotFeaturePoint6d::
COMPUTATION_FRAME_DEFAULT = FRAME_DESIRED;

sotFeaturePoint6d::
sotFeaturePoint6d( const string& pointName )
  : sotFeatureAbstract( pointName )
    ,computationFrame( COMPUTATION_FRAME_DEFAULT )
    ,positionSIN( NULL,"sotFeaturePoint6d("+name+")::input(matrixHomo)::position" )
    ,articularJacobianSIN( NULL,"sotFeaturePoint6d("+name+")::input(matrix)::Jq" )
{
  jacobianSOUT.addDependancy( positionSIN );
  jacobianSOUT.addDependancy( articularJacobianSIN );

  errorSOUT.addDependancy( positionSIN );

  activationSOUT.removeDependancy( desiredValueSIN );

  signalRegistration( positionSIN<<articularJacobianSIN );
}


/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

unsigned int& sotFeaturePoint6d::
getDimension( unsigned int & dim, int time ) 
{
  sotDEBUG(25)<<"# In {"<<endl;

  const sotFlags &fl = selectionSIN.access(time);

  dim = 0;
  for( int i=0;i<6;++i ) if( fl(i) ) dim++;

  sotDEBUG(25)<<"# Out }"<<endl;
  return dim;
}


/** Compute the interaction matrix from a subset of
 * the possible features. 
 */
ml::Matrix& sotFeaturePoint6d::
computeJacobian( ml::Matrix& J,int time )
{
  sotDEBUG(15)<<"# In {"<<endl;

  const ml::Matrix & Jq = articularJacobianSIN(time);
  const int & dim = dimensionSOUT(time);
  const sotFlags &fl = selectionSIN(time);

  sotDEBUG(25)<<"dim = "<<dimensionSOUT(time)<<" time:" << time << " "
              << dimensionSOUT.getTime() << " " << dimensionSOUT.getReady() << endl;
  sotDEBUG(25)<<"selec = "<<selectionSIN(time)<<" time:" << time << " "
              << selectionSIN.getTime() << " " << selectionSIN.getReady() << endl;

  sotDEBUG(15) <<"Dimension="<<dim<<std::endl;

  const unsigned int cJ = Jq.nbCols();
  J.resize(dim,cJ) ;
  ml::Matrix LJq(6,cJ);
  
  if( FRAME_CURRENT==computationFrame )
    {
      ml::Matrix Lx(6,6);  
      for(unsigned int i=0;i<6;i++) 
	{for(unsigned int j=0;j<6;j++)
	  if (i==j) Lx(i,j) = -1; else Lx(i,j) = 0;}
      
      const ml::Vector & err = errorSOUT(time);
      const double & X=err(0), &Y=err(1), &Z=err(2);
      
      Lx(0,4) = -Z ;
      Lx(0,5) = Y ;
      Lx(1,3) = Z ;
      Lx(1,5) = -X ;
      Lx(2,3) = -Y ;
      Lx(2,4) = X ;
      sotDEBUG(15) << "Lx= "<<Lx<<endl;
      
      Lx.multiply(Jq,LJq);
    } else { LJq = Jq; }

  
  unsigned int rJ = 0;
  for( unsigned int r=0;r<6;++r )
    if( fl(r) )
      {
	for( unsigned int c=0;c<cJ;++c )
	  J(rJ,c)=LJq(r,c);
	rJ ++;
      }

  sotDEBUG(15)<<"# Out }"<<endl;
  return J;
}

#define SOT_COMPUTE_H1MH2(wMh,wMhd,hMhd) {                 \
	sotMatrixHomogeneous hMw; wMh.inverse(hMw);        \
	sotDEBUG(15)<<"hMw = "<<hMw<<endl;                 \
	hMw.multiply( wMhd,hMhd );                         \
	sotDEBUG(15)<<"hMhd = "<<hMhd<<endl;               \
      }


/** Compute the error between two visual features from a subset
 * a the possible features.
 */
ml::Vector&
sotFeaturePoint6d::computeError( ml::Vector& error,int time )
{
  sotDEBUGIN(15);

  const sotFlags &fl = selectionSIN(time);
  sotFeatureAbstract * sdesAbs = desiredValueSIN(time);
  sotFeaturePoint6d * sdes = dynamic_cast<sotFeaturePoint6d*>(sdesAbs);
  
  const sotMatrixHomogeneous& wMhd = sdes->positionSIN(time);
  const sotMatrixHomogeneous& wMh = positionSIN(time);
  sotDEBUG(15)<<"wMh = "<<wMh<<endl;
  sotDEBUG(15)<<"wMhd = "<<wMhd<<endl;
  
  /* Computing only translation:                                        *
   * trans( hMw wMhd ) = htw + hRw wthd                                 *
   *                   = -hRw wth + hrW wthd                            *
   *                   = hRw ( wthd - wth )                             *
   * The second line is obtained by writting hMw as the inverse of wMh. */

  sotMatrixHomogeneous hMhd;	 
  switch(computationFrame)
    {
    case FRAME_CURRENT:
      SOT_COMPUTE_H1MH2(wMh,wMhd,hMhd); break;
    case FRAME_DESIRED:
      SOT_COMPUTE_H1MH2(wMhd,wMh,hMhd); break;
    default: break; /*TODO: error*/
    };      
  
  sotDEBUG(25)<<"dim = "<<dimensionSOUT(time)<<" time:" << time << " "
              << dimensionSOUT.getTime() << " " << dimensionSOUT.getReady() << endl;
  sotDEBUG(25)<<"selec = "<<selectionSIN(time)<<" time:" << time << " "
              << selectionSIN.getTime() << " " << selectionSIN.getReady() << endl;
  error.resize(dimensionSOUT(time)) ;

  unsigned int cursor = 0;
  for( unsigned int i=0;i<3;++i )
    { if( fl(i) ) error(cursor++) = hMhd(i,3); } 

  /* TODO: UTHETA FROM CMO */
  sotMatrixRotation hRhd; hMhd.extract( hRhd );
  sotVectorUTheta hrhd; hrhd.fromMatrix( hRhd );
  for( unsigned int i=0;i<3;++i )
    { if( fl(i+3) ) error(cursor++) = hrhd(i); }

  sotDEBUGOUT(15);
  return error ;
}

// {
//   sotDEBUGIN(15);

//   const sotFlags &fl = selectionSIN(time);
//   const sotMatrixHomogeneous& wMh = positionSIN(time);
//   sotFeatureAbstract * sdesAbs = desiredValueSIN(time);
//   sotFeaturePoint6d * sdes = dynamic_cast<sotFeaturePoint6d*>(sdesAbs);
//   const sotMatrixHomogeneous& wMhd = sdes->positionSIN(time);

// //   if( (pos.nbCols()!=4)||(pos.nbRows()!=4)||(posdes.nbCols()!=4)||(posdes.nbRows()!=4) )
//   sotDEBUG(15)<<"wMh = "<<wMh<<endl;
//   sotDEBUG(15)<<"wMhd = "<<wMhd<<endl;

//   sotMatrixHomogeneous hdMw; wMhd.inverse(hdMw); 
//   sotDEBUG(15)<<"hdMw = "<<hdMw<<endl;

//   sotMatrixHomogeneous hdMh; hdMw.multiply( wMh,hdMh );
//   sotDEBUG(15)<<"hdMh = "<<hdMh<<endl;

//   error.resize(dimensionSOUT(time)) ;
//   hdMh=hdMh.inverse();//DEBUG

//   unsigned int cursor = 0;
//   for( unsigned int i=0;i<3;++i )
//     { if( fl(i) ) error(cursor++) = -hdMh(i,3); }

//   /* TODO: UTHETA FROM CMO */
//   sotMatrixRotation hdRh; hdMh.extract( hdRh );
//   sotVectorUTheta hdrh; hdrh.fromMatrix( hdRh );
//   for( unsigned int i=0;i<3;++i )
//     { if( fl(i+3) ) error(cursor++) = -hdrh(i); }

//   sotDEBUGOUT(15);
//   return error ;
// }


/** Compute the error between two visual features from a subset
 * a the possible features.
 */
ml::Vector&
sotFeaturePoint6d::computeActivation( ml::Vector& act,int time )
{
  selectionSIN(time);
  act.resize(dimensionSOUT(time)) ; act.fill(1);
  return act ;
}


static const char * featureNames  []
= { "X ",
    "Y ",
    "Z ",
    "RX",
    "RY",
    "RZ"  };
void sotFeaturePoint6d::
display( std::ostream& os ) const
{
  os <<"Point6d <"<<name<<">: (" ;

  try{ 
    const sotFlags &fl = selectionSIN.accessCopy();
    bool first = true;
    for( int i=0;i<6;++i ) 
      if( fl(i) ) 
	{
	  if( first ) { first = false; } else { os << ","; } 
	  os << featureNames[i];
	}
    os<<") ";
  }  catch(sotExceptionAbstract e){ os<< " selectSIN not set."; }
}



void sotFeaturePoint6d::
commandLine( const std::string& cmdLine,
	     std::istringstream& cmdArgs,
	     std::ostream& os )
{
  if( cmdLine=="help" )
    {
      os << "FeaturePoint: "<<endl
	 << "  - frame [{desired|current}]: get/set the computation frame."<<endl;
      sotEntity::commandLine( cmdLine,cmdArgs,os );
    }
  else if( cmdLine=="frame" )
    {
      cmdArgs>>std::ws; 
      if(cmdArgs.good())
	{
	  std::string frameStr; cmdArgs >> frameStr;
	  if( frameStr =="current" ) { computationFrame = FRAME_CURRENT; }
	  else { computationFrame = FRAME_DESIRED; }
	}
      else {
	os << "frame = "; 
	if( FRAME_DESIRED==computationFrame ) os << "desired" << std::endl;
	else if( FRAME_CURRENT==computationFrame ) os << "current" << std::endl;
      }
    }
  else  //sotFeatureAbstract::
    sotEntity::commandLine( cmdLine,cmdArgs,os );

}



/*
 * Local variables:
 * c-basic-offset: 2
 * End:
 */
