/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotMotion.cpp
 * Project:   SOT
 * Author:    Evrard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 * Generates a reference velocity to track a trajectory in the
 * Cartesian space.
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/* LibC */
#include <cmath>

/* STL */
#include <fstream>

#include <sot/sotMotion.h>

using namespace std;

#include <sot/sotFactory.h>
SOT_FACTORY_ENTITY_PLUGIN(sotMotion,"Motion");

const double sotMotion::dt = 0.005;

sotMotion::
sotMotion( const string& fName )
  : sotEntity( fName )
  ,durationSIN( NULL,"sotMotion("+name+")::input(double)::duration" )
  ,completionSOUT( boost::bind(&sotMotion::computeCompletion,this, _1, _2),
		 sotNOSIGNAL,
		"sotMotion("+name+")::output(double)::completion" )
  ,positionSOUT( boost::bind(&sotMotion::computeOutput,this, _1, _2),
		 durationSIN<<completionSOUT,
		"sotMotion("+name+")::output(vector)::position" )

  ,xf(1)
  ,x(0)
  ,startTime(0)
  ,saveStartTime(false)
  ,started(false)
{
  signalRegistration( durationSIN<<positionSOUT<<completionSOUT );
  completionSOUT.setNeedUpdateFromAllChildren(true);
}

ml::Vector&
sotMotion::computeOutput( ml::Vector& res,int time )
{
  res.resize(3); res.fill(0);
  if(keys.size()){ res = keys[0]; }

  if(saveStartTime) {
    startTime = time;
    saveStartTime = false;
  }

  if((started == true) && (polys.size() > 0)) {
    double t = dt * (time - startTime);
    double T = durationSIN.access(time);
    double y = t / T;
    double y2 = y*y;
    double y3 = y*y2;

    if(t < T) { 
      x = xf * (10 * y*y2 - 15 * y2*y2 + 6 * y2*y3);
      int index = floor(x);
      double xt = x - index;
      res = polys[index].compute(res, xt);
    }
    else {
      x = xf;
      res = polys.back().compute(res, 1);
    }
  }

  return res;
}

double& sotMotion::computeCompletion( double& res, int time )
{
  res = x / xf;
  if( res > 1){ res = 1; }
  return res;
}

void sotMotion::display( std:: ostream& os ) const
{
  os << "LeadFollow<" << name << ">" << endl;
  os << " - started: " << started << endl;
  os << " - npolys: " << polys.size() << endl;
}

void sotMotion::commandLine( const std::string& cmdLine,
				 std::istringstream& cmdArgs,
				 std::ostream& os )
{
  if ( cmdLine=="help") {
    os << "LeadFollow:" << endl;
    os << " - start" << endl;
    os << " - stop" << endl;
    os << " - keys" << endl;
    os << " - key i" << endl;
    os << " - nkeys" << endl;
    os << " - addKey" << endl;
    os << " - clear" << endl;
    os << " - compute" << endl;
  }
  else if ( cmdLine=="start" ) {
    started = true;
    saveStartTime = true;
    xf = static_cast<double>(polys.size());
    x = 0;
  }
  else if ( cmdLine=="stop" ) {
    started = false;
  }
  else if ( cmdLine=="keys" ) {
    copy(keys.begin(), keys.end(), ostream_iterator<ml::Vector>(os, " "));
    os << endl;
  }
  else if ( cmdLine=="key" ) {
    size_t index = 0;
    if( (cmdArgs>>index) && (index<keys.size()) ){ os << keys[index] << endl; }
  }
  else if ( cmdLine=="nkeys" ) {
    os << keys.size() << endl;
  }
  else if ( cmdLine=="addKey" ) {
    ml::Vector tmp;
    if( cmdArgs>>tmp ){ keys.push_back(tmp); }
  }
  else if ( cmdLine=="clear" ) {
    keys.clear();
  }
  else if ( cmdLine=="compute" ) {
    computeParameters();
  }
  else { sotEntity::commandLine( cmdLine,cmdArgs,os ); }
}


void sotMotion::spline(Poly& Pi, const ml::Vector& b0, const ml::Vector& b1,
		   const ml::Vector& b2, const ml::Vector& df)
{
  Pi.a0 = b0;
  Pi.a1 = df;

  ml::Vector a2_plus_a3 = b2 - Pi.a0 - Pi.a1;
  Pi.a2 = 8 * (b1 - Pi.a0) - 4 * Pi.a1 - a2_plus_a3;
  Pi.a3 = a2_plus_a3 - Pi.a2;
  Pi.order = 3;
}

void sotMotion::spline(Poly& Pi, const std::vector< ml::Vector >& b)
{
  Pi.a0 = b[0];
  Pi.a1 = -5.5*b[0] + 9*b[1] - 4.5*b[2] + b[3];
  Pi.a2 = 9*b[0] - 22.5*b[1] + 18*b[2] - 4.5*b[3];
  Pi.a3 = -4.5*b[0] + 13.5*b[1] - 13.5*b[2] + 4.5*b[3];
  Pi.order = 3;
}

void sotMotion::quadratic(Poly& Pi, const std::vector< ml::Vector >& b)
{
  Pi.a0 = b[0];
  ml::Vector a1_plus_a2 = b[2] - Pi.a0;
  Pi.a1 = 4 * b[1] - 4 * Pi.a0 - a1_plus_a2;
  Pi.a2 = a1_plus_a2 - Pi.a1;
  Pi.order = 2;
}

void sotMotion::computeParameters()
{
  polys.clear();

  Poly Pi;
  unsigned nkeys = keys.size();

  if(nkeys == 0){ return; }
  else if(nkeys == 1) {
    Pi.order = 0;
    Pi.a0 = keys[0];
    polys.push_back(Pi);
    return;
  }
  else if(nkeys == 2) {
    Pi.order = 1;
    Pi.a0 = keys[0];
    Pi.a1 = keys[1] - keys[0];
    polys.push_back(Pi);
    return;
  }
  else if(nkeys == 3) {
    quadratic(Pi, keys);
    polys.push_back(Pi);
    return;
  }
  else if(nkeys == 4) {
    spline(Pi, keys);
    polys.push_back(Pi);
    return;
  }

  ml::Vector tmp(3); tmp.fill(0);

  unsigned i = 0;
  for(; i < nkeys-2; i+=2) {
    spline(Pi, keys[i], keys[i+1], keys[i+2], tmp);
    tmp = Pi.derivative(tmp, 1);
    polys.push_back(Pi);
  }

  /* there is a remaining key point */
  if(i == nkeys-2) {
    i -= 2;
    Pi.a0 = keys[i];
    Pi.a1 = tmp;
    Pi.a2 = keys[i+1] - Pi.a0 - Pi.a1;
    Pi.order = 2;
    polys.push_back(Pi);
  }
}

ml::Vector& sotMotion::Poly::compute(ml::Vector& res, double t)
{
  res.resize(3); res.fill(0);

  if(order == 0) {
    res = a0;
  }
  else if(order == 1) {
    res = a0 + a1 * t;
  }
  else if(order == 2) {
    res = a0 + a1 * t + a2 * t * t;
  }
  else if(order == 3) {
    double t2 = t*t;
    res = a0 + a1*t + a2*t2 + a3*t2*t;
  }

  return res;
}

ml::Vector& sotMotion::Poly::derivative(ml::Vector& res, double t)
{
  res.resize(3); res.fill(0);

  if(order == 1) {
    res = a1;
  }
  else if(order == 2) {
    res = a1 + a2 * t * 2;
  }
  else if(order == 3) {
    res = a1 + a2*t*2 + a3*t*t*3;
  }

  return res;
}
