/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotSot.cpp
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* SOT */
// #define VP_DEBUG 
// #define VP_DEBUG_MODE 5
// #include <sot/sotDebug.h>
// class sotSOT__INIT 
// {
// public:sotSOT__INIT( void ) { sotDebugTrace::openFile(); }
// };
// sotSOT__INIT sotSOT_initiator;


#include <sot/sotSOT.h>
#include <sot/sotPool.h>
#include <sot/sotTask.h>


using namespace std;

/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */


#include <sot/sotFactory.h>
SOT_FACTORY_ENTITY_PLUGIN(sotSOT,"SOT");


const double sotSOT::INVERSION_THRESHOLD_DEFAULT = 1e-4;
const unsigned int sotSOT::NB_JOINTS_DEFAULT = 46;

/* --------------------------------------------------------------------- */
/* --- CONSTRUCTION ---------------------------------------------------- */
/* --------------------------------------------------------------------- */
sotSOT::
sotSOT( const std::string& name )
  :sotEntity(name)
  ,stack()
  ,constraintList()
  ,ffJointIdFirst( FF_JOINT_ID_DEFAULT )
  ,ffJointIdLast( FF_JOINT_ID_DEFAULT+6 )
  
  ,nbJoints( NB_JOINTS_DEFAULT )
  ,taskGradient(0)
  ,q0SIN( NULL,"sotSOT("+name+")::input(double)::q0" )
   ,inversionThresholdSIN( NULL,"sotSOT("+name+")::input(double)::damping" )
   ,constraintSOUT( boost::bind(&sotSOT::computeConstraintProjector,this,_1,_2),
		   sotNOSIGNAL,
		    "sotSOT("+name+")::output(matrix)::constraint" )
  ,controlSOUT( boost::bind(&sotSOT::computeControlLaw,this,_1,_2),
		constraintSOUT<<inversionThresholdSIN<<q0SIN,
		"sotSOT("+name+")::output(vector)::control" )
{
  inversionThresholdSIN = INVERSION_THRESHOLD_DEFAULT;

  signalRegistration( inversionThresholdSIN<<controlSOUT<<constraintSOUT<<q0SIN );
}

/* --------------------------------------------------------------------- */
/* --- STACK MANIPULATION --- */
/* --------------------------------------------------------------------- */
void sotSOT::
push( sotTaskAbstract& task )
{
  stack.push_back( &task );
  controlSOUT.addDependancy( task.taskSOUT );
  controlSOUT.addDependancy( task.jacobianSOUT );
  //controlSOUT.addDependancy( task.featureActivationSOUT );
  controlSOUT.setReady();
}
sotTaskAbstract& sotSOT::
pop( void )
{
  sotTaskAbstract* res = stack.back();
  stack.pop_back();
  controlSOUT.removeDependancy( res->taskSOUT );
  controlSOUT.removeDependancy( res->jacobianSOUT );
  controlSOUT.removeDependancy( res->featureActivationSOUT );
  controlSOUT.setReady();
  return *res;
}
bool sotSOT::
exist( const sotTaskAbstract& key )
{
  std::list<sotTaskAbstract*>::iterator it;
  for ( it=stack.begin();stack.end()!=it;++it )
    {
      if( *it == &key ) { return true; }
    }
  return false; 
}
void sotSOT::
remove( const sotTaskAbstract& key )
{
  bool find =false; std::list<sotTaskAbstract*>::iterator it;
  for ( it=stack.begin();stack.end()!=it;++it )
    {
      if( *it == &key ) { find=true; break; }
    }
  if(! find ){ return; }

  stack.erase( it );
  removeDependancy( key );
}

void sotSOT::
removeDependancy( const sotTaskAbstract& key )
{
  controlSOUT.removeDependancy( key.taskSOUT );
  controlSOUT.removeDependancy( key.jacobianSOUT );
  //controlSOUT.removeDependancy( key.featureActivationSOUT );
  controlSOUT.setReady();
}

void sotSOT::
up( const sotTaskAbstract& key )
{
  bool find =false; std::list<sotTaskAbstract*>::iterator it;
  for ( it=stack.begin();stack.end()!=it;++it )
    {
      if( *it == &key ) { find=true; break; }
    }
  if( stack.begin()==it ) { return; }
  if(! find ){ return; }

  std::list<sotTaskAbstract*>::iterator pos=it; pos--;
  sotTaskAbstract * task = *it;
  stack.erase( it );
  stack.insert( pos,task );
  controlSOUT.setReady();
}
void sotSOT::
down( const sotTaskAbstract& key )
{
  bool find =false; std::list<sotTaskAbstract*>::iterator it;
  for ( it=stack.begin();stack.end()!=it;++it )
    {
      if( *it == &key ) { find=true; break; }
    }
  if( stack.end()==it ) { return; }
  if(! find ){ return; }

  std::list<sotTaskAbstract*>::iterator pos=it; pos++;
  sotTaskAbstract* task=*it;
  stack.erase( it );
  if( stack.end()==pos ){ stack.push_back(task); }
  else
    {
      pos++;
      stack.insert( pos,task );
    }
  controlSOUT.setReady();
}

void sotSOT::
clear( void )
{ 
  for (  std::list<sotTaskAbstract*>::iterator it=stack.begin();stack.end()!=it;++it )
    {
      removeDependancy( **it );
    }
  stack.clear();  
  controlSOUT.setReady();
}

/* --------------------------------------------------------------------- */
/* --- CONSTRAINTS ----------------------------------------------------- */
/* --------------------------------------------------------------------- */
 
void sotSOT::
addConstraint( sotConstraint& constraint )
{
  constraintList.push_back( &constraint ); 
  constraintSOUT.addDependancy( constraint.jacobianSOUT );
}

void sotSOT::
removeConstraint( const sotConstraint& key )
{
  bool find =false; ConstraintListType::iterator it;
  for ( it=constraintList.begin();constraintList.end()!=it;++it )
    {
      if( *it == &key ) { find=true; break; }
    }
  if(! find ){ return; }

  constraintList.erase( it );

  constraintSOUT.removeDependancy( key.jacobianSOUT );
}
void sotSOT::
clearConstraint( void ) 
{
  for (  ConstraintListType::iterator it=constraintList.begin();
	 constraintList.end()!=it;++it )
    {
      constraintSOUT.removeDependancy( (*it)->jacobianSOUT );
    }
  constraintList.clear(); 
}

void sotSOT::
defineFreeFloatingJoints( const unsigned int& first,const unsigned int& last )
{
  ffJointIdFirst = first ;
  if( last>0 ) ffJointIdLast=last ;
  else ffJointIdLast=ffJointIdFirst+6;
}

/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

static ml::Matrix & computeJacobianConstrained( const sotTaskAbstract& task,
						const ml::Matrix& K )
{
  const ml::Matrix &Jac = task.jacobianSOUT;
  const unsigned int nJ = Jac.nbRows();
  const unsigned int mJ = K.nbCols();
  const unsigned int FF_SIZE = Jac.nbCols() - mJ;

  ml::Matrix &Jff = task.Jff;
  ml::Matrix &Jact = task.Jact;
  ml::Matrix &JK = task.JK;

  for( unsigned int i=0;i<nJ;++i )
    {
      for( unsigned int j=0;j<FF_SIZE;++j ) Jff(i,j)=Jac(i,j);
      for( unsigned int j=FF_SIZE;j<Jac.nbCols();++j ) 
	Jact(i,j-FF_SIZE)=Jac(i,j);
    }
  Jff.multiply(K,JK);
  JK+=Jact;

  return JK;
}

static void computeJacobianActivated( sotTask* taskSpec, 
				      ml::Matrix& Jt,
				      const int& iterTime )
{
  if( NULL!=taskSpec )
    {
      const sotFlags& controlSelec = taskSpec->controlSelectionSIN( iterTime );
      sotDEBUG(25) << "Control selection = " << controlSelec <<endl;
      if( controlSelec )
	{
	  if(!controlSelec)
	    {
	      sotDEBUG(15) << "Control selection."<<endl;
	      for( unsigned int i=0;i<Jt.nbCols();++i )
		{
		  if(! controlSelec(i) )
		    {for( unsigned int j=0;j<Jt.nbRows();++j ) { Jt(j,i)=0.; }}
		}
	    }
	  else
	    { 
	      sotDEBUG(15) << "S is equal to Id."<<endl;
	    }
	}
      else 
	{
	  sotDEBUG(15) << "Task not activated."<<endl;
	  Jt *= 0; 
	}
    }
  else { /* No selection specification: nothing to do. */ }
}


/* --------------------------------------------------------------------- */
/* --- CONTROL --------------------------------------------------------- */
/* --------------------------------------------------------------------- */


// #define WITH_CHRONO


#ifdef  WITH_CHRONO
	#ifndef WIN32
	#include <sys/time.h>
	#else /*WIN32*/
	// When including Winsock2.h, the MAL must be included first
	//#include <MatrixAbstractLayer/boost.h>
	#include <sot/sotUtilsWindows.h>
	#include <Winsock2.h>
	#endif /*WIN32*/
#endif /*WITH_CHRONO*/


#ifdef  WITH_CHRONO
#   define sotINIT_CHRONO1 struct timeval t0,t1; double dt
#   define sotSTART_CHRONO1  gettimeofday(&t0,NULL)
#   define sotCHRONO1(a)			\
      gettimeofday(&t1,NULL);\
      dt = ( (t1.tv_sec-t0.tv_sec) * 1000.* 1000.\
	     + (t1.tv_usec-t0.tv_usec+0.)  );\
      sotDEBUG(1) << "dt_" << a << ": "<< dt << std::endl
#   define sotINITPARTCOUNTERS  struct timeval tpart0
#   define sotSTARTPARTCOUNTERS  gettimeofday(&tpart0,NULL)
#   define sotCOUNTER(nbc1,nbc2) \
	  gettimeofday(&tpart##nbc2,NULL); \
	  dt##nbc2 += ( (tpart##nbc2.tv_sec-tpart##nbc1.tv_sec) * 1000.* 1000. \
		   + (tpart##nbc2.tv_usec-tpart##nbc1.tv_usec+0.)  )
#   define sotINITCOUNTER(nbc1) \
   struct timeval tpart##nbc1; double dt##nbc1=0;
#   define sotPRINTCOUNTER(nbc1)  sotDEBUG(1) << "dt" << nbc1 << " = " << dt##nbc1 << std::endl
#else // #ifdef  WITH_CHRONO
#   define sotINIT_CHRONO1
#   define sotSTART_CHRONO1
#   define sotCHRONO1(a)
#   define sotINITPARTCOUNTERS
#   define sotSTARTPARTCOUNTERS
#   define sotCOUNTER(nbc1,nbc2)
#   define sotINITCOUNTER(nbc1) 
#   define sotPRINTCOUNTER(nbc1)
#endif // #ifdef  WITH_CHRONO



ml::Vector& sotSOT::
computeControlLaw( ml::Vector& control,const int& iterTime )
{
  sotDEBUGIN(15);

  sotINIT_CHRONO1; sotINITPARTCOUNTERS;
  sotINITCOUNTER(1); sotINITCOUNTER(2);sotINITCOUNTER(3);sotINITCOUNTER(4);
  sotINITCOUNTER(5); sotINITCOUNTER(6);sotINITCOUNTER(7);sotINITCOUNTER(8);
  sotINITCOUNTER(9);

  sotSTART_CHRONO1;

  const double &th = inversionThresholdSIN(iterTime);
  const ml::Matrix &K = constraintSOUT(iterTime);
  const unsigned int mJ = K.nbCols();

  try { 
    control = q0SIN( iterTime ); 
    sotDEBUG(15) << "initial velocity q0 = " << control << endl; 
    if( mJ!=control.size() ) { control.resize( mJ ); control.fill(.0); }
  } 
  catch (...) 
    {
      if( mJ!=control.size() ) { control.resize( mJ ); }
      control.fill(0.);
      sotDEBUG(25) << "No initial velocity." <<endl; 
    }

  sotDEBUGF(5, " --- Time %d -------------------", iterTime );
  sotCHRONO1("init");
   
  unsigned int iterTask = 0; 
  for( StackType::iterator iter = stack.begin(); iter!=stack.end();++iter )
    {
      sotSTARTPARTCOUNTERS;
      sotDEBUGF(5,"Rank %d.",iterTask); 
      sotTaskAbstract & task = **iter;
      sotDEBUG(15) << "Task: e_" << task.getName() << std::endl;
      const ml::Matrix &Jac = task.jacobianSOUT(iterTime);
      const ml::Vector &err = task.taskSOUT(iterTime);
      /***/sotCOUNTER(0,1); // Direct Dynamic
        
      unsigned int rankJ;
      const unsigned int nJ = Jac.nbRows();

      ml::Matrix &Jp = task.Jp;
      ml::Matrix &V = task.V;
      ml::Matrix &JK = task.JK;
      ml::Matrix &Jt = task.Jt;

      Jp.resize( mJ,nJ ); 
      V.resize( mJ,mJ );
      Jt.resize( nJ,mJ );
      task.JK.resize( nJ,mJ );

      if( (task.jacobianSOUT.getTime()>task.jacobianInvSINOUT.getTime())
	  ||(task.jacobianInvSINOUT.accessCopy().nbRows()!=mJ)
	  ||(task.jacobianInvSINOUT.accessCopy().nbCols()!=nJ)
	  ||(task.jacobianSOUT.getTime()>task.jacobianConstrainedSINOUT.getTime())
	  ||(task.jacobianSOUT.getTime()>task.rankSINOUT.getTime())
	  ||(task.jacobianSOUT.getTime()>task.singularBaseImageSINOUT.getTime()) )
      {
	sotDEBUG(5) <<"Recompute inverse."<<endl;

	/* --- FIRST ALLOCS --- */
	ml::Vector &S = task.S;

	sotDEBUG(15) << "Size = " 
		    << S.size() + task.Jff.nbCols()*task.Jff.nbRows()
	  + task.Jact.nbCols()*task.Jact.nbRows() << std::endl;

	S.resize( min(nJ,mJ) );
	task.Jff.resize( nJ,Jac.nbCols()-mJ );
	task.Jact.resize( nJ,mJ );

	sotDEBUG(15) << "Size = " 
		    << S.size() + task.Jff.nbCols()*task.Jff.nbRows()
	  + task.Jact.nbCols()*task.Jact.nbRows() << std::endl;

	/***/sotCOUNTER(1,2); // first allocs 
	
	/* --- COMPUTE JK --- */
	computeJacobianConstrained( task,K );
	/***/sotCOUNTER(2,3); // compute JK
	
	/* --- COMPUTE Jt --- */
	if( 0<iterTask ) JK.multiply(Proj,Jt); else { Jt = JK; } 
	/***/sotCOUNTER(3,4); // compute Jt
	
	/* --- COMPUTE S --- */
	computeJacobianActivated( dynamic_cast<sotTask*>( &task ),Jt,iterTime );
	/***/sotCOUNTER(4,5); // Jt*S
	
	/* --- PINV --- */
	Jt.dampedInverse( Jp,th,NULL,&S,&V );
	/***/sotCOUNTER(5,6); // PINV
	sotDEBUG(45) << "V after dampedInverse." << V <<endl;
	/* --- RANK --- */
	{
	  const unsigned int Jmax = S.size(); rankJ=0;
	  for( unsigned i=0;i<Jmax;++i ) { if( S(i)>th ) rankJ++; }
	} 
	
	sotDEBUG(45) << "control"<<iterTask<<" = "<<control<<endl;
	sotDEBUG(25) << "J"<<iterTask<<" = "<<Jac<<endl;
	sotDEBUG(25) << "JK"<<iterTask<<" = "<<JK<<endl;
	sotDEBUG(25) << "Jt"<<iterTask<<" = "<<Jt<<endl;
	sotDEBUG(15) << "Jp"<<iterTask<<" = "<<Jp<<endl;
	sotDEBUG(15) << "e"<<iterTask<<" = "<<err<<endl;
	sotDEBUG(45) << "JJp"<<iterTask<<" = "<< JK*Jp <<endl;
	//sotDEBUG(45) << "U"<<iterTask<<" = "<< U<<endl;
	sotDEBUG(45) << "S"<<iterTask<<" = "<< S<<endl;
	sotDEBUG(45) << "V"<<iterTask<<" = "<< V<<endl;

	task.jacobianInvSINOUT = Jp;
	task.jacobianInvSINOUT.setTime( iterTime );
	task.jacobianConstrainedSINOUT = JK;
	task.jacobianConstrainedSINOUT.setTime( iterTime );
	task.jacobianProjectedSINOUT = Jt;
	task.jacobianProjectedSINOUT.setTime( iterTime );
	task.singularBaseImageSINOUT = V;
	task.singularBaseImageSINOUT.setTime( iterTime );
	task.rankSINOUT = rankJ;
	task.rankSINOUT.setTime( iterTime );

	sotDEBUG(25)<<"Inverse recomputed."<<endl;
      } else {
	sotDEBUG(2)<<"Inverse not recomputed."<<endl;
	rankJ = task.rankSINOUT.accessCopy();
	Jp = task.jacobianInvSINOUT.accessCopy();
	V = task.singularBaseImageSINOUT.accessCopy();
	JK = task.jacobianConstrainedSINOUT;
	Jt = task.jacobianProjectedSINOUT;
      }
      /***/sotCOUNTER(6,7); // TRACE


      /* --- COMPUTE QDOT AND P --- */
      /*DEBUG: normally, the first iter (ie the test below)
      * is the same than the other, starting with control_0 = q0SIN. */
      if( iterTask==0 ) control += Jp*err; else 
	control += Jp*(err - JK*control);
      /***/sotCOUNTER(7,8); // QDOT 
      

      /* --- OPTIMAL FORM: To debug. --- */
      if( 0==iterTask ) 
	{ Proj.resize( mJ,mJ ); Proj.setIdentity(); } 
      
      /* --- OLIVIER START  --- */    
      if (1)
       {
	 double *p,*v1,*v2,*vtmp1,*vtmp2;
	 p = traits::matrix_storage(Proj.matrix);
	 v1 = traits::matrix_storage(V.matrix);
	 v2 = traits::matrix_storage(V.matrix);
	 vtmp1 = traits::matrix_storage(V.matrix);
	 
	 for( unsigned int i=0;i<mJ;++i )
	   {
	     vtmp2 = traits::matrix_storage(V.matrix); 
	     for( unsigned int j=0;j<mJ;++j )
	       {
		 v1 = vtmp1;   v2 =vtmp2;
		 for( unsigned int k=0;k<rankJ;++k )
		   //for( unsigned int k=0;k<mJ;++k )
		   { 
		     (*p) -=( *v1) * (*v2); 
		     v2+=mJ;v1+=mJ;
		   } 
		 p++; vtmp2 ++;
	       }
	     vtmp1++;
	   }
       }
      else
	{ Proj-=Jp*Jt;}
      /***/sotCOUNTER(8,9); // P
      
      /* --- OLIVIER END --- */
      
      sotDEBUG(15) << "q"<<iterTask<<" = "<<control<<std::endl;
      sotDEBUG(25) << "P"<<iterTask<<" = "<< Proj <<endl;
      iterTask++;
      
      sotPRINTCOUNTER(1);     sotPRINTCOUNTER(2);    sotPRINTCOUNTER(3);
      sotPRINTCOUNTER(4);     sotPRINTCOUNTER(5);    sotPRINTCOUNTER(6);
      sotPRINTCOUNTER(7);     sotPRINTCOUNTER(8);    sotPRINTCOUNTER(9);   
      
    }
  
  sotCHRONO1("tasks");

  if( 0!=taskGradient )
    {
      const ml::Vector & err = taskGradient->taskSOUT.access(iterTime);
      const ml::Vector & h = taskGradient->featureActivationSOUT.access(iterTime);
      const ml::Matrix & Jac = taskGradient->jacobianSOUT.access(iterTime);

      const unsigned int nJ = Jac.nbRows();

      sotDEBUG(45) << "K = " << K <<endl;
      sotDEBUG(45) << "Jff = " << Jac <<endl;

      /* --- MEMORY INIT --- */
      ml::Matrix &Jp = taskGradient ->Jp; 
      ml::Matrix &PJp = taskGradient ->PJp; 
      ml::Matrix &Jt = taskGradient ->Jt; 

      taskGradient->JK.resize( nJ,mJ );
      taskGradient->Jt.resize( nJ,mJ );
      taskGradient->Jff.resize( nJ,Jac.nbCols()-mJ );
      taskGradient->Jact.resize( nJ,mJ );
      Jp.resize( mJ,nJ ); 
      PJp.resize( nJ,mJ );

     /* --- COMPUTE JK --- */
      ml::Matrix &JK = computeJacobianConstrained( *taskGradient,K);

      /* --- COMPUTE Jinv --- */
      sotDEBUG(35) << "grad = " << err <<endl;
      sotDEBUG(35) << "h = " << h <<endl;
      sotDEBUG(35) << "Jgrad = " << JK <<endl;

      // Use optimized-memory Jt to do the p-inverse.
      Jt=JK; Jt.dampedInverse( Jp,th ); 
      Proj.multiply( Jp,PJp );

      /* --- COMPUTE ERR --- */
      ml::Vector Herr( err.size() );
      for( unsigned int i=0;i<err.size(); ++i )
	{
	  if( h(i)<=0 ) Herr(i) = 0.;
	  else if( h(i) >= 1. ) Herr(i) = err(i);
	  else Herr(i) = err(i) * h(i) ;
	}

      /* --- COMPUTE CONTROL --- */
      control += PJp*Herr;

      /* ---  TRACE  --- */
      sotDEBUG(45) << "Pgrad = " << (PJp*Herr) <<endl;
      sotDEBUG(45) << "P = " << Proj <<endl;
      sotDEBUG(45) << "Jp = " << Jp <<endl;
      sotDEBUG(45) << "PJp = " << PJp <<endl;
    }

  sotCHRONO1("total");
  sotDEBUGOUT(15);
  return control;
}



ml::Matrix& sotSOT::
computeConstraintProjector( ml::Matrix& ProjK, const int& time )
{
  sotDEBUGIN(15);
  const ml::Matrix *Jptr;
  if( 0==constraintList.size() )
    { 
      const unsigned int FF_SIZE = ffJointIdLast-ffJointIdFirst;
      ProjK.resize( FF_SIZE,nbJoints-FF_SIZE ); ProjK.fill(.0);
      sotDEBUGOUT(15);
      return ProjK;
    }
  else if( 1==constraintList.size() )
    { Jptr = &(*constraintList.begin())->jacobianSOUT(time); }
  else 
    {
      SOT_THROW sotExceptionTask( sotExceptionTask::EMPTY_LIST, 
				  "Not implemented yet." );
    }

  const ml::Matrix &J = *Jptr;
  sotDEBUG(12) << "J = "<< J;

  const unsigned int nJc = J.nbCols();
  ml::Matrix Jff( J.nbRows(),ffJointIdLast-ffJointIdFirst );
  ml::Matrix Jc( J.nbRows(),nJc-ffJointIdLast+ffJointIdFirst );
  
  for( unsigned int i=0;i<J.nbRows();++i )
    {
      if( ffJointIdFirst ) 
	for( unsigned int j=0;j<ffJointIdFirst;++j )
	  Jc(i,j)=J(i,j);
      if( ffJointIdLast<nJc )
	for( unsigned int j=ffJointIdLast;j<nJc;++j )
	    Jc(i,j+ffJointIdFirst-ffJointIdLast)=J(i,j);
      for( unsigned int j=ffJointIdFirst;j<ffJointIdLast;++j )
	Jff( i,j-ffJointIdFirst )=J(i,j);
    }
  sotDEBUG(25) << "Jc = "<< Jc;
  sotDEBUG(25) << "Jff = "<< Jff;

  ml::Matrix Jffinv( Jff.nbCols(),Jff.nbRows() );
  Jff.pseudoInverse( Jffinv );   Jffinv *= -1;

  ml::Matrix& Jffc = ProjK;
  Jffc.resize( Jffinv.nbRows(),Jc.nbCols() );
  Jffinv.multiply( Jc,Jffc );
  sotDEBUG(15) << "Jffc = "<< Jffc;

  sotDEBUGOUT(15);
  return ProjK;
}


/* --------------------------------------------------------------------- */
/* --- DISPLAY --------------------------------------------------------- */
/* --------------------------------------------------------------------- */
void sotSOT::
display( std::ostream& os ) const
{
  
  os << "+-----------------"<<std::endl<<"+   SOT     "
     << std::endl<< "+-----------------"<<std::endl;
  for ( std::list<sotTaskAbstract*>::const_iterator it=this->stack.begin();
	this->stack.end()!=it;++it )
    {
      os << "| " << (*it)->getName() <<std::endl;
    }
  os<< "+-----------------"<<std::endl;
  if( taskGradient )
    {
      os << "| {Grad} " <<taskGradient->getName() << std::endl;
      os<< "+-----------------"<<std::endl;
    }
  for( ConstraintListType::const_iterator it = constraintList.begin();
       it!=constraintList.end();++it )
    { os<< "| K: "<< (*it)->getName() << endl; }


}

std::ostream& 
operator<< ( std::ostream& os,const sotSOT& sot )
{
  sot.display(os);
  return os;
}

/* --------------------------------------------------------------------- */
/* --- COMMAND --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

void sotSOT::
commandLine( const std::string& cmdLine,std::istringstream& cmdArgs,
	     std::ostream& os )
{
  sotDEBUGIN(15);

  if( cmdLine == "help") 
    {
      os << "Stack of Tasks: "<<endl
	 << " - push <task>"<< endl
	 << " - pop"<< endl
	 << " - down <task>"<< endl
	 << " - up <task>"<< endl
	 << " - rm <task>"<< endl
	 << " - clear"<< endl
	 << " - display"<< endl

	 << " - addConstraint <constraint> "<<endl
	 << " - rmConstraint <constraint> "<<endl
	 << " - clearConstraint"<<endl
	 << " - printConstraint "<<endl

	 << " - nbJoints <nb> "<<endl;
      sotEntity::commandLine( cmdLine,cmdArgs,os );
    }
  else if( cmdLine == "clear") 
    {
      clear(); 
    }
  else if( cmdLine == "push") 
    {
      std::string tname; cmdArgs >> tname; 
      sotTaskAbstract & task = sotPool.getTask( tname );
      push(task); 
    }
  else if( cmdLine == "gradient") 
    {
      std::string tname; cmdArgs >> std::ws;
      if( cmdArgs.good() )
	{
	  cmdArgs >> tname; 
	  if( ( "0"==tname )||( "rm"==tname ) )
	    { taskGradient = 0; }
	  else 
	    {
	      sotTaskAbstract & task = sotPool.getTask( tname );
	      taskGradient = &task;
	    }
	}
      else 
	{
	  os << "gradient = ";
	  if( taskGradient ) os << (*taskGradient) << std::endl;
	  else os << "undef. " << std::endl;
	}	
      controlSOUT.setReady();
    }
  else if( cmdLine == "up") 
    {
      std::string tname; cmdArgs >> tname; 
      sotTaskAbstract & task = sotPool.getTask( tname );
      up(task);
    }
  else if( cmdLine == "down") 
    {
      std::string tname; cmdArgs >> tname; 
      sotTaskAbstract & task = sotPool.getTask( tname );
      down(task);
    }
  else if( cmdLine == "rm") 
    {
      std::string tname; cmdArgs >> tname; 
      sotTaskAbstract & task = sotPool.getTask( tname );
      remove(task);
    }
  else if( cmdLine == "pop") 
    {
      sotTaskAbstract& task = pop();
      os << "Remove : "<< task << std::endl;
    }
  
  else if( cmdLine == "addConstraint" )
    {
      std::string cstname; cmdArgs >> cstname;
      sotConstraint &cs = dynamic_cast<sotConstraint&>(sotPool.getTask( cstname ));
      addConstraint( cs );
      constraintSOUT.setReady();
    }
  else if( cmdLine == "rmConstraint" )
    {
      std::string cstname; cmdArgs >> cstname;
      sotConstraint &cs = dynamic_cast<sotConstraint&>(sotPool.getTask( cstname ));
      removeConstraint( cs );
      constraintSOUT.setReady();
    }
  else if( cmdLine == "clearConstraint" )
    {
      clearConstraint( );
      constraintSOUT.setReady();
    }
  else if( cmdLine == "printConstraint" )
    {
      os<< "Constraints: "<<std::endl;
      for( ConstraintListType::iterator it = constraintList.begin();
	   it!=constraintList.end();++it )
	{ os<< "  - "<< (*it)->getName() << endl; }
    }
  else if( cmdLine == "nbJoints") 
    {
      cmdArgs>>ws;
      if( cmdArgs.good() )
	{
	  cmdArgs >> nbJoints;
	} else { os << "nbJoints = "<< nbJoints <<endl; }
      constraintSOUT.setReady();
      controlSOUT.setReady();
    }
  else if( cmdLine == "display") 
    {
      display(os);
    }
  else
    sotEntity::commandLine( cmdLine,cmdArgs,os );
  
  
  sotDEBUGOUT(15);
}

std::ostream& sotSOT::
writeGraph( std::ostream& os ) const
{
  std::list<sotTaskAbstract *>::const_iterator iter;
  for(  iter = stack.begin(); iter!=stack.end();++iter )
    {
      const sotTaskAbstract & task = **iter;
      std::list<sotTaskAbstract *>::const_iterator nextiter =iter;
      nextiter++;
      
      if (nextiter!=stack.end())
	{
	  sotTaskAbstract & nexttask = **nextiter;
	  os << "\t\t\t" << task.getName() << " -> " << nexttask.getName() << " [color=red]" << endl;
	}

    }

  os << "\t\tsubgraph cluster_Tasks {" <<endl;
  os << "\t\t\tsubgraph cluster_" << getName() << " {" << std::endl;
  os << "\t\t\t\tcolor=lightsteelblue1; label=\"" << getName() <<"\"; style=filled;" << std::endl;
  for(  iter = stack.begin(); iter!=stack.end();++iter )
    {
      const sotTaskAbstract & task = **iter;
      os << "\t\t\t\t" << task.getName() 
		<<" [ label = \"" << task.getName() << "\" ," << std::endl 
		<<"\t\t\t\t   fontcolor = black, color = black, fillcolor = magenta, style=filled, shape=box ]" << std::endl; 
      
    }
  os << "\t\t\t}" << std::endl;
  os << "\t\t\t}" <<endl;
  return os;
}
