/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotReproConstraints.cpp
 * Project:   SOT
 * Author:    Paul Evrard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 * Outputs a freeze signal when velocity / force / workspace
 * constraints are violated.
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <sot/sotReproConstraints.h>
#include <sot/sotTaskLearningTools.h>

using namespace std;

#include <sot/sotFactory.h>
SOT_FACTORY_ENTITY_PLUGIN(sotReproConstraints, "ReproConstraints");


sotReproConstraints::
sotReproConstraints( const std::string& fName )
  : sotEntity( fName )

    /*
     * INPUT SIGNALS
     */
  ,positionSIN( NULL,"sotReproConstraints("+name+")::input(vector)::positionIN" )
  ,velocitySIN( NULL,"sotReproConstraints("+name+")::input(vector)::velocityIN")
  ,forceSIN( NULL,"sotReproConstraints("+name+")::input(vector)::forceIN")

    /*
     * OUTPUT SIGNALS
     */
  ,freezeSOUT( boost::bind(&sotReproConstraints::computeOutput,this,_1,_2),
		 positionSIN<<velocitySIN<<forceSIN,
		"sotReproConstraints("+name+")::output(vector)::freeze" )  

    /*
     * DATA
     */
  ,workspace(6)
  ,maxVelocity(6)
  ,maxForce(6)
  ,freezed(false)
{
  // x, y, z in m
  workspace(0) = -0.2; workspace(1) = 0.2;
  workspace(2) = -0.1; workspace(3) = 0.1;
  workspace(4) = -0.1; workspace(5) = 0.5;

  // force in N, torque in N.m
  maxForce(0) = 150; maxForce(1) = 150; maxForce(2) = 150;
  maxForce(3) = 20; maxForce(4) = 20; maxForce(5) = 20;

  // vel in m/s, ang vel in rad/s
  maxVelocity(0) = 1; maxVelocity(1) = 1; maxVelocity(2) = 1;
  maxVelocity(3) = 1.57; maxVelocity(4) = 1.57; maxVelocity(5) = 1.57;

  signalRegistration( positionSIN<<velocitySIN<<forceSIN<<freezeSOUT );
}

int& sotReproConstraints::computeOutput( int& res, int time )
{
  const ml::Vector& pos = positionSIN.access(time);
  const ml::Vector& vel = velocitySIN.access(time);
  const ml::Vector& force = forceSIN.access(time);

  freezed = false;

  if( !between(pos(0), workspace(0), workspace(1)) ||
      !between(pos(1), workspace(2), workspace(3)) ||
      !between(pos(2), workspace(4), workspace(5)) ) {
    freezed = true;
  }

  for( int i=0; (i<6) && !freezed; ++i ) {
    if( (abs(vel(i)) >= maxVelocity(i)) ||
	(abs(force(i)) >= maxForce(i)) ) {
      freezed = true;
    }
  }

  return (res = freezed ? 1 : 0);
}

void sotReproConstraints::display(std:: ostream& os ) const
{
  os << "ReproConstraints<" << name << ">" << endl;
  os << " - freeze: " << freezed << endl;
}

void sotReproConstraints::commandLine( const std::string& cmdLine,
		    std::istringstream& cmdArgs,
				   std::ostream& os )
{
  if (cmdLine == "help")
  {
    os << "ReproConstraints: " << endl
       << " -setWorkspace [minx maxx miny maxy minz maxz]" << endl
       << " -maxForce [maxf]" << endl
       << " -maxVelocity [maxv]" << endl
       << " -unfreeze" << endl;
  }
  else if (cmdLine == "workspace") {
    readVector<6>(cmdArgs, os, "workspace", workspace);
  }
  else if (cmdLine == "maxForce") {
    readVector<6>(cmdArgs, os, "maxForce", maxForce);
  }
  else if (cmdLine == "maxVelocity") {
    readVector<6>(cmdArgs, os, "maxVelocity", maxVelocity);
  }
  else if(cmdLine == "unfreeze") {
    freezed = false;
  }
  else { sotEntity::commandLine( cmdLine,cmdArgs,os ); }
}
