/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotTaskLearning.cpp
 * Project:   SOT
 * Author:    Who can it be?
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 * What can it do?
 *
 * !!! WARNING !!! THIS CODE IS CURRENTLY SPECIFIC TO SOMETHING
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#include <sot/sotTaskLearning.h>
#include <sot/sotTaskLearningTools.h>

using namespace std;

#include <sot/sotFactory.h>
SOT_FACTORY_ENTITY_PLUGIN(sotTaskLearning,"TaskLearning");

sotTaskLearning::
sotTaskLearning( const string& fName )
  : sotEntity( fName )

    /*
     * INPUT SIGNALS
     */
  ,freezeSIN( NULL,"sotTaskLearning("+name+")::input(vector)::freezeIN" )
  ,positionSIN( NULL,"sotTaskLearning("+name+")::input(vector)::positionIN" )
  ,velocitySIN( NULL,"sotTaskLearning("+name+")::input(vector)::velocityIN")
  ,forceSIN( NULL,"sotTaskLearning("+name+")::input(vector)::forceIN")

    /*
     * OUTPUT SIGNALS
     */
  ,velocitySOUT( boost::bind(&sotTaskLearning::computeOutput,this,_1,_2),
		 positionSIN<<velocitySIN<<forceSIN,
		"sotTaskLearning("+name+")::output(vector)::velocity" )  

    /*
     * DATA
     */
  ,pos(0)
  ,dpos(0)
  ,force(0)
  ,nIter(0)
  ,run(false)
  ,freezed(false)

  ,mode(MODE_KINEMATIC)
  ,nIterMax(1000)
  ,dt(0.005)
  ,beta(0.03)
{
  signalRegistration( freezeSIN<<positionSIN<<velocitySIN
		      <<forceSIN<<velocitySOUT );
}

ml::Vector&
sotTaskLearning::computeOutput( ml::Vector& res,int time )
{
  res.resize(6); res.fill(0);

  int freezeIN = freezeSIN.access(time);

  if(run && !freezeIN) {
    const ml::Vector& posIN = positionSIN.access(time);
    const ml::Vector& forceIN = forceSIN.access(time);
    for( int i=0;i<3;++i ){ pos[i]=posIN(i); force[i]=forceIN(i); }

    // if first iteration, init GMR
    if(!nIter) {
      for( int i=0;i<3;++i ){ dpos[i] = 0; }
      lifting->SetInitialPosition(pos.buf(), dpos.buf());
    }

    // perform GMR
    if(mode == MODE_KINEMATIC){ dpos = lifting->Step(pos.buf()); }
    else if(mode == MODE_COMPLETE){ dpos = lifting->Step(pos.buf(), force.buf()); }

    // the output is a position increment, and the sot expects a velocity.
    const double invdt = 200;
    for( int i=0;i<3;++i ){ res(i) = dpos[i] * invdt; }

    if(++nIter > nIterMax){ stop(); }
  }
  else if(freezeIN){ stop(); freezed = true; }

  return res;
}

void sotTaskLearning::display(std:: ostream& os ) const
{
  os << "TaskLearning<" << name << ">" << endl;
  os << " - run: " << run << endl;
  os << " - file: " << file << endl;
  os << " - mode: " << mode << endl;
  os << " - dt: " << dt << endl;
  os << " - beta: " << beta << endl;
  os << " - nIterMax: " << nIterMax << endl;
  os << " - freezed: " << freezed << endl;
  os << " - nIter: " << nIter << endl;
}

void sotTaskLearning::commandLine( const std::string& cmdLine,
		    std::istringstream& cmdArgs,
				   std::ostream& os )
{
  if(cmdLine == "help")
  {
    os << "TaskLearning: " << endl
       << "reproduces a task using GMR"<< endl
       << " -start" << endl
       << " -stop" << endl
       << " -file" << endl
       << " -mode" << endl
       << " -dt" << endl
       << " -beta" << endl
       << " -nIterMax" << endl;
  }
  else if(cmdLine == "start"){ start(); }
  else if(cmdLine == "stop"){ stop(); }
  else if(cmdLine == "file"){ readElm<std::string>(cmdArgs, os, "file", file); }
  else if(cmdLine == "mode"){ setMode(cmdArgs, os); }
  else if(cmdLine == "dt"){ readElm<double>(cmdArgs, os, "dt", dt); }
  else if(cmdLine == "beta"){ readElm<double>(cmdArgs, os, "beta", beta); }
  else if(cmdLine == "nIterMax"){ readElm<int>(cmdArgs, os, "nIterMax", nIterMax); }
  else { sotEntity::commandLine( cmdLine,cmdArgs,os ); }
}

void sotTaskLearning::start()
{
  if(file.length()) {
    char* filename = const_cast<char*>(file.c_str());
    int nbVar = 1;
    const int dim = 1;

    switch(mode)
    {
    case MODE_KINEMATIC:
    case MODE_FORCE: nbVar = 1; break;
    case MODE_COMPLETE: nbVar = 2; break;
    default: nbVar = 1;
    }

    lifting.reset(new Motion(nbVar, dim, filename, dt, beta, mode));

    freezed = false;
    run = true;
    nIter = 0;
  }
}

void sotTaskLearning::stop()
{
  freezed = false;
  run = false;
  nIter = 0;
}

void sotTaskLearning::setFile( std::istringstream& cmdArgs, std::ostream& os)
{
  std::string tmp;
  if((cmdArgs>>tmp) && tmp.length()) {
    file = tmp;
  }
  os << "file: " << file << endl;
}

void sotTaskLearning::setMode( std::istringstream& cmdArgs, std::ostream& os)
{
  std::string tmp;
  if((cmdArgs>>tmp) && tmp.length()) {
    if(tmp == "kinematic"){ mode = MODE_KINEMATIC; }
    else if(tmp == "force"){ mode = MODE_FORCE; }
    else if(tmp == "complete"){ mode = MODE_COMPLETE; }
  }
  os << "mode: " << mode << " (" << tmp << ")" << endl;
}
