/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotVec.h
 * Project:   SOT
 * Author:    Paul Evrard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 * Wrapper for a buffer of T with arithmetic operations,
 * useful for the new_buf method which is much like
 * the stringstream::str method (it assigns another buffer).
 * The reason for this class is that the GMR library from EPFL
 * returns a vector as dynamically allocated buffer, pointed
 * by a raw pointer. The use of this calss to store the
 * output through a call to Vec::new_buf ensured proper
 * deletion of the buffer.
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

#ifndef _SOT_VEC_HH_
#define _SOT_VEC_HH_

#include <algorithm>
#include <iterator>

template<int N, typename T = double>
class Vec
{
public:

  typedef T type_t;
  enum { size = N };
  typedef Vec<size, type_t> Vec_t;

public:

  Vec(): buf_(new type_t[size]) {}

  Vec(type_t args): buf_(new type_t[size])
  {
    for(int i = 0; i < N; ++i){ buf_[i] = args; }
  }

  Vec(const Vec_t& v): buf_(new type_t[size])
  {
    std::copy(v.buf_, v.buf_ + size, buf_);
  }

  Vec_t& operator=(const Vec_t& v)
  {
    std::copy(v.buf_, v.buf_ + size, buf_);
    return *this;
  }

  Vec_t& operator=(const type_t* buf)
  {
    std::copy(buf, buf + size, buf_);
    return *this;
  }

public:

  ~Vec() { delete[] buf_; }

public:

  void print(std::ostream& os) const
  {
    std::copy(buf_, buf_ + size, std::ostream_iterator<type_t>(os, " "));
  }

  void new_buf(type_t* buf)
  {
    delete[] buf_;
    buf_ = buf;
  }

  type_t& operator[](size_t i) { return buf_[i]; }
  const type_t& operator[](size_t i) const { return buf_[i]; }

  type_t* buf() { return buf_; }
  const type_t* buf() const { return buf_; }

public:

  Vec_t& operator+=(const Vec_t& v)
  {
    for(int i = 0; i < size; ++i){ buf_[i] += v.buf_[i]; }
    return *this;
  }

  Vec_t& operator-=(const Vec_t& v)
  {
    for(int i = 0; i < size; ++i){ buf_[i] -= v.buf_[i]; }
    return *this;
  }

public:

  Vec_t& operator+=(type_t el)
  {
    for(int i = 0; i < size; ++i){ buf_[i] += el; }
    return *this;
  }

  Vec_t& operator-=(type_t el)
  {
    for(int i = 0; i < size; ++i){ buf_[i] -= el; }
    return *this;
  }

  Vec_t& operator*=(type_t el)
  {
    for(int i = 0; i < size; ++i){ buf_[i] *= el; }
    return *this;
  }

  Vec_t& operator/=(type_t el)
  {
    for(int i = 0; i < size; ++i){ buf_[i] /= el; }
    return *this;
  }

private:

  type_t* buf_;
};


template<int N, typename T>
Vec<N, T>& operator+(const Vec<N, T>& v1, const Vec<N, T>& v2)
{
  Vec<N, T> res(v1);
  return res += v2;
}

template<int N, typename T>
Vec<N, T>& operator-(const Vec<N, T>& v1, const Vec<N, T>& v2)
{
  Vec<N, T> res(v1);
  return res -= v2;
}

template<int N, typename T>
Vec<N, T>& operator+(const Vec<N, T>& v1, T el)
{
  Vec<N, T> res(v1);
  return res += el;
}

template<int N, typename T>
Vec<N, T>& operator+(T el, const Vec<N, T>& v2)
{
  Vec<N, T> res(v2);
  return res += el;
}

template<int N, typename T>
Vec<N, T>& operator-(const Vec<N, T>& v1, T el)
{
  Vec<N, T> res(v1);
  return res -= el;
}

template<int N, typename T>
Vec<N, T>& operator*(const Vec<N, T>& v1, T el)
{
  Vec<N, T> res(v1);
  return res *= el;
}

template<int N, typename T>
Vec<N, T>& operator*(T el, const Vec<N, T>& v2)
{
  Vec<N, T> res(v2);
  return res *= el;
}

template<int N, typename T>
Vec<N, T>& operator/(const Vec<N, T>& v1, T el)
{
  Vec<N, T> res(v1);
  return res /= el;
}

template<int N, typename T>
std::ostream& operator<<(std::ostream& os, const Vec<N, T>& v)
{
  v.print(os);
  return os;
}

#endif
