/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotCollisionDetector.cpp
 * Project:   SOT
 * Author:    Paul Evrard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* --- STL  --- */
#include <algorithm>
#include <iterator>

/* --- COLLISION DETECTION --- */
#include <CollisionDetection/HumanoidRobotCollisionDetection.h>

/* --- Matrix Abstract Layer v1 ---*/
#include <MatrixAbstractLayer/MatrixAbstractLayer.h>

/* --- SOT --- */
#include <sot/sotPool.h>
#include <sot/sotDynamic.h>
#include <sot/sotCollisionDetector.h>
#include <sot/sotExceptionFeature.h>
#include <sot/sotDebug.h>
using namespace std;

#include <sot/sotFactory.h>
SOT_FACTORY_ENTITY_PLUGIN(sotCollisionDetector,"CollisionDetector");

/* --------------------------------------------------------------------- */
/* --- COLLISION DETECTION --------------------------------------------- */
/* --------------------------------------------------------------------- */

struct CollisionDetectionImpl
{
  sotDynamic* dynamic;
  HumanoidRobotCollisionDetection collisionDetection;
  std::vector<PaireOfBodies> setOfPairs;
  std::vector<int> jointsToFreeze;
  double threshold; // safety margin in meters

  static const double DEFAULT_THRESHOLD;
  static const char* DEFAULT_DYN_NAME;

  CollisionDetectionImpl()
    : dynamic(NULL)
    , threshold(DEFAULT_THRESHOLD)
  {
    collisionDetection.LoadStructureForTheHumanoid("data");
    try {
      sotEntity& entity = sotPool.getEntity(DEFAULT_DYN_NAME);
      dynamic = dynamic_cast<sotDynamic*>(&entity);
    }
    catch(...) {
      sotERROR << "Info: Could not find entity <dyn>. Don't forget to set <dyn> entity" <<endl;
      dynamic = NULL;
    }
  }
};

std::ostream& operator<<(std::ostream& os, const PaireOfBodies& pob)
{
  os << pob.body1 << " " << pob.body2;
  return os;
}

/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */

const double CollisionDetectionImpl::DEFAULT_THRESHOLD = 0.01;
const char* CollisionDetectionImpl::DEFAULT_DYN_NAME = "dyn";


sotCollisionDetector::
sotCollisionDetector( const string& fName )
  : sotEntity( fName )
  ,jointSIN( NULL,"sotCollisionDetector("+name+")::input(vector)::joint" )
  ,controlSIN( NULL,"sotCollisionDetector("+name+")::input(vector)::controlIN" )
  ,controlSOUT( boost::bind(&sotCollisionDetector::computeControl,this,_1,_2),
		jointSIN<<controlSIN,
		"sotCollisionDetector("+name+")::output(vector)::control" )  

  ,cdimpl(new CollisionDetectionImpl)
{
  signalRegistration( jointSIN<<controlSIN<<controlSOUT );
}

/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

ml::Vector&
sotCollisionDetector::computeControl( ml::Vector& qdot,int time )
{
  sotDEBUGIN(15);

  const ml::Vector q = jointSIN.access(time);
  const ml::Vector qdotIN = controlSIN.access(time);
  unsigned int SIZE = qdotIN.size();

  qdot.resize(SIZE);
  for( unsigned i=0; i<SIZE; ++i ){ qdot(i) = qdotIN(i); }

  if(!cdimpl->dynamic) {
    sotERROR << "CollisionDetector "<<name<<"  error: dyn not set!" <<endl;
    return qdot;
  }

  Vect3 aP2;
  Mat3 aR2;
  VclipPose aX;
  djj::DynamicMultiBody* aDMB = cdimpl->dynamic->aHDMB;

  for( int i=0; i<28; i++ ) {
      matrix4d aCurrentM = aDMB->GetJointFromVRMLID(i)->currentTransformation();

      aP2[0] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,0,3);
      aP2[1] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,1,3);
      aP2[2] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,2,3);

      aR2[0][0] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,0,0); 
      aR2[0][1] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,0,1); 
      aR2[0][2] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,0,2);
      aR2[1][0] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,1,0); 
      aR2[1][1] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,1,1); 
      aR2[1][2] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,1,2);
      aR2[2][0] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,2,0);
      aR2[2][1] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,2,1);
      aR2[2][2] = MAL_S4x4_MATRIX_ACCESS_I_J(aCurrentM,2,2);

      aX.set(aR2,aP2);
      cdimpl->collisionDetection.SetBodyPose(i+1,aX);
  } // Yeepeekayay

  double dmin = cdimpl->collisionDetection.ComputeSelfCollision();
  if( dmin < cdimpl->threshold ) {
    for( unsigned i=0; i<cdimpl->jointsToFreeze.size(); ++i ){ qdot(cdimpl->jointsToFreeze[i]) = 0; }
  }

  sotDEBUGOUT(15);
  return qdot;
}

void sotCollisionDetector::
display( std::ostream& os ) const
{
  os <<"CollisionDetector <"<<name<<">" <<endl
     << " - safety margin is " << cdimpl->threshold << " m" <<endl
     << " - joints to freeze are: ";

  std::copy(cdimpl->setOfPairs.begin(), cdimpl->setOfPairs.end(), 
	    std::ostream_iterator<PaireOfBodies>(os, "\n"));

  os <<endl
     << " - registered pairs are" <<endl;

  std::copy(cdimpl->jointsToFreeze.begin(), cdimpl->jointsToFreeze.end(),
	    std::ostream_iterator<int>(os, " "));

  os <<endl;
}

void sotCollisionDetector::
commandLine( const std::string& cmdLine,
	     std::istringstream& cmdArgs,
	     std::ostream& os )
{
  if( cmdLine == "help" )
    {
      os << "CollisionDetector: "<<endl
	 << "  - jointsToFreeze <index1, index2, ...> " <<endl
	 << "  - addPair index1 index2 " <<endl
	 << "  - threshold <thres> " <<endl
	 << "  - pairs" <<endl
	 << "  - dynName name: loads dynamic entity with given name" <<endl
	;
    }
  else if( cmdLine == "jointsToFreeze" )
    {
      int index = 0;
      if( !(cmdArgs>>index) ) {
	os <<endl;
	std::copy(cdimpl->jointsToFreeze.begin(), cdimpl->jointsToFreeze.end(),
		  std::ostream_iterator<int>(os, " "));
	os <<endl;
      }
      else {
	do{ cdimpl->jointsToFreeze.push_back(index); } while( cmdArgs>>index );
      }
    }
  else if( cmdLine == "addPair" )
    {
      int index1, index2;
      if( !(cmdArgs>>index1) || !(cmdArgs>>index2) ) { 
	os << "Usage: addPair index1 index2" <<endl
	   << "where index1 and index2 are integers" <<endl;
      }
      else {
	PaireOfBodies pob = {index1, index2};
	cdimpl->setOfPairs.push_back(pob);
	cdimpl->collisionDetection.InsertACollisionPair(pob);
      }
    }
  else if( cmdLine == "pairs" )
    {
      os <<endl;
      std::copy(cdimpl->setOfPairs.begin(), cdimpl->setOfPairs.end(), 
		std::ostream_iterator<PaireOfBodies>(os, "\n"));
    }
  else if ( cmdLine == "threshold" )
    {
      double tmp = 0;
      if( !(cmdArgs>>tmp) ) {
	os << "safety margin is " << cdimpl->threshold << " m" <<endl;
      }
      else{ cdimpl->threshold = tmp; }
    }
  else if( cmdLine == "dynName" )
    {
      std::string dynname;
      if( !(cmdArgs>>dynname) ) {
	os << "could not parse entity name" <<endl;
      }
      else {
	try {
	  sotEntity& entity = sotPool.getEntity(dynname); 
	  cdimpl->dynamic = dynamic_cast<sotDynamic*>(&entity);
	}
	catch(...) {
	  os << "Could not find entity named " << dynname <<endl;
	  cdimpl->dynamic = NULL;
	}
      }
    }
  else { sotEntity::commandLine( cmdLine,cmdArgs,os ); }
}
