/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotMotionPeriod.cpp
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* --- SOT --- */
#include <sot/sotMotionPeriod.h>
#include <sot/sotExceptionFeature.h>
#include <sot/sotDebug.h>
using namespace std;

#include <sot/sotFactory.h>
SOT_FACTORY_ENTITY_PLUGIN(sotMotionPeriod,"MotionPeriod");

#include <MatrixAbstractLayer/MatrixAbstractLayer.h>

/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */

sotMotionPeriod::
sotMotionPeriod( const string& fName )
  : sotEntity( fName )
    ,motionParams( 0 )
    ,motionSOUT( boost::bind(&sotMotionPeriod::computeMotion,this,_1,_2),
		 sotNOSIGNAL,
		 "sotMotionPeriod("+name+")::output(vector)::motion" )
{
  signalRegistration( motionSOUT );
  motionSOUT.setNeedUpdateFromAllChildren( true );
  resize(0);
}


/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

ml::Vector&
sotMotionPeriod::computeMotion( ml::Vector& res,const int& time )
{
  sotDEBUGIN(15);

  res.resize( size );
  for( unsigned int i=0;i<size;++i )
    {
      const sotMotionParam & p = motionParams[i];
      double x= ( ( ((time-p.initPeriod) % p.period) +0.0 ) 
		  / (p.period+0.0) );
      res(i)=p.initAmplitude;
      switch( p.motionType )
	{
	case MOTION_CONSTANT: { res(i)+= p.amplitude;  break; }
	case MOTION_SIN: { res(i)+= p.amplitude*sin(M_PI*2*x); break; }
	case MOTION_COS: { res(i)+= p.amplitude*cos(M_PI*2*x); break; }
	  //case MOTION_: {res(i)+= p.amplitude; break}
	}
    }

  sotDEBUGOUT(15);
  return res;
}

void sotMotionPeriod::
resize( const unsigned int & _size ) 
{
  size = _size;
  motionParams.resize(size);
  for( unsigned int i=0;i<size;++i )
    {
      motionParams[i] .motionType = MOTION_CONSTANT;
      motionParams[i] .amplitude = 0;
      motionParams[i] .initPeriod = 0;
      motionParams[i] .period = 1;
      motionParams[i] .initAmplitude = 0;
    }
}


void sotMotionPeriod::
display( std::ostream& os ) const
{
  os <<"MotionPeriod <"<<name<<"> ... TODO";
}


#define SOT_PARAMS_CONFIG(ARGname,ARGtype)                                              \
  else if( cmdLine == #ARGname )                                                        \
   {                                                                                    \
     unsigned int rank; ARGtype period;                                                 \
     cmdArgs >> rank >>std::ws;                                                         \
     if( rank>=this->size ) { os <<"!! Error: size size too large." << std::endl; }     \
     if( cmdArgs.good() )                                                               \
     { cmdArgs>>period; motionParams[rank].ARGname = period; }                             \
     else { os << #ARGname << "[" << rank << "] = "                                     \
               <<  motionParams[rank].ARGname << std::endl; }                              \
    }


void sotMotionPeriod::
commandLine( const std::string& cmdLine,
	     std::istringstream& cmdArgs,
	     std::ostream& os )
{
  if( cmdLine == "help" )
    {
      os << "MotionPeriod:"<<endl
	 <<"  - size  " <<endl
	 <<"  - period  <rank> [<value>]" <<endl
	 <<"  - amplitude  <rank> [<value>] " <<endl
	 <<"  - init  <rank> [<value>] " <<endl
	 <<"  - type  <rank> [const|sin|cos] " <<endl;
    }
//   else if( cmdLine == "period" )
//     {
//       unsigned int rank; unsigned int period; 
//       cmdArgs >> rank >>std::ws; 
//       if( rank>=this->size ) { os <<"!! Error size too large." << std::endl; }
//       if( cmdArgs.good() ) 
// 	{ cmdArgs>>period; motionParams[rank].period = period; }
//       else { os << "period[" << rank << "] = " <<  motionParams[rank].period << std::endl; }
//     }
  SOT_PARAMS_CONFIG(period,unsigned int)
    SOT_PARAMS_CONFIG(amplitude,double)
    SOT_PARAMS_CONFIG(initPeriod,unsigned int)
    SOT_PARAMS_CONFIG(initAmplitude,double)
    else if( cmdLine == "size" )
      {
	unsigned int rank; 
	cmdArgs >> std::ws; 
	if( cmdArgs.good() ) 
	  { cmdArgs>>rank; resize( rank ); }
	else { os << "size = " << rank << std::endl; }
      }
  else if( cmdLine == "type" )
    {
      unsigned int rank; std::string typeName;
      cmdArgs >> rank >>std::ws; 
      if( rank>=this->size ) { os <<"!! Error size too large." << std::endl; }
      if( cmdArgs.good() ) 
	{
	  cmdArgs>> typeName;
	  if( "const" == typeName ) { motionParams[rank].motionType = MOTION_CONSTANT; }
	  else if( "sin" == typeName ) { motionParams[rank].motionType = MOTION_SIN; }
	  else if( "cos" == typeName ) { motionParams[rank].motionType = MOTION_COS; }
	  //else if( "" == typeName ) { motionParams[rank].motionType = MOTION_; }
	  else { os << "!! Motion type invalid." << std::endl; }
	}
      else 
	{ 
	  os << "type[" << rank << "] = ";
	  switch( motionParams[rank].period )
	    {
	    case MOTION_CONSTANT: { os << "const" << std::endl; break; }
	    case MOTION_SIN: { os << "sin" << std::endl; break; }
	    case MOTION_COS: { os << "cos" << std::endl; break; }
	      //case MOTION_: { os << "" << std::endl; break; }
	    }
	}
    }
  else { sotEntity::commandLine( cmdLine,cmdArgs,os ); }
}



/*
 * Local variables:
 * c-basic-offset: 2
 * End:
 */
