/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Copyright Projet JRL-Japan, 2007
 *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *
 * File:      sotTracer.cpp
 * Project:   SOT
 * Author:    Nicolas Mansard
 *
 * Version control
 * ===============
 *
 *  $Id$
 *
 * Description
 * ============
 *
 *
 * ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/


/* --------------------------------------------------------------------- */
/* --- INCLUDE --------------------------------------------------------- */
/* --------------------------------------------------------------------- */

/* SOT */
#include <sot/sotTracer.h>
#include <sot/sotDebug.h>
#include <sot/sotPool.h>

#include <boost/bind.hpp>

using namespace std;

#include <sot/sotFactory.h>
SOT_FACTORY_ENTITY_PLUGIN(sotTracer,"Tracer");

/* --------------------------------------------------------------------- */
/* --- CLASS ----------------------------------------------------------- */
/* --------------------------------------------------------------------- */

sotTracer::sotTracer( const std::string n )
  :sotEntity(n)
   ,toTraceSignals()
   ,traceStyle(TRACE_STYLE_DEFAULT)
   ,frequency(1)
   ,basename()
   ,suffix(".dat")
   ,rootdir()
   ,namesSet( false )
   ,files()
   ,names()
   ,play(false)
   ,timeStart(0)
   ,triger( boost::bind(&sotTracer::recordTrigger,this,_1,_2),
	    sotNOSIGNAL,
	   "Tracer("+n+")::triger" )
{
  signalRegistration( triger );
}  

/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */


void sotTracer::
addSignalToTrace( const sotSignalAbstract<int>& sig,
		  const string& filename )
{
  sotDEBUGIN(15);
  toTraceSignals.push_back( &sig ); sotDEBUGF(15,"%p",&sig);
  names.push_back( filename );
  if( namesSet ) openFile( sig,filename );
  triger.addDependancy( sig );
  sotDEBUGOUT(15);
} 

/*! Empty the list of signals to trace. This function
 * does not modify the file list (it does not close
 * the files in particular. 
 */
void sotTracer::
clearSignalToTrace( void )
{
  toTraceSignals.clear();
  triger.clearDependancies();
}


// void sotTracer::
// parasite( sotSignalAbstract<int>& sig )
// {
//   triger.parasite(sig);
// }

void sotTracer::
openFiles( const std::string& rootdir_, const std::string& basename_,
	   const std::string& suffix_  )
{
  sotDEBUGIN(15);
  int n = rootdir_.length();
  rootdir=rootdir_;
  if( (0<n)&('/'!=rootdir[n-1]) ) rootdir+='/';
  
  basename=basename_;
  suffix=suffix_;

  if( files.size() ) closeFiles();

  SignalList::const_iterator iter = toTraceSignals.begin();
  NameList::const_iterator iterName = names.begin();
  while( toTraceSignals.end()!=iter )
    {
      sotDEBUG(15) << "Open <" << (*iter)->getName() 
		   << "> in <" << *iterName << ">." << std::endl;
      openFile( **iter,*iterName );    
      ++iter; ++iterName;
    }

  namesSet = true;
  sotDEBUGOUT(15);
}

void sotTracer::
openFile( const sotSignalAbstract<int> & sig,
	  const string& givenname )
{
  sotDEBUGIN(15);
  string signame;
  if( givenname.length() )  
    { signame = givenname;  } else { signame = sig.shortName(); }

  string filename = rootdir + basename + signame + suffix;

  sotDEBUG(5) << "Sig <"<< sig.getName() << ">: new file "<< filename << endl;
  std::ofstream * newfile = new std::ofstream( filename.c_str() );
  files.push_back( newfile );
  sotDEBUGOUT(15);
}


void sotTracer::
closeFiles( void )
{
  sotDEBUGIN(15);
  
  for( FileList::iterator iter = files.begin();files.end()!=iter;++iter )
    {
      std::ostream * filePtr = *iter;
      delete filePtr;
    }
  files.clear();

  sotDEBUGOUT(15);
}

/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */

void sotTracer::
record( void )
{
  if(! play) { sotDEBUGINOUT(15); return;} 

  sotDEBUGIN(15);
 
  if( files.size()!=toTraceSignals.size() )
    { SOT_THROW sotExceptionTraces( sotExceptionTraces::NOT_OPEN,
				    "No files open for tracing"," (file=%d != %d=sig).",
				    files.size(),toTraceSignals.size()); }

  FileList::iterator iterFile = files.begin();
  SignalList::iterator iterSig = toTraceSignals.begin();

  while( toTraceSignals.end()!=iterSig )
    {
      sotDEBUG(45) << "Try..." <<endl;
      recordSignal( **iterFile,**iterSig ); 
      ++iterSig;      ++iterFile;
    }
  sotDEBUGOUT(15);
}

void sotTracer::
recordSignal(  std::ostream& os,
	       const sotSignalAbstract<int>& sig )
{
  sotDEBUGIN(15);
  
  try { 
    if( sig.getTime()>timeStart ) 
      {
	os<< sig.getTime() << "\t"; 
	sig.trace(os); os<<endl;
      }
  }
  catch( sotExceptionAbstract& exc ) { os << exc << std::endl; }
  catch( ... ) { os << "Unknown error occured while reading signal." << std::endl; }
  
   sotDEBUGOUT(15);

}


int& sotTracer::
recordTrigger( int& dummy, const int& time )
{
  sotDEBUGIN(15) << "    time="<<time <<endl;
  record();
  sotDEBUGOUT(15);
  return dummy;
}


void sotTracer::
trace( void )
{
}

/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */
/* --------------------------------------------------------------------- */


void sotTracer::
display( std::ostream& os ) const
{
  os << CLASS_NAME << " " << name << " [mode=" << (play?"play":"pause") 
     << "] : "<< endl
     << "  - Dep list: "<<endl;
  for( SignalList::const_iterator iter = toTraceSignals.begin();
       toTraceSignals.end()!=iter;++iter )
    {      os << "     -> "<<(*iter)->getName()<<endl;    } 
}


std::ostream& operator<< ( std::ostream& os,const sotTracer& t )
{
  t.display(os);
  return os;
}


void sotTracer::
commandLine( const std::string& cmdLine
	     ,std::istringstream& cmdArgs
	     ,std::ostream& os )
{
  if( cmdLine=="help" )
    {
      os << "Tracer: "<<endl
	 << "  - add <obj.signal> [<filename>=<signal>] "<<endl
	 << "  - clear"<<endl
	 << "  - open  [<root> [<file> [<suffix>=.dat]]]"<<endl
	 << "  - close "<<endl
	 << "  - record "<<endl
	 << "  - trace "<<endl
	 << "  - start/stop" << endl;
	//	 << "  - parasite <obj.signal> "<<endl;
	
      sotEntity::commandLine( cmdLine,cmdArgs,os );
    }
  else if( cmdLine=="add" )
    {
      sotSignalAbstract<int> &sig = sotPool.getSignal(cmdArgs);
      string r; cmdArgs>>ws>>r;
      addSignalToTrace(sig,r);
      sotDEBUG(14)<<"Add <" <<sig.getName()<<"> with nick \""<<r<<"\""<<endl;
    }
  else if( cmdLine=="clear" ) 
    { closeFiles(); toTraceSignals.clear(); }
  else if( cmdLine=="open" )
    {
      string n,r="",s=".dat"; 
      cmdArgs>>ws>>r;
      if( cmdArgs.good() )
	{
	  cmdArgs>>ws>>n;
	  if( cmdArgs.good() )
	    {
	      cmdArgs>>ws>>s;
	    }
	}

      //>>r>>s; 
      sotDEBUGF( 15,"Close files.");
      closeFiles();
      sotDEBUGF( 15,"Open files \"%s\" \"%s\" \"%s\".",
		r.c_str(),n.c_str(),s.c_str());
      openFiles(r,n,s);
    }
  else if( cmdLine=="close" ) { closeFiles(); }
  else if( cmdLine=="trace" ) { trace(); }
  else if( cmdLine=="record" ) 
    { 
      //unsigned int t;
      //cmdArgs >> ws>>t; if(! cmdArgs.good() ) t=0;
      record();
    }
//   else if( cmdLine=="parasite" ) 
//     {
//        sotSignalAbstract<int> &sig = sotPool.getSignal( cmdArgs );
//        parasite(sig);
//     }
  else if( cmdLine == "start" )    {  play=true; }
  else if( cmdLine == "stop" )    {  play=false; }
  else if( cmdLine == "timeStart" )    
    {
      cmdArgs >> std::ws; if(! cmdArgs.good() )
	{ os << "timeStart = " << timeStart << std::endl; }
      else { cmdArgs >> timeStart; }
    }
    

  else  //sotTaskAbstract::
    sotEntity::commandLine( cmdLine,cmdArgs,os );



}
